;;;; 	copyright (C) 1995 Free Software Foundation, Inc.
;;;; 
;;;; This program is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2, or (at your option)
;;;; any later version.
;;;; 
;;;; This program is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with this software; see the file COPYING.  If not, write to
;;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
;;;; 


(define (struct-type-printer obj port writing)
  (display "#(struct-type " port)
  (display (struct-type-name obj) port)
  (do ((c (struct-type-super obj) (struct-type-super c)))
      ((not c) #t)
    (display " : " port)
    (display (struct-type-name c)))
  (display ")" port))
  
(define struct-type-type (make-record-type 'struct-type '(name
							  keyvector
							  fields
							  super
							  predicate)
					   struct-type-printer))
(define struct-type-constructor (record-constructor struct-type-type))
(define struct-type-keyvector (record-accessor struct-type-type 'keyvector))
(define struct-type? (record-predicate struct-type-type))
(define (struct-type-is-subtype? t maybe-super)
  (and maybe-super t
       (or (eq? t maybe-super)
	   (struct-type-is-subtype? (struct-type-super t) maybe-super))))
(define (struct-predicate type)
  (or (struct-type-predicate type)
      (begin
	(set-struct-type-predicate!
	 type
	 (lambda (obj)
	   (and (struct? obj)
		(struct-type-is-subtype?  type)))))))

(define (make-struct-type name fields . opts)
  (let* ((super (and (not (eq? opts '())) (car opts)))
	 (answer (struct-type-constructor name
					  (make-struct-type-keyvector super fields)
					  fields
					  super
					  #f)))
    (vector-set! (struct-type-keyvector answer)
		 (+ -1 (length (struct-type-keyvector answer)))
		 answer)
    answer))

(define (make-struct struct-type . inits)
  (check-struct-type struct-type)
  (let* ((kv (struct-type-keyvector struct-type))
	 (v (if (null? inits)
		(make-vector (+ -1 (length kv)) #f)
		(let ((v (apply vector #f inits)))
		  (vector-set-length! v (length kv))
		  v))))
    (vector-set! v 0 kv)
    (lock-vector! v)
    v))

(define struct-type-name (record-accessor struct-type-type 'name))
(define struct-type-fields (record-accessor struct-type-type 'fields))
(define struct-type-super (record-accessor struct-type-type 'super))

(define (subclass-instance? obj type)
  (let loop ((t (struct-type obj)))
    (or (eq? t type)
	(and t (loop (struct-type-super t))))))

(define (struct-constructor type)
  (let* ((fields (struct-type-all-fields type))
	 (setters (map (lambda (f) (struct-fixpos-setter type f)) fields)))
    (lambda args
      (let ((it (make-struct type)))
	(map (lambda (s! v) (s! it v)) setters args)
	it))))



;;; {Converting Unexpected Types Into Errors}
;;;

(define (check-struct obj)
  (or (struct? obj)
      (error 'not-a-struct obj)))

(define (check-struct-type obj)
  (or (struct-type? obj)
      (error 'not-a-struct-type obj)))

(define (check-subclass-instance obj type)
  (or (subclass-instance? obj type)
      (throw 'wrong-type obj type)))



;;; {Reflective Stuff Used Internally}
;;;

(define struct-user-field-offset 1)

(define (field-posn type field)
  (+ struct-user-field-offset
     (or (list-index (struct-type-all-fields type) field)
	 (error 'field-not-defined type field))))

;; An fixpos accessor works without regard
;; to the type of the structure.  It always returns
;; the data at a particular offset, provided that
;; that field has the same name as the same field in type.
;;
(define (struct-fixpos-accessor type field)
  (let* ((n (field-posn type field))
	 (key (vector-ref (struct-type-keyvector type) n)))
    (lambda (obj) (lvector-ref obj key n))))

;; An fixpos setter works without regard
;; to the type of the structure.  (cf fixpos-accessor)
;;
(define (struct-fixpos-setter type field)
  (let* ((n (field-posn type field))
	 (key (vector-ref (struct-type-keyvector type) n)))
    (lambda (obj val) (lvector-set! obj key n val))))

(define (struct-type-lineage type)
  (let loop ((answer '())
	     (t type))
    (if t
	(loop (cons t answer)
	      (struct-type-super t))
	answer)))

(define (struct-type-all-fields type)
  (apply append (map struct-type-fields (struct-type-lineage type))))

(define struct-lvector-hooks
  (let ((kv (make-vector lvector-hook-slots lvector-ref)))
    (vector-set! kv 0 #f)
    (let ((hooks (make-vector lvector-hook-slots #f)))
      (vector-set! hooks lvector-hook-ref-fn
		   (lambda (lv key index . opt)
		     (if (null? opt)
			 (throw 'illegal-access lv key index)
			 (let* ((cache (car opt))
				(new-index (field-posn (struct-type lv) key)))
			   (set-cdr! cache new-index)
			   (lvector-ref lv key new-index)))))
      (vector-set! hooks lvector-hook-set-fn
		   (lambda (lv key index val . opt)
		     (if (null? opt)
			 (throw 'illegal-access lv key index)
			 (let* ((cache (car opt))
				(new-index (field-posn (struct-type lv) key)))
			   (set-cdr! cache new-index)
			   (lvector-set! lv key new-index val)))))
      (vector-set! hooks lvector-hook-print-fn #f)
      (vector-set! hooks lvector-hook-equal-fn #f)
      (vector-set! hooks lvector-hook-isa-fn #f)
      (vector-set! hooks 0 kv)
      (lock-vector! hooks)
      hooks)))

(define (make-struct-type-keyvector super fields)
  (let* ((keys (cons struct-lvector-hooks
		     (append! (struct-type-all-fields super)
			      (append! fields (list #f)))))
	 (keyvec (apply vector keys)))
    keyvec))
    
(define (struct-type obj)
  (check-struct obj)
  (vector-ref (lvector-keys obj) (+ -1 (length (lvector-keys obj)))))

(define <struct> (make-struct-type 'struct '()))
(define (struct? obj)
  (and (lvector? obj)
       (let ((keys (lvector-keys obj)))
	 (struct-type? (vector-ref keys (+ -1 (length keys)))))))

(define (struct-accessor field)
  (let ((cache (cons field 0)))
    (lambda (obj)
      (lvector-ref2 obj (car cache) (cdr cache) cache))))

(define (struct-modifier field)
  (let ((cache (cons field 0)))
    (lambda (obj val)
      (lvector-set! obj (car cache) (cdr cache) val cache))))

(defmacro -> (field struct)
  (list (struct-accessor field) struct))

(defmacro set->! (field struct val)
  (list (struct-modifier field) struct val))


