/*
** Copyright (c) 1993-1995 Xerox Corporation.  All Rights Reserved.
**
** Unlimited use, reproduction, and distribution of this software is
** permitted.  Any copy of this software must include both the above
** copyright notice of Xerox Corporation and this paragraph.  Any
** distribution of this software must comply with all applicable United
** States export control laws.  This software is made available AS IS,
** and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
** INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
** AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
** PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
** THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
** CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
** XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
*/
/* $Id: ilu.c,v 1.157 1997/01/31 19:47:42 janssen Exp $ */
/* Last tweaked by Mike Spreitzer January 16, 1997 1:51 pm PST */

#include <stdio.h>

#include <stdarg.h>

#if ((defined(WIN32) || defined(WIN16)) && defined(_WINIO))
#include <winiodef.h>
#endif /* ((defined(WIN32) || defined(WIN16)) && defined(_WINIO)) */

#include <string.h>
#include <stdlib.h>		/* for exit */

#include "ilucpvt.h"

#include <oscalls.h>		/* for OS_SLEEP */
#include <iludebug.h>		/* for _ilu_DebugLevel and LSR_DEBUG */
 
#include <iluhash.h>

#define  TRUE  1
#define  FALSE 0
#define  AND   &&
#define  OR    ||
#define  NOT   !

#define OPTIONAL(x)	x
#define PASS(x)		x
#define RETAIN(x)	x
#define GLOBAL(x)	x
 
#define SEQUENCE_INCREMENT 5

#define IOFNS_TABLE_SIZE	137

#define MOST_SPECIFIC_ILU_CLASS(cobj)	((cobj)->type)[0]->c
#define ILU_CLASS_OF_C_CLASS(cclass)	(cclass)[0]->c

/* L1, L2 unconstrained */

#define ILU_C_Logging (_ilu_DebugLevel & LSR_DEBUG)

static ILU_C_Object	_ILU_C_ORB[1];
CORBA_Object		ILU_C_ORB = _ILU_C_ORB;

static ilu_boolean c_registered = FALSE;
static ilu_cardinal c_lang = 47;

static ilu_Server defaultServer = ILU_NIL;

static ilu_Server GetDefaultServer(void);

static ilu_boolean threaded = ilu_FALSE;
static ilu_boolean threadedSet = ilu_FALSE;

static void     (*Fork) (void (*proc) (void *arg), void *arg) = 0;

static void MonitorOutgoingConnection(void *rock);

static ilu_cardinal MyLangIdx(void)
{
  if (!c_registered) {
    c_lang = ilu_RegisterLanguage("ANSI-C");
    c_registered = TRUE;
  }
  return c_lang;
}

#define ERRFMTSIZE 128

static const char *FmtError(char buf[ERRFMTSIZE], ilu_Error * err)
{
  int             mc, est;
  char            nbuf[24], *fmt;
  const char     *errname = ILU_ERR_NAME(*err);
  const char     *errfile = ilu_ErrorFile(err);
  static char    *fmt0 = "%s from line %d of %s";
  static char    *fmt1 = "%s (minor=%s) from line %d of %s";
  static char    *fmt2 = "interrupted (bitset=%s) from line %d of %s";
  static char    *fmt3 = "no_memory (nbytes=%s) from line %d of %s";
  ILU_ERR_SWITCH(*err) {
    ILU_SUCCESS_CASE return "SUCCESS";
    ILU_ERR_CASE(no_memory, x) {
      sprintf(nbuf, "%lu", x->nbytes);
      fmt = fmt3;
    }
    ILU_ERR_CASE(interrupted, x) {
      sprintf(nbuf, "0x%x", x->ilu_interruptSet);
      fmt = fmt2;
    }
    ILU_ERR_CASE(bad_param, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(imp_limit, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(comm_failure, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(inv_objref, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(internal, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(marshal, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(bad_typecode, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(bad_operation, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_CASE(no_resources, x) fmt = fmt1, mc = x->minor;
    ILU_ERR_ELSE    fmt = fmt0;
  } ILU_ERR_ENDSWITCH;
  if (fmt == fmt1)
    sprintf(nbuf, "%d", mc);
  est = (strlen(fmt) + strlen(errname) + 20 + strlen(errfile) +
	 ((fmt != fmt0) ? strlen(nbuf) : 0));
  if (est > ERRFMTSIZE)
    return errname;
  if (fmt == fmt0)
    sprintf(buf, fmt, errname, ilu_ErrorLine(err), errfile);
  else
    sprintf(buf, fmt, errname, nbuf, ilu_ErrorLine(err), errfile);
  return buf;
}

static void DisposeErrFull(ilu_Error * err, char *doing, int lineno)
{
  if (ILU_ERRNOK(*err)) {
    char            ef[ERRFMTSIZE];
    ILU_ERRPRINTF("Warning: %s ignored at line %d of %s while %s.\n",
		  FmtError(ef, err), lineno, __FILE__, doing);
    ILU_HANDLED(*err);
    ILU_CLER(*err);
  }
}

#define DisposeErr(e,d) DisposeErrFull((e), (d), __LINE__)


struct typeRecord_s {
  ilu_Class iluclass;
  ILU_C_ObjectType cclass;
  struct typeRecord_s *next;
};

typedef struct typeRecord_s *typeRecord;

static typeRecord TypeRegistry = ILU_NIL;

void _ILU_C_RegisterSurrogateCType (ilu_Class c, ILU_C_ObjectType t)
{
  typeRecord      r;

  r = (typeRecord) ilu_must_malloc(sizeof(*r));
  r->iluclass = c;
  r->cclass = t;
  r->next = TypeRegistry;
  TypeRegistry = r;
}

/* Inside(obj->server, obj->class) */
static ILU_C_Object *
_ILU_C_CreateSurrogateFromRegistry(ilu_Class c, ilu_Object obj)
{
  typeRecord      p;
  ILU_C_Object   *lspo = ILU_NIL;

  for (p = TypeRegistry; p != ILU_NIL; p = p->next)
    if (c == p->iluclass)
      break;
  if (p != ILU_NIL) {
    lspo = (ILU_C_Object *) ilu_must_malloc(sizeof(ILU_C_Object));
    lspo->server = ilu_ServerOfObject(obj);
    lspo->instanceId = (ilu_Object) obj;
    lspo->data = 0;
    lspo->type = p->cclass;
    lspo->interruptH = ILU_NIL;
    ilu_RegisterLanguageSpecificObject(obj, lspo, MyLangIdx());
  } else {
    ilu_DebugPrintf (
	    "ILU_C:  attempt to create surrogate instance of class \"%s\", but surrogate code is not loaded.\n",
	    c->cl_name);
  }
  return (lspo);
}
 
#ifdef ADD_VARIANT_SUPPORT

struct CORBA_any_s {
  CORBA_TypeCode _type;		/* NIL to indicate valid pickle instead of any */
  void * _value;		/* if _type is non-NIL, points to LSR value; otherwise undefined */
  ilu_Pickle _pickle;		/* if _type is NIL, contains ILU pickle; otherwise undefined */
};

ilu_string CORBA_TypeCode_id (CORBA_TypeCode tc, ILU_C_ENVIRONMENT *env)
{
  env->_major = ILU_C_NO_EXCEPTION;
  return (tc->type_id);
}

ilu_boolean CORBA_TypeCode_equal (CORBA_TypeCode tc1, CORBA_TypeCode tc2, ILU_C_ENVIRONMENT *env)
{
  env->_major = ILU_C_NO_EXCEPTION;
  return (strcmp(tc1->type_id, tc2->type_id) == 0);
}

void CORBA_TypeCode_free_value (CORBA_TypeCode tc, void *c_value, ILU_C_ENVIRONMENT *env)
{
  env->_major = ILU_C_NO_EXCEPTION;
  if (tc->type_kind == ilu_object_tk) {
  } else {
    if (tc->freeFn != ((ILU_C_FreeFn) 0)) {
      (*tc->freeFn)(c_value);
    };
    ilu_free(c_value);
  }
}

/*****  IoFns for predefined types *****/

#define OutputMacroC(fn, ktype, ctype) \
static void _ILU_C_Output ## fn (ilu_Call call, void *p, ilu_Error *e) \
{ \
    ilu_Output ## fn (call, (ktype) (*(ctype *)p), e); \
} \

#define OutputMacroS(fn, ktype, ctype) \
static void _ILU_C_Output ## fn (ilu_Call call, void *p, ilu_Error *e) \
{ \
    ilu_Output ## fn (call, (*(ktype *)p), e); \
} \

#define SizeOfMacroC(fn, ktype, ctype) \
static ilu_cardinal _ILU_C_SizeOf ## fn (ilu_Call call, void *p, ilu_Error *e) \
{ \
    return (ilu_SizeOf ## fn (call, (ktype) (*(ctype *)p), e)); \
} \

#define SizeOfMacroS(fn, ktype, ctype) \
static ilu_cardinal _ILU_C_SizeOf ## fn (ilu_Call call, void *p, ilu_Error *e) \
{ \
    return (ilu_SizeOf ## fn (call, (*(ktype *)p), e)); \
} \

#define InputMacro(fn, ktype, ctype) \
static void * _ILU_C_Input ## fn (ilu_Call call, void *p, ilu_Error * e) \
{ \
    ktype _temp1; ctype * _temp2; \
    if (p == ILU_NIL) { \
      _temp2 = (ctype *) ilu_MallocE(sizeof(ctype),e); \
      if (ILU_ERRNOK(*e)) return ILU_NIL; \
    } else { _temp2 = (ctype *) p; }; \
    ilu_Input ## fn (call, &_temp1, e); \
    if (ILU_ERROK(*e)) { *((ctype *) _temp2) = _temp1; return (void *) _temp2; }\
    else { if (((void *) _temp2) != p) ilu_free(_temp2); return ILU_NIL; } \
} \


#ifdef _MSC_VER

/* workaround for vc++ treating references imported from another dll as non-static */

#define RegisterIofnsMacro(fn, ktype, ctype) \
struct _ILU_C_IoFnsRegistration_s _ ## ctype ## __IoFns = { \
  ktype ## _tk, 0, \
  (ilu_Class) sizeof(ktype), \
  _ILU_C_SizeOf ## fn, _ILU_C_Output ## fn, _ILU_C_Input ## fn, 0 }; \

#define InitFnsRegistrationTypeIdMacro(ktype, ctype) \
	_ ## ctype ## __IoFns.type_id = (ilu_string) &ilu_TypeID_ ## ktype;

#else

#define RegisterIofnsMacro(fn, ktype, ctype) \
struct _ILU_C_IoFnsRegistration_s _ ## ctype ## __IoFns = { \
  ktype ## _tk, (ilu_string) &ilu_TypeID_ ## ktype, \
  (ilu_Class) sizeof(ktype), \
  _ILU_C_SizeOf ## fn, _ILU_C_Output ## fn, _ILU_C_Input ## fn, 0 }; \

#endif

#define ScalarTypeMacro(fn, ktype, ctype)	OutputMacroC(fn,ktype,ctype) \
						SizeOfMacroC(fn,ktype,ctype) \
						InputMacro(fn,ktype,ctype) \
						RegisterIofnsMacro(fn,ktype,ctype) \

#define SameTypeMacro(fn, ktype, ctype)	OutputMacroS(fn,ktype,ctype) \
					SizeOfMacroS(fn,ktype,ctype) \
					InputMacro(fn,ktype,ctype) \
					RegisterIofnsMacro(fn,ktype,ctype) \

ScalarTypeMacro(Boolean,ilu_boolean,CORBA_boolean)

ScalarTypeMacro(Byte,ilu_byte,CORBA_octet)

ScalarTypeMacro(ShortCharacter,ilu_shortcharacter,CORBA_char)
SameTypeMacro(Character,ilu_character,ilu_character)

ScalarTypeMacro(ShortCardinal,ilu_shortcardinal,CORBA_unsigned_short)
ScalarTypeMacro(Cardinal,ilu_cardinal,CORBA_unsigned_long)
SameTypeMacro(LongCardinal,ilu_longcardinal,ilu_longcardinal)

ScalarTypeMacro(ShortInteger,ilu_shortinteger,CORBA_short)
ScalarTypeMacro(Integer,ilu_integer,CORBA_long)
SameTypeMacro(LongInteger,ilu_longinteger,ilu_longinteger)

ScalarTypeMacro(ShortReal,ilu_shortreal,CORBA_float)
ScalarTypeMacro(Real,ilu_real,CORBA_double)
SameTypeMacro(LongReal,ilu_longreal,ilu_longreal)

static void
  _ILU_C_CString__Output (ilu_Call call,
			 ilu_string *string,
			 ilu_Error *err)
{
  ilu_OutputString(call, *string, _ILU_C_SafeStrlen(*string), 0, err);
}

static ilu_cardinal
  _ILU_C_CString__SizeOf (ilu_Call call,
			 ilu_string *string,
			 ilu_Error *err)
{
  return ilu_SizeOfString(call, *string, _ILU_C_SafeStrlen(*string), 0, err);
}

struct _ILU_C_IoFnsRegistration_s _CORBA_string__IoFns = {
	ilu_sequence_tk,
#ifdef _MSC_VER
		/* workaround for vc++ treating references imported from another dll as non-static */
		0,
#else
		(ilu_string) &ilu_TypeID_ilu_CString[0],
#endif
		(ilu_Class) (sizeof(ilu_string)),
		(ILU_C_SizeFn) _ILU_C_CString__SizeOf,
		(ILU_C_OutputFn) _ILU_C_CString__Output,
		(ILU_C_InputFn) _ilu_CString__Input,
		(ILU_C_FreeFn) ilu_CString__Free };

struct _ILU_C_IoFnsRegistration_s _CORBA_any__IoFns = {
	ilu_pickle_tk,
#ifdef _MSC_VER
	/* workaround for vc++ treating references imported from another dll as non-static */
	0,
#else
	(ilu_string) &ilu_TypeID_ilu_pickle[0],
#endif
	(ilu_Class) (sizeof(CORBA_any)),
	(ILU_C_SizeFn) _CORBA_any__SizeOf,
	(ILU_C_OutputFn) _CORBA_any__Output,
	(ILU_C_InputFn) _CORBA_any__Input,
	(ILU_C_FreeFn) CORBA_any__Free };

static HashTable _IoFns = ILU_NIL;

void _ILU_C_RegisterIoFns (ILU_C_IoFnsRegistration r)
{
  if (_IoFns == ILU_NIL)
    {
      _IoFns = _ilu_hash_MakeNewTable (IOFNS_TABLE_SIZE,
				       _ilu_hash_HashString,
				       _ilu_hash_StringCompare);
      if (_IoFns == ILU_NIL)
	{
	  ilu_DebugPrintf("Can't create type I/O functions registration table!\n");
	  exit(1);
	}

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_boolean[0], &_CORBA_boolean__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_byte[0], &_CORBA_octet__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_pickle[0], &_CORBA_any__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_shortinteger[0], &_CORBA_short__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_integer[0], &_CORBA_long__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_longinteger[0], &_ilu_longinteger__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_shortcardinal[0], &_CORBA_unsigned_short__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_cardinal[0], &_CORBA_unsigned_long__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_longcardinal[0], &_ilu_longcardinal__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_shortcharacter[0], &_CORBA_char__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_character[0], &_ilu_character__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_shortreal[0], &_CORBA_float__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_real[0], &_CORBA_double__IoFns);
      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_longreal[0], &_ilu_longreal__IoFns);

      _ilu_hash_AddToTable (_IoFns, (ilu_string) &ilu_TypeID_ilu_CString[0], &_CORBA_string__IoFns);

    }
  if (_ilu_hash_FindInTable(_IoFns, r->type_id) == ILU_NIL) {
    _ilu_hash_AddToTable(_IoFns, r->type_id, r);
  };
}

ILU_C_IoFnsRegistration _ILU_C_LookupIoFns (ilu_string type_uid)
{
  if (_IoFns == ILU_NIL)
    return ILU_NIL;
  return (ILU_C_IoFnsRegistration) _ilu_hash_FindInTable(_IoFns, type_uid);
}

static void
  _ILU_C_OutputValue (CORBA_TypeCode tc,
		      ilu_Call call,
		      void *value_ptr,
		      ilu_Error *err)
{
  if (tc->type_kind != ilu_object_tk)
    (*tc->outFn)(call, value_ptr, err);
  else
    _ILU_C_OutputObject (call, value_ptr,
			 tc->properties.object_class,
			 ilu_FALSE, err);
}

static ilu_cardinal
  _ILU_C_SizeOfValue (CORBA_TypeCode tc,
		      ilu_Call call,
		      void *value_ptr,
		      ilu_Error *err)
{
  if (tc->type_kind != ilu_object_tk)
    return ((*tc->sizeFn)(call, value_ptr, err));
  else
    return (_ILU_C_SizeOfObject (call, value_ptr,
				 tc->properties.object_class,
				 ilu_FALSE, err));
}

static void *
  _ILU_C_InputValue (CORBA_TypeCode tc,
		     ilu_Call call,
		     void *value_ptr,
		     ilu_Error *err)
{
  void *ptr;

  if (value_ptr != ILU_NIL)
    ptr = value_ptr;
  else if (tc->type_kind != ilu_object_tk) {
    ptr = ilu_MallocE (tc->properties.value_size, err);
    if (ILU_ERRNOK(*err)) return ILU_NIL;
  }
  if (tc->type_kind != ilu_object_tk)
    (*tc->inputFn)(call, ptr, err);
  else
    ptr = _ILU_C_InputObject (call, tc->properties.object_class, ilu_FALSE, err);
  return ptr;
}

#endif /* ADD_VARIANT_SUPPORT */

/* L1, L2 unconstrained */

ilu_CString ILU_C_ClassName (ILU_C_Object *o)
{
  ilu_Class c = ILU_C_ClassRecordOfInstance (o);
  if (o == ILU_NIL)
    return (ILU_NIL);
  else
    return (c->cl_name);
}

ilu_CString ILU_C_ClassID (ILU_C_Object *o)
{
  ilu_Class c = ILU_C_ClassRecordOfInstance (o);
  if (o == ILU_NIL)
    return (ILU_NIL);
  else
    return (c->cl_unique_id);
}

ilu_Class ILU_C_FindILUClassByTypeName (ilu_CString typename)
{
  return (ilu_FindClassFromName(typename));
}

ilu_Class ILU_C_FindILUClassByTypeID (ilu_CString typeID)
{
  return (ilu_FindClassFromID(typeID));
}

ilu_Class ILU_C_ClassRecordOfInstance(ILU_C_Object *o)
{
  if (o == ILU_NIL OR o->type == ILU_NIL)
    return (ILU_NIL);
  return ((o->type)[0]->c);
}

void *_ILU_C_FindMethod (ILU_C_Object *h, ilu_Class cls, int ndx)
{
  ILU_C_ObjectType c;
  _ILU_C_MethodBlock *d;
 
  c = h->type;
  while ((d = *c++) != ILU_NIL) {
    if (d->c == cls)
      {
	return(( void * ) d->methods[ ndx ]);
      }
  }
  return(( void * ) ILU_NIL );
}
 
/* Main Invariant holds */

ilu_CString ILU_C_SBHOfObject (ILU_C_Object *obj)
{
  ilu_CString sbh = ILU_NIL;

  ilu_Object kobj = _ILU_C_KernelObjOfObj (obj);
  if (kobj != ILU_NIL)
    {
      sbh = ilu_SBHOfObject (kobj);
      ilu_ExitServer (obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
    }
  return (sbh);
}

#ifdef IIOP_PROTOCOL

ilu_CString ILU_C_IOROfObject (ILU_C_Object *obj)
{
  ilu_Error       err;
  ilu_CString     sbh = ILU_NIL;

  ilu_Object      kobj = _ILU_C_KernelObjOfObj(obj);
  sbh = ilu_IOROfObject(kobj, &err);
  if (ILU_ERRNOK(err)) {
    DisposeErr(&err, "creating IOR (in ILU_C_IOROfObject)");
    sbh = ILU_NIL;
  }
  if (kobj != ILU_NIL)
    ilu_ExitServer(obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  return (sbh);
}

#endif

ILU_C_Object   *
ILU_C_SBHToObject(ilu_CString sbh,
		  ilu_Class static_type,
		  ILU_C_ENVIRONMENT * env)
{
  ILU_C_Object   *h = ILU_NIL;
  ilu_Object      obj;
  ILU_ERRS((bad_locks, broken_locks, inv_objref,
	    no_memory, internal)) lerr;
  env->_major = ILU_C_NO_EXCEPTION;
  obj = ilu_ObjectOfSBH(sbh, static_type, &lerr);
  if (ILU_ERRNOK(lerr))
    _ILU_C_ConvertError(env, &lerr, CORBA_COMPLETED_NO);
  if (obj == ILU_NIL)
    return (ILU_NIL);
  h = ilu_GetLanguageSpecificObject(obj, MyLangIdx());
  if (h == ILU_NIL)
    h = _ILU_C_CreateSurrogateFromRegistry(ilu_ClassOfObject(obj), obj);
  ilu_ExitServer(ilu_ServerOfObject(obj), static_type);
  return (h);
}

void
ILU_C_PingObject(ILU_C_Object * obj, ILU_C_ENVIRONMENT * env)
{
  ilu_Object      kobj = _ILU_C_KernelObjOfObj(obj);
  ilu_Class       mst;
  ilu_Error       err = ILU_INIT_NO_ERR;
  ilu_boolean     ok;
  ilu_Connection  newconn;
  if (kobj == ILU_NIL) {
    ILU_C_RAISE_SYSTEM(env, BAD_PARAM, 0, NO);
    return;
  }
  mst = ilu_ClassOfObject(kobj);
  err = ilu_DeltaHolds(kobj, 1);
  ilu_ExitServer(obj->server, mst);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE(GcRegFailed, v) {
      ILU_C_RAISE_SYSTEM(env, COMM_FAILURE, 0, NO);
      return;
    }
    ILU_ERR_ELSE {
      ILU_C_RAISE_SYSTEM(env, INTERNAL, 0, NO);
      return;
    }
  } ILU_ERR_ENDSWITCH;
  ok = ilu_PingObject(kobj, &newconn);
  if (newconn != ILU_NIL)
    (*Fork) (MonitorOutgoingConnection, newconn);
  ilu_EnterServer(obj->server, mst);
  err = ilu_DeltaHolds(kobj, -1);
  ilu_ExitServer(obj->server, mst);
  ILU_ERR_SWITCH(err) {
    ILU_SUCCESS_CASE;
    ILU_ERR_CASE(GcRegFailed, v) {
      ILU_C_RAISE_SYSTEM(env, COMM_FAILURE, 0, MAYBE);
      return;
    }
    ILU_ERR_ELSE {
      ILU_C_RAISE_SYSTEM(env, INTERNAL, 0, MAYBE);
      return;
    }
  } ILU_ERR_ENDSWITCH;
  if (ok)
    env->_major = ILU_C_NO_EXCEPTION;
  else
    ILU_C_RAISE_SYSTEM(env, COMM_FAILURE, 0, MAYBE);
  return;
}

void ILU_C_DestroyObject(ILU_C_OBJECT obj)
{
  ilu_Object      kobj = _ILU_C_KernelObjOfObj(obj);
  ilu_Class       mst;
  if (kobj == ILU_NIL)
    return;
  mst = ilu_ClassOfObject(kobj);
  ilu_RegisterLanguageSpecificObject(kobj, ILU_NIL, MyLangIdx());
  obj->instanceId = ILU_NIL;
  ilu_ExitServer(obj->server, mst);
  return;
}

/***********************************************************************
 *  This is support for ilu.CORBA-Object, the standard base object for
 *  all ILU objects defined in OMG IDL.
 ***********************************************************************/

ilu_Class _ilu_CORBA_Object__ILUType = NULL;

#ifdef ADD_VARIANT_SUPPORT
struct _ILU_C_IoFnsRegistration_s _ilu_CORBA_Object__IoFns = { ilu_object_tk, "IDL:omg.org/CORBA/Object:1.0", ILU_NIL, 0, 0, 0, 0 };
#endif

void ilu_CORBA_Object__Free (ilu_CORBA_Object* _val)
{
  /* What you put in the freeRoutine member of a CORBA_Environment for an exception parameterized by a ilu_CORBA_Object */
  /* frees allocated storage inside _val (if any), but does not free(_val) */
  CORBA_Environment env = {0};
  CORBA_Object_release(*_val, &env);
  *_val = 0;
}

struct _ilu_CORBA_Object__MethodBlock_s _ilu_CORBA_Object__SurrogateMethodBlock = {
  ILU_NIL /* will be set to _ilu_CORBA_Object__ILUType */};

static _ILU_C_MethodBlock *_ilu_CORBA_Object__SurrogateTypeVector[] = {
  (_ILU_C_MethodBlock *) &_ilu_CORBA_Object__SurrogateMethodBlock,
  ILU_NIL
};

ilu_CORBA_Object ilu_CORBA_Object__CreateFromSBH (char *sbh, ILU_C_ENVIRONMENT * env)
{
  return (ilu_CORBA_Object) (ILU_C_SBHToObject(sbh, _ilu_CORBA_Object__ILUType, env));
}

static struct _ilu_CORBA_Object__MethodBlock_s _ilu_CORBA_Object_ilu_CORBA_Object__TrueMethodBlock = {
  NULL /* will be set to _ilu_CORBA_Object__ILUType */};
static _ILU_C_MethodBlock *_ilu_CORBA_Object__TrueTypeVector[] = {
  (_ILU_C_MethodBlock *) &_ilu_CORBA_Object_ilu_CORBA_Object__TrueMethodBlock,
  ILU_NIL
};

ILU_C_OBJECT ilu_CORBA_Object__CreateTrue  (ilu_string instance_handle, ilu_Server server, void * data)
{
  return (_ILU_C_CreateTrueObject (_ilu_CORBA_Object__TrueTypeVector, instance_handle, server, data, ilu_FALSE));
}

ILU_C_OBJECT ilu_CORBA_Object__OTCreateTrue  (ilu_string instance_handle, ilu_Server server, void * data)
{
  if (server == ILU_NIL || instance_handle == ILU_NIL) return ILU_NIL;
  return (_ILU_C_CreateTrueObject (_ilu_CORBA_Object__TrueTypeVector, instance_handle, server, data, ilu_TRUE));
}

static void init_CORBA_Object()
{
  ilu_Class cl;
  ilu_Error lerr;

  cl = ilu_DefineObjectType("ilu.CORBA-Object",
			    ILU_NIL,
			    (char *) ilu_TypeID_ilu_CORBA_Object,
			    ILU_NIL,	/* singleton */
			    ilu_TRUE,	/* optional */
			    ilu_FALSE,	/* collectible */
			    ILU_NIL,	/* authentication */
			    0,		/* method count */
			    0,		/* superclass count */
			    ILU_NIL,	/* superclasses */
			    &lerr);
  if (ILU_ERRNOK(lerr))
    goto fail2;
  _ilu_CORBA_Object__ILUType = cl;

  /* surrogate support */
  _ilu_CORBA_Object__SurrogateMethodBlock.c = cl;
  _ILU_C_RegisterSurrogateCType (cl, _ilu_CORBA_Object__SurrogateTypeVector);

  /* true support */
  _ilu_CORBA_Object_ilu_CORBA_Object__TrueMethodBlock.c = cl;

#ifdef ADD_VARIANT_SUPPORT
  _ilu_CORBA_Object__IoFns.properties.object_class = cl;
  _ILU_C_RegisterIoFns (&_ilu_CORBA_Object__IoFns);
#endif /* ADD_VARIANT_SUPPORT */

fail2:
  ILU_MUST_BE_SUCCESS(lerr);
}

void _ILU_C_InitializeCRuntime(void)
{
	static ilu_boolean initted = ilu_FALSE;
	if (! initted) {
		
#if (defined(ADD_VARIANT_SUPPORT) && defined(_MSC_VER))
		/* workaround for vc++ treating references imported from another dll as non-static */
			InitFnsRegistrationTypeIdMacro(ilu_boolean,CORBA_boolean)
			InitFnsRegistrationTypeIdMacro(ilu_byte,CORBA_octet)
			InitFnsRegistrationTypeIdMacro(ilu_shortcharacter,CORBA_char)
			InitFnsRegistrationTypeIdMacro(ilu_character,ilu_character)
			InitFnsRegistrationTypeIdMacro(ilu_shortcardinal,CORBA_unsigned_short)
			InitFnsRegistrationTypeIdMacro(ilu_cardinal,CORBA_unsigned_long)
			InitFnsRegistrationTypeIdMacro(ilu_longcardinal,ilu_longcardinal)
			InitFnsRegistrationTypeIdMacro(ilu_shortinteger,CORBA_short)
			InitFnsRegistrationTypeIdMacro(ilu_integer,CORBA_long)
			InitFnsRegistrationTypeIdMacro(ilu_longinteger,ilu_longinteger)
			InitFnsRegistrationTypeIdMacro(ilu_shortreal,CORBA_float)
			InitFnsRegistrationTypeIdMacro(ilu_real,CORBA_double)
			InitFnsRegistrationTypeIdMacro(ilu_longreal,ilu_longreal)
			InitFnsRegistrationTypeIdMacro(ilu_CString,CORBA_string)
			InitFnsRegistrationTypeIdMacro(ilu_pickle,CORBA_any)
#endif
			
			initted = ilu_TRUE;
		init_CORBA_Object();
	};
}

/***********************************************************************
 *  End of support for ilu.CORBA-Object.
 ***********************************************************************/

typedef struct {
  ilu_cardinal    ncobjs, nviobjs;
}               DestructCounters;

/* Inside(kobj's server, ilu_rootClass) */
static int DestroyIt(ilu_Object kobj, ilu_refany rock)
{
  DestructCounters *dc = (DestructCounters *) rock;
  ilu_cardinal    mli = MyLangIdx();
  ILU_C_OBJECT    cobj = ilu_GetLanguageSpecificObject(kobj, mli);
  if (ilu_VeryInterested(kobj))
    dc->nviobjs++;
  if (cobj != ILU_NIL) {
    dc->ncobjs++;
    ilu_SetLSO(kobj, ilu_rootClass, ILU_NIL, mli);
    cobj->instanceId = ILU_NIL;
  }
  return 0;
}

/* Main Invariant holds */

void 
ILU_C_CloseServer(ilu_Server s,
		  ilu_cardinal * ncobjs, ilu_cardinal * nviobjs,
		  ilu_cardinal * nconns, ILU_C_ENVIRONMENT * env)
{
  int             res;
  DestructCounters dc = {0, 0};
  res = ilu_BankAndScanServer(s, DestroyIt, &dc, nconns);
  if (ncobjs != ILU_NIL)
    *ncobjs = dc.ncobjs;
  if (nviobjs != ILU_NIL)
    *nviobjs = dc.nviobjs;
  env->_major = ILU_C_NO_EXCEPTION;
  return;
}

void
ILU_C_DestroyObjectAndServer(ILU_C_OBJECT obj,
			     ILU_C_ENVIRONMENT * env)
{
  ilu_Server      s = obj->server;
  ILU_C_DestroyObject(obj);
  ILU_C_CloseServer(s, ILU_NIL, ILU_NIL, ILU_NIL, env);
  return;
}

ilu_boolean
ILU_C_ValidateOrDestroyObjSvr(ILU_C_OBJECT obj,
			      ILU_C_ENVIRONMENT * env)
{
  ILU_C_PingObject(obj, env);
  switch (env->_major) {
  case ILU_C_NO_EXCEPTION:
    return ilu_TRUE;
  case ILU_C_SYSTEM_EXCEPTION:
    if (ILU_C_EXCEPTION_ID(env) != ILU_C_STDEX(COMM_FAILURE))
      return ilu_FALSE;
    break;
  default:			/* can't happen */
    ILU_C_RAISE_SYSTEM(env, INTERNAL, 0, MAYBE);
    return ilu_FALSE;
  }
  ILU_C_EXCEPTION_FREE(env);
  ILU_C_DestroyObjectAndServer(obj, env);
  return ilu_FALSE;
}

typedef struct CallCons *CallList;
struct CallCons {
  /* L1, L2 unconstrained */

  ilu_Call        head;
  CallList        tail;
};

struct ILU_C_InterruptHandle_s {
  /* L1, L2 unconstrained */
  CallList        calls;
};

/* L1, L2 unconstrained */

static ilu_boolean AddCallToIH(ilu_Call call, ILU_C_InterruptHandle h)
{
  CallList        this = (CallList) ilu_malloc(sizeof(*this));
  if (this == ILU_NIL)
    return FALSE;
  this->head = call;
  this->tail = h->calls;
  h->calls = this;
  return TRUE;
}

static void RemCallFromIH(ilu_Call call, ILU_C_InterruptHandle h)
{
  CallList       *pcur;
  for (pcur = &(h->calls); *pcur != ILU_NIL; pcur = &((*pcur)->tail)) {
    if ((*pcur)->head == call) {
      CallList        doomed = *pcur;
      *pcur = doomed->tail;
      ilu_free(doomed);
      return;
    }
  }
}

/* Main Invariant holds */

ILU_C_InterruptHandle ILU_C_NewInterruptHandle(void)
{
  ILU_C_InterruptHandle h;
  h = (ILU_C_InterruptHandle) ilu_malloc(sizeof(*h));
  if (h != ILU_NIL)
    h->calls = ILU_NIL;
  return h;
}

void
ILU_C_SetObjectInterruptHandle(ILU_C_Object * obj,
			       ILU_C_InterruptHandle h)
{
  if (obj != ILU_NIL)
    obj->interruptH = h;
}

void ILU_C_InterruptHandleCalls(ILU_C_InterruptHandle h)
{
  CallList        cl;
  ilu_Error       lerr;
  for (cl = h->calls; cl != ILU_NIL; cl = cl->tail) {
    ilu_InterruptCall(cl->head, &lerr);
    ILU_HANDLED(lerr);
  }
}

ilu_CString ILU_C_PublishObject (ILU_C_Object *obj)
{
  ilu_Object kobj = _ILU_C_KernelObjOfObj (obj);
  if (kobj != ILU_NIL)
    return(ilu_PublishObject (kobj));
  else
    return (ILU_NIL);
}

ilu_boolean ILU_C_WithdrawObject (ILU_C_Object *obj, ilu_CString proof)
{
  ilu_Object kobj = _ILU_C_KernelObjOfObj (obj);
  if (kobj != ILU_NIL)
    return (ilu_WithdrawObject (kobj, proof));
  else
    return (ilu_FALSE);
}

OPTIONAL(GLOBAL(ILU_C_Object *))
ILU_C_LookupObject(RETAIN(char *) sid, RETAIN(char *) ih,
		   ilu_Class static_type)
{
  int             change;
  return ILU_C_ReLookupObject(sid, ih, static_type, &change);
}

OPTIONAL(GLOBAL(ILU_C_Object *))
ILU_C_ReLookupObject(RETAIN(char *) sid, RETAIN(char *) ih,
		     ilu_Class static_type, int *change)
{
  ILU_C_Object   *h = ILU_NIL;
  ilu_Object      obj;

  *change = ilu_ReLookupObject(sid, ih, static_type, &obj);
  if (obj == ILU_NIL)
    return (ILU_NIL);
  h = ilu_GetLanguageSpecificObject(obj, MyLangIdx());
  if (h == ILU_NIL)
    h = _ILU_C_CreateSurrogateFromRegistry(ilu_ClassOfObject(obj), obj);
  ilu_ExitServer(ilu_ServerOfObject(obj), static_type);
  return (h);
}

/*L1, L2 unconstrained*/

PASS(ilu_Passport)
ILU_C_CreatePassport (OPTIONAL(PASS(ilu_IdentityInfo)) info,
		      ILU_ERRS((no_memory)) *err)
{
  return ilu_CreatePassport(info, err);
}

PASS(ilu_IdentityInfo)
ILU_C_CopyIdentity (RETAIN(const struct _ilu_IdentityInfo_s *) info,
		    ILU_ERRS((no_memory)) *err)
{
  return ilu_CopyIdentity(info, err);
}

ilu_boolean
ILU_C_AddIdentity (RETAIN(ilu_Passport) pp,
		   PASS(const struct _ilu_IdentityInfo_s *) info,
		   ilu_Error *err)
{
  return ilu_AddIdentity (pp, info, err);
}

OPTIONAL(RETAIN(ilu_IdentityInfo))
ILU_C_FindIdentity (RETAIN(ilu_Passport) pp,
		    ilu_IdentityType type)
{
  return ilu_FindIdentity (pp, type);
}

ilu_boolean
ILU_C_DestroyPassport (PASS(ilu_Passport) pp,
		       ilu_Error *err)
{
  return ilu_DestroyPassport(pp, err);
}

#ifdef SECURE_TRANSPORT

ilu_IdentityInfo
ILU_C_AcquireGSSIdentity (gss_cred_id_t cred, ilu_Error *err)
{
  return ilu_AcquireGSSIdentity (cred, err);
}

ilu_boolean
ILU_C_DecodeGSSIdentity (ilu_IdentityInfo info,	/* input; retain; info to decode */
			 gss_name_t * name,	/* output; name in identity */
			 ilu_FineTime *till,	/* output; good-till; seconds past Unix epoch */
			 gss_OID mech,		/* input; actual mechanism desired; optional */
			 ilu_boolean *local,	/* if TRUE, local; otherwise remote */
			 ilu_cardinal *flags,	/* connection flags, as in gss_inquire_context */
			 ilu_Error *err)
{
  return ilu_DecodeGSSIdentity (info, name, till, mech, local, flags, err);
}

gss_cred_id_t
ILU_C_AcquireGSSCredForName (char *name,		/* name */
			     ilu_cardinal lifetime,	/* lifetime */
			     gss_OID mech,		/* secmech */
			     ilu_boolean accept_only,	/* accept_only */
			     ilu_Error *err		/* err */)
{
  return ilu_AcquireGSSCredForName (name, lifetime, mech, accept_only, err);
}

PASS(ilu_string)
ILU_C_GSSNameToString (gss_name_t name,
		       ilu_Error *err)
{
  return ilu_GSSNameToString (name, err);
}

#endif /* def SECURE_TRANSPORT */

/**before: Main Invariant, Call-Hi(call);
    after: Call-Invariant(call, err) && call->ca_ms == ilu_cmsHi*/
ILU_C_Object   *
_ILU_C_GetServerSingleton(ilu_Call call, ILU_ERRS((bad_param)) * err)
{
  ILU_C_Object   *h;
  ilu_Object      kobj;

  if ((kobj = ilu_GetCallSingleton(call, err)) != ILU_NIL) {
    h = (ILU_C_Object *) ilu_GetLanguageSpecificObject(kobj, MyLangIdx());
    ilu_ExitServer(ilu_ServerOfObject(kobj),
		   ilu_IntroTypeOfCall(call));
    return (h);
  }
  return (ILU_NIL);
}

/*before: not Inside (cobj->server, cobj->type->c) */
/*after:  return != ILU_NIL => Inside(cobj->server, cobj->type->c) */
ilu_Object _ILU_C_KernelObjOfObj (ILU_C_Object *cobj)
{
  ilu_EnterServer (cobj->server, MOST_SPECIFIC_ILU_CLASS(cobj));
  if (cobj->instanceId != ILU_NIL)
    return (cobj->instanceId);
  else
    {
      ilu_ExitServer (cobj->server, MOST_SPECIFIC_ILU_CLASS(cobj));
      return (ILU_NIL);
    }
}

struct _ILU_C_ObjectTable_struct {
  /* for slot access: L1, L2 unconstrained */

  ILU_C_ObjectTable_ApplyProc *ot_object_of_ih;
  ILU_C_ObjectTable_FreeProc *ot_free_ot;
  ilu_private     ot_user_data;
  ilu_ObjectTable ot_kernel_ot;
};

/**L1 >= {table's server};
   L1 >= {gcmu} if result is true and collectible;
   L2 unconstrained*/
static          ilu_Object
_ILU_C_OT_ObjectOfIh(ilu_ObjectTable self,
		     ilu_string ih)
{
  ILU_C_ObjectTable cot = (ILU_C_ObjectTable) (self->ot_rock);
  ILU_C_Object   *h;
  h = (cot->ot_object_of_ih) (ih, cot->ot_user_data);
  if (h == ILU_NIL)
    return (ILU_NIL);
  else
    return (h->instanceId);
}

/* L1 >= {table's server}; L2 unconstrained */
static void _ILU_C_OT_FreeSelf(ilu_ObjectTable self)
{
  ILU_C_ObjectTable cot = (ILU_C_ObjectTable) (self->ot_rock);
  (*cot->ot_free_ot) (cot->ot_user_data);
  ilu_free(cot);
}

/* L1, L2 unconstrained */

ILU_C_ObjectTable
ILU_C_CreateObjectTable(ILU_C_ObjectTable_ApplyProc * object_of_ih,
			ILU_C_ObjectTable_FreeProc * free_ot,
			ilu_private user_data)
{
  ILU_C_ObjectTable cot = ilu_must_malloc(sizeof(*cot));
  ilu_ObjectTable ot = ilu_must_malloc(sizeof(*ot));
  cot->ot_object_of_ih = object_of_ih;
  cot->ot_free_ot = free_ot;
  cot->ot_user_data = user_data;
  cot->ot_kernel_ot = ot;
  ot->ot_object_of_ih = _ILU_C_OT_ObjectOfIh;
  ot->ot_free_self = _ILU_C_OT_FreeSelf;
  ot->ot_rock = cot;
  return (cot);
}

void *_ILU_C_MallocFailure (ilu_cardinal nbytes)
{
  ilu_DebugPrintf("*** ilu_malloc(%u) failed.  Returning NIL.\n",
		  nbytes);
  return ILU_NIL;
}

#define BLKSIZE 4

ilu_CString ILU_C_Strdup (ilu_CString s)
{
  unsigned        len, fullen, i;
  ilu_CString     s2;
  if (s == ILU_NIL)
    return (ILU_NIL);
  len = strlen(s);
  fullen = len + BLKSIZE - (len % BLKSIZE);
  s2 = ilu_must_malloc(fullen);
  strcpy(s2, s);
  for (i = len + 1; i < fullen; i++)
    s2[i] = 0;
  return (s2);
}

ilu_boolean _ILU_C_IsSingleton (ilu_Class c)
{
  return (c->cl_singleton != ILU_NIL);
}

/* Main invariant holds */
ILU_C_Object *
  ILU_C_CreateSurrogateObject (ilu_Class type,
			       RETAIN(ilu_string) ih,
			       ilu_Server server,
			       ILU_C_ENVIRONMENT *env)
/* Create and return an instance of the specified type,
   with the specified ih, on the specified server */
{
  ILU_C_Object   *h = ILU_NIL;
  ilu_Object      obj;
  ILU_ERRS((bad_locks, broken_locks, inv_objref,
	    no_memory, internal)) lerr;
  env->_major = ILU_C_NO_EXCEPTION;
  obj = ilu_FindOrCreateSurrogate ((server == ILU_NIL) ? GetDefaultServer() : server,
				   ih, type, &lerr);
  if (ILU_ERRNOK(lerr))
    _ILU_C_ConvertError(env, &lerr, CORBA_COMPLETED_NO);
  if (obj == ILU_NIL)
    return (ILU_NIL);
  h = ilu_GetLanguageSpecificObject(obj, MyLangIdx());
  if (h == ILU_NIL)
    h = _ILU_C_CreateSurrogateFromRegistry(ilu_ClassOfObject(obj), obj);
  ilu_ExitServer(ilu_ServerOfObject(obj), type);
  return (h);
}

/* insideServer => Inside(server, class) */
/* otherwise, Main invariant holds */
ILU_C_Object   *
_ILU_C_CreateTrueObject(ILU_C_ObjectType class,
			OPTIONAL(ilu_CString) instance_handle,
			OPTIONAL(ilu_Server) server,
			void *client_data,
			ilu_boolean insideServer)
{
  ILU_C_Object   *lspo;
  ilu_Class       iluclass;
  ilu_CString     id;
  char            idbuf[10];
  static unsigned long idcounter = 0;

  if (server == ILU_NIL)
    server = GetDefaultServer();

  iluclass = (*class)->c;

  lspo = (ILU_C_Object *) ilu_must_malloc(sizeof(ILU_C_Object));
  lspo->type = class;
  lspo->server = server;
  lspo->data = client_data;

  if (instance_handle != ILU_NIL)
    id = instance_handle;
  else {
    sprintf(idbuf, "%lu", ++idcounter);
    id = idbuf;
  }

  if (!insideServer)
    ilu_EnterServer(lspo->server, iluclass);
  lspo->instanceId = ilu_FindOrCreateTrueObject(id, lspo->server,
						iluclass, lspo);
  if (!insideServer)
    ilu_ExitServer(lspo->server, iluclass);
  if (lspo->instanceId == ILU_NIL) {
    ilu_free(lspo);
    ilu_DebugPrintf (
	    "ILU_C:  Can't create kernel object for true object of type \"%s\".\n",
	    iluclass->cl_name);
    return (ILU_NIL);
  } else
    return (lspo);
}

/*Main invariant, Call-Hi(call)*/
ILU_C_Object   *
_ILU_C_InputObject(ilu_Call call, ilu_Class putative_class,
		   ilu_boolean discriminator,
		   ILU_ERRS((IoErrs)) * err)
{
  ilu_Object      obj = ILU_NIL;
  ILU_C_Object   *o = ILU_NIL;
  ilu_Class       c;

  ilu_InputObjectID(call, &obj, discriminator,
		    putative_class, err);
  if (ILU_ERRNOK(*err))
    return ILU_NIL;
  if (obj == ILU_NIL)
    return (ILU_NIL);
  /* now Inside(obj->server, putative_class) */
  o = (ILU_C_Object *) ilu_GetLanguageSpecificObject(obj, MyLangIdx());
  if (o == ILU_NIL) {
    if ((c = ilu_ClassOfObject(obj)) != ILU_NIL)
      o = _ILU_C_CreateSurrogateFromRegistry(c, obj);
  }
  ilu_ExitServer(ilu_ServerOfObject(obj), putative_class);
  return (o);
}
 
/*Main invariant, Call-Hi(call)*/
ilu_boolean 
_ILU_C_OutputObject(ilu_Call call, ILU_C_Object * obj,
		    ilu_Class putative_class,
		    ilu_boolean discriminator,
		    ILU_ERRS((IoErrs)) * err)
{
  ILU_CLER(*err);
  if (obj != ILU_NIL)
    ilu_EnterServer(obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  ilu_OutputObjectID(call,
		     (obj == ILU_NIL) ? ILU_NIL : obj->instanceId,
		     discriminator, putative_class, err);
  return ILU_ERROK(*err);
}

/* Main invariant holds */
ilu_cardinal 
_ILU_C_SizeOfObject(ilu_Call call, ILU_C_Object * obj,
		    ilu_Class putative_class,
		    ilu_boolean discriminator,
		    ILU_ERRS((IoErrs)) *err)
{
  ilu_cardinal    sz;

  if (obj != ILU_NIL)
    ilu_EnterServer(obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  sz = ilu_SizeOfObjectID(call,
		      (obj == ILU_NIL) ? ILU_NIL : obj->instanceId,
			  discriminator, putative_class, err);
  if (obj != ILU_NIL)
    ilu_ExitServer(obj->server, MOST_SPECIFIC_ILU_CLASS(obj));
  return (sz);
}
 
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

/*		Server code					       */

/***********************************************************************/
/***********************************************************************/
/***********************************************************************/
/***********************************************************************/

#include <stdio.h>      /* I/O defs (including popen and pclose) */

#if ((defined(WIN32) || defined(WIN16)) && defined(_WINIO))
#include <winiodef.h>
#endif /* ((defined(WIN32) || defined(WIN16)) && defined(_WINIO)) */

#include <sys/types.h>
#include <errno.h>
#if !(defined(WIN32) || defined(WIN16))
/* dll no sys/errno in VC++2 */
#include <sys/errno.h>
#endif /* not WIN32 or WIN16 */

#if (defined(WIN32) || defined(WIN16))
/* for WSAGetLastError */
#include <winsock.h>
#endif /* WIN32 or WIN16 */


/*Main Invariant; L2 disjoint {conn's callmu, iomu, waitmu}*/
static void     _ILU_C_ReadConnectionRequests(void *arg);

/* Main invariant holds; L2 disjoint {port's iomu, waitmu} */
static void     _ILU_C_ReadConnectionRequest(ilu_private);

/*Main Invariant holds; L2 disjoint {arg's waitmu, callmu, iomu}*/
static void     _ILU_C_ReadServiceRequest(ilu_private);

/*Main Invariant holds; L2 disjoint {conn's waitmu, callmu, iomu}*/
static ilu_boolean 
_ILU_C_FinalServiceRequest(ilu_Connection conn,
			   ilu_boolean single_threaded);

/*Main Invariant holds*/
static void PassNewConnections(void *rock);

/*L1, L2 unconstrained */

extern ilu_Exception	ex_ilu_ProtocolError;

static char *DefaultProtocol = "sunrpc";
static ilu_string DefaultTransport[] = { "sunrpcrm", "tcp_0_0", ILU_NIL };
static ilu_string inmemTinfo[] = { "inmem_", ILU_NIL };

/* Main invariant holds */
ilu_boolean ILU_C_SetFork(ILU_C_ForkProc fork)
{
  ILU_ERRS((internal)) lerr;
  if (threadedSet)
    return ilu_FALSE;
  threadedSet = ilu_TRUE;
  Fork = fork;
  threaded = ilu_TRUE;
  (void) MyLangIdx();
  (void) GetDefaultServer();
  (*fork) (PassNewConnections, ILU_NIL);
  (void) ilu_NewConnectionGetterForked(&lerr);
  ILU_MUST_BE_SUCCESS(lerr);
  return (ilu_TRUE);
}

/*
 * We don't want to mention ilu_OSForkNewThread() or
 * ilu_InitializeOSThreading explicitly, in case the application
 * isn't using them, to avoid linking that object module
 * unnecessarily.  So we indirect through KernelForkProc. If the
 * application code uses ILU_C_USE_OS_THREADS, the right things are
 * automatically invoked.
 */

static ILU_C_ErrableForkProc errableFork = 0;

static void ErrlessFork(ILU_C_WorkProc threadproc, void *threadarg)
{
  ILU_ERRS((no_memory, no_resources, internal)) err;
  (void) ((*errableFork) (threadproc, threadarg, &err));
  ILU_MUST_BE_SUCCESS(err);
}

/*Main invariant holds*/
ilu_boolean
ILU_C_EnableThreads(ILU_C_ThreadSetupProc s,
		    ILU_C_ErrableForkProc f)
{
  ILU_ERRS((bad_param, no_memory, no_resources,
	    internal)) err;

  if (s == 0 || f == 0) {
    ilu_DebugPrintf("ilu.c:ILU_C_EnableThreads given silly arguments!\n");
    return ilu_FALSE;
  }
  if (!(*s) (&err)) {
    DisposeErr(&err,
	       "attempting to set ILU kernel multi-threaded (in ILU_C_EnableThreads)");
    return ilu_FALSE;
  }
  errableFork = f;
  ILU_C_SetFork(ErrlessFork);
  return ilu_TRUE;
}

/*Main invariant holds */

static void MonitorOutgoingConnection(void *rock)
{
  ilu_Connection  conn = (ilu_Connection) rock;
  ILU_ERRS((IoErrs)) lerr;
  (void) ilu_OutgoingConnectionThreadProc(conn, &lerr);
  ILU_MUST_BE_SUCCESS(lerr);
}

/* Main invariant holds */
static void PassNewConnections(void *rock)
{
  ilu_Connection  nu;
  ILU_ERRS((internal)) lerr;
  while (1) {
    nu = ilu_OtherNewConnection(&lerr);
    if (nu != ILU_NIL)
      (*Fork) (MonitorOutgoingConnection, nu);
    ILU_MUST_BE_SUCCESS(lerr);
  }
}

/*
 * Create a true server then open an ILU port so we've got
 * something to export objects on. This procedure should be called
 * only once.
 */
/* Main invariant holds */
ilu_Server
ILU_C_InitializeServer(char *serverID, ILU_C_ObjectTable obj_tab,
		       char *protocol, ilu_TransportInfo transport,
		       ilu_Passport pp, ilu_boolean setdefaultport)
{
  ilu_Server      server;
  ilu_Port        port;
  ilu_Error	 err;
  char           *sid;
  char           *pinfo = ((protocol == ILU_NIL) ? DefaultProtocol
			   : protocol);
  ilu_TransportInfo tinfo = ((transport == ILU_NIL) ? DefaultTransport
			     : transport);
  threadedSet = ilu_TRUE;
  if (serverID == ILU_NIL)
    sid = ilu_InventID();
  else
    sid = ILU_C_Strdup(serverID);
  if (sid == ILU_NIL) {
    ilu_DebugPrintf (
       "ilu.c:  Unable to allocate for copy of server ID \"%s\"\n",
	    serverID ? serverID : "(unknown gensymed ID)");
    exit(1);
  }
  server = ilu_CreateTrueServer(sid,
	    (obj_tab == ILU_NIL) ? ILU_NIL : obj_tab->ot_kernel_ot,
				MyLangIdx());
  if (server == ILU_NIL) {
    ilu_DebugPrintf (
	"ilu.c:  Couldn't create true server from (\"%s\", %p).\n",
	    sid, obj_tab);
    exit(1);
  }
  if (protocol != ILU_NIL OR transport != ILU_NIL OR setdefaultport) {
    port = ilu_CreatePort(server, pinfo, tinfo, pp, &err);
    if (port == ILU_NIL) {
      char **ti;
      ilu_DebugPrintf(
		      "ilu.c:  Error <%s> (from line %d of %s) opening port with protocol = \"%s\" and transport =",
		      ILU_ERR_NAME(err), ilu_ErrorLine(&err),
		      ilu_ErrorFile(&err), pinfo);
      ILU_HANDLED(err);
      for (ti = tinfo;  *ti != ILU_NIL;  ti++)
	ilu_DebugPrintf (" \"%s\"", *ti);
      ilu_DebugPrintf (".\n");
      ilu_BankServer(server);
      return (ILU_NIL);
    }
    if (setdefaultport)
      ilu_SetServerDefaultPort(server, port);
    if (threaded) {
      Fork(_ILU_C_ReadConnectionRequests, port);
    } else {
      ILU_ERRS((no_memory, imp_limit,
		no_resources, bad_param,
		bad_locks, internal,
		broken_locks)) lerr;
      if (!ilu_SetConnectionRequestHandler(port,
				      _ILU_C_ReadConnectionRequest,
					   port, &lerr)) {
	ilu_DebugPrintf (
	  "ilu.c:  Can't register connection request handler for new server;\n");
	ilu_DebugPrintf ("sid=%s, k-err=%s\n", sid, ILU_ERR_NAME(lerr));
	ILU_HANDLED(lerr);
	ilu_BankServer(server);
	return (ILU_NIL);
      }
    }
  }
  port = ilu_CreatePort(server, DefaultProtocol, inmemTinfo, pp, &err);
  if (port == ILU_NIL) {
    ilu_DebugPrintf (
	    "ilu.c:  Error <%s> creating local port (protocol=\"%s\", tinfo=\"%s\")\n",
	    ILU_ERR_NAME(err), DefaultProtocol, inmemTinfo[0]);
    ILU_HANDLED(err);
  } else {
    ILU_ERRS((no_memory, imp_limit,
	      no_resources, bad_param,
	      bad_locks, internal,
	      broken_locks)) lerr;
    if (!ilu_SetConnectionRequestHandler(port,
				      _ILU_C_ReadConnectionRequest,
					 port, &lerr)) {
      ilu_DebugPrintf (
	      "ilu.c:  Can't register connection request handler for local port of new server;\n");
      ilu_DebugPrintf ("sid=%s, k-err=%s\n", sid, ILU_ERR_NAME(lerr));
      ILU_HANDLED(lerr);
    }
  }
  return (server);
}

/*
 * Create a true server then open an ILU port so we've got
 * something to export objects on.  This function should be called
 * at most once.
 */
/* Main invariant holds */
static ilu_Server GetDefaultServer(void)
{
  if (defaultServer != ILU_NIL)
    return defaultServer;
  defaultServer = ILU_C_InitializeServer(ILU_NIL, ILU_NIL, ILU_NIL,
					 ILU_NIL, ILU_NIL, TRUE);
  if (defaultServer == ILU_NIL) {
    ilu_DebugPrintf("ilu.c:  Error:  Couldn't create default server.\n");
    exit(1);
  }
  return defaultServer;
}

/* Main invariant holds; L2 disjoint {arg's waitmu, callmu, iomu} */
static void _ILU_C_RunConnection(void *arg)
{
  ilu_Connection  conn = (ilu_Connection) arg;
  ilu_boolean     closed = ilu_FALSE;
  while (NOT closed) {
    closed = NOT ilu_BlockingWaitForInputOnConnection(conn, ILU_NIL);
    if (closed) {
      ILU_ERRS((bad_param, broken_locks, bad_locks, internal)) lerr;
      if (ILU_C_Logging)
	ilu_DebugPrintf("ILU_C done serving connection %p.\n",
			conn);
      ilu_DoneServingConnection(conn, &lerr);
      DisposeErr(&lerr, "cleaning up after serving a connection");
    } else
      closed = _ILU_C_FinalServiceRequest(conn, ilu_FALSE);
  }
  return;
}

extern void _ilu_WaitForInputOnFD (int, ilu_boolean *, void *);

/*Main Invariant; L2 disjoint {conn's callmu, iomu, waitmu}*/
static void _ILU_C_ReadConnectionRequests(void *arg)
{
  ilu_Port        p = (ilu_Port) arg;
  ilu_boolean     closed = 0;
  ilu_Connection  conn;
  char            ef[ERRFMTSIZE];
  ILU_ERRS((IoErrs, bad_locks)) lerr;

  while (1) {
    if (ILU_C_Logging)
      ilu_DebugPrintf("Port %p waiting for connection request.\n", p);
    ilu_WaitForPortConnectionRequest(p, &lerr);
    if (ILU_ERRNOK(lerr)) {
      if (ILU_C_Logging)
	ilu_DebugPrintf(
	  "No longer waiting for connection requests on port %p,\n"
	   "because ilu_WaitForPortConnectionRequest raised %s.\n",
			p, FmtError(ef, &lerr));
      ILU_HANDLED(lerr);
      break;
    }
    conn = ilu_HandleNewConnection(p, &closed, &lerr);
    if (closed) {
      if (ILU_C_Logging)
	ilu_DebugPrintf(
	    "No longer waiting for connection requests on port %p,"
			"because it's closed.\n",
			p);
      break;
    }
    if (conn == ILU_NIL) {
      ILU_ERR_SWITCH(lerr) {
	ILU_SUCCESS_CASE {
	  if (ILU_C_Logging)
	    ilu_DebugPrintf("Spurrious input call on port %p.\n",
			    p);
	  goto nechst;
	}
	ILU_ERR_CASE(no_resources, x)
	  if (ILU_C_Logging)
	  ilu_DebugPrintf("Connection request on port %p"
			" ran over FD limit --- port abandoned.\n",
			  p);
	ILU_ERR_ELSE
	  if (ILU_C_Logging)
	  ilu_DebugPrintf("Got error %s"
			  " trying to accept connection on port %p;"
			  " abandoning port.\n",
			  FmtError(ef, &lerr), p);
      } ILU_ERR_ENDSWITCH;
      ILU_HANDLED(lerr);
      goto dun;
    } else {
      /* fork thread to run new connection */
      Fork(_ILU_C_RunConnection, conn);
    }
nechst:0;
  }
dun:
  if (!ilu_DoneWithPort(p, &lerr))
    DisposeErr(&lerr, "cleaning up after animating a port");
}

/*Main Invariant holds; L2 disjoint {conn's iomu, callmu, waitmu}*/
static ilu_boolean EnableRequests(ilu_Connection conn, ilu_Call call)
{
  ilu_Server      s = ilu_ServerOfConnection(conn);
  ilu_string      sid = ilu_IDOfServer(s);
  ilu_boolean     ans;
  ILU_ERRS((no_memory, no_resources)) lerr;
  call->ca_reqs_enabled = ilu_TRUE;
  ans = ilu_SetConnectionInputHandler(conn, _ILU_C_ReadServiceRequest,
				      conn, &lerr);
  if (!ans) {
    if (ILU_C_Logging)
      printf("ilu.c:  Can't register request handler for conn %p server %s!\nError = %s\n",
	     conn, sid, ILU_ERR_NAME(lerr));
    else
      ilu_DebugPrintf(
	    "ilu.c:  Can't register request handler for conn %p server %s!\nError = %s\n",
	    conn, sid, ILU_ERR_NAME(lerr));
    ILU_HANDLED(lerr);
  } else if (ILU_C_Logging)
    printf("Connection %p server %s awaiting requests.\n",
	   conn, sid);
  return ans;
}

/*Main Invariant holds; L2 disjoint {conn's iomu, callmu, waitmu}*/
static ilu_boolean DisableRequests(ilu_Connection conn, ilu_Call call)
{
  ilu_Server      s = ilu_ServerOfConnection(conn);
  ilu_string      sid = ilu_IDOfServer(s);
  ilu_boolean     ans;
  ILU_ERRS((no_memory, no_resources)) lerr;
  call->ca_reqs_enabled = ilu_FALSE;
  ans = ilu_SetConnectionInputHandler(conn, NULL, ILU_NIL, &lerr);
  if (!ans) {
    if (ILU_C_Logging)
      printf("ilu.c:  failure (%s) detected when unregistering request handler for conn %p server %s!\n",
	     ILU_ERR_NAME(lerr), conn, sid);
    else
      ilu_DebugPrintf("ilu.c:  failure (%s) detected when unregistering request handler for conn %p server %s!\n",
		      ILU_ERR_NAME(lerr), conn, sid);
    ILU_HANDLED(lerr);
  } else if (ILU_C_Logging)
    printf("Connection %p server %s not awaiting requests.\n",
	   conn, sid);
  return ans;
}

/* Main invariant holds; L2 disjoint {port's iomu, waitmu} */
static void
_ILU_C_ReadConnectionRequest(ilu_private arg)
{
  ilu_Port        p = (ilu_Port) arg;
  ilu_boolean     closed = 0;
  ilu_Connection  conn;
  ilu_Call_s      nocall;
  ILU_ERRS((IoErrs, bad_locks, no_resources)) lerr;
  char            ef[ERRFMTSIZE];
  if (ILU_C_Logging)
    printf("Port %p handling connection request.\n", p);
  conn = ilu_HandleNewConnection(p, &closed, &lerr);
  if (closed) {
    ilu_DebugPrintf("Port %p closed.\n", p);
    goto abandon;
  }
  if (conn == ILU_NIL) {
    ILU_ERR_SWITCH(lerr) {
      ILU_SUCCESS_CASE {
	if (ILU_C_Logging)
	  ilu_DebugPrintf("Spurrious input call on port %p.\n", p);
	goto ok;
      }
      ILU_ERR_CASE(no_resources, x)
	if (ILU_C_Logging)
	ilu_DebugPrintf("Connection request on port %p"
			" ran over FD limit --- port abandoned.\n",
			p);
      ILU_ERR_ELSE
	if (ILU_C_Logging)
	ilu_DebugPrintf("Got error %s"
			" trying to accept connection on port %p;"
			" abandoning port.\n",
			FmtError(ef, &lerr), p);
    } ILU_ERR_ENDSWITCH;
    ILU_HANDLED(lerr);
    goto abandon;
  }
  EnableRequests(conn, &nocall);
ok:
  return;
abandon:
  ilu_DoneWithPort(p, &lerr);
  DisposeErr(&lerr, "ceasing to animate a port");
}

/** Main Invariant holds;
    before: Call-Locking(call, IHi);
    after:  Call-Locking(call, VNo) */
typedef void (*cstubproc) (ilu_Call, ilu_Error *);

typedef struct {
  /* L1, L2 unconstrained */

  ilu_Call_s      call_s;
  ilu_Method      method;
}               C_Invokn;

/** Main Invariant holds;
    before: Call-Locking(call, IHi);
    after:  Call-Locking(call, VNo) */
static void DoInvocation(void *arg)
{
  C_Invokn       *ci = (C_Invokn *) arg;
  ILU_ERRS((bad_locks, IoErrs)) lerr;
  (*((cstubproc) ci->method->me_stubproc)) (&(ci->call_s), &lerr);
  ilu_free(ci);
  return;
}

/* Returns TRUE iff connection closed. */
/*Main Invariant holds; L2 disjoint {conn's waitmu, callmu, iomu}*/
/*If single_threaded: requests enabled for fd*/
static          ilu_boolean
_ILU_C_FinalServiceRequest(ilu_Connection conn,
			   ilu_boolean single_threaded)
{
  ilu_Call_s      call_s, *call;
  ilu_boolean     initted;
  ilu_Class       class;
  ilu_Method      method;
  ilu_cardinal    SN;
  ilu_RcvReqStat  stat;
  ILU_ERRS((bad_locks, IoErrs)) lerr = ILU_INIT_NO_ERR;
  C_Invokn       *ci = ILU_NIL;

  if (ILU_C_Logging)
    printf("Activity on connection %p:  ", conn);
  if (single_threaded) {
    call = &call_s;
    DisableRequests(conn, call);
  } else if (ilu_ThreadPerRequest(conn)) {
    ci = ilu_MallocE(sizeof(*ci), &lerr);
    if (ILU_ERRNOK(lerr)) {	/* out of memory */
      DisposeErr(&lerr,
	   "allocating memory to fork a thread to service a call");
      ilu_DebugPrintf("ilu.c: abandonding connection %p"
		      " for lack of memory.\n",
		      conn);
      ilu_DoneServingConnection(conn, &lerr);
      DisposeErr(&lerr,
	     "closing a connection in _ILU_C_FinalServiceRequest");
      return ilu_TRUE;
    }
    call = &(ci->call_s);
  } else
    call = &call_s;
  stat = ilu_ReceiveRequest(call, &initted, conn, &class, &method,
			    &SN, &lerr);
  if (ILU_C_Logging) {
    if (stat == ilu_RcvReqStat_request)
      printf("Received request SN %lu, method \"%s\" of class \"%s\"\n",
	     (unsigned long) SN, method->me_name, class->cl_name);
    else
      printf("ReceiveRequest => %s, *initted = %s, *err = %s\n",
	     ((stat == ilu_RcvReqStat_noop) ? "noop" :
	      ((stat == ilu_RcvReqStat_quit) ? "quit" :
	       "invalid result code")),
	     ((initted) ? "T" : "F"),
	     (ILU_ERROK(lerr) ? "SUCCESS" : ILU_ERR_NAME(lerr)));
  }
  if (stat == ilu_RcvReqStat_request) {
    /* Call-Locking(call, IHi) */
    if (ci) {
      ci->method = method;
      (*Fork) (DoInvocation, ci);
      /* L2 disjoint {conn's callmu, iomu} (handed off to new thd) */
    } else {
      (*((cstubproc) method->me_stubproc)) (call, &lerr);
      /* L2 disjoint {conn's waitmu, callmu, iomu} */
      /* single_threaded != threaded */
      /* !threaded <=> requests enabled */
    }
    return (ilu_FALSE);
  } else if (initted) {
    ilu_FinishCall(call, &lerr);
  }
  if (ci)
    ilu_free(ci);

  DisposeErr(&lerr,
	     "serving a call (in _ILU_C_FinalServiceRequest)");

  if (stat == ilu_RcvReqStat_quit) {
    if (ilu_ConnectionServingP(conn)) {
      if (ILU_C_Logging)
	ilu_DebugPrintf("ILU_C finished serving connection %p in nested call.\n",
			conn);
    } else {
      if (ILU_C_Logging)
	ilu_DebugPrintf("ILU_C done serving connection %p.\n",
			conn);
      ilu_DoneServingConnection(conn, &lerr);
      DisposeErr(&lerr,
	   "closing a connection (in _ILU_C_FinalServiceRequest)");
    }
    return ilu_TRUE;
  }
  if (single_threaded)
    EnableRequests(conn, call);
  return ilu_FALSE;
}

/* Returns TRUE iff connection closed. */
/*Main Invariant holds; L2 disjoint {arg's waitmu, callmu, iomu}*/
static void 
_ILU_C_ReadServiceRequest(ilu_private arg)
{
  ilu_Connection  conn = (ilu_Connection) arg;
  (void) _ILU_C_FinalServiceRequest(conn, ilu_TRUE);
  return;
}


/**before: Call-Locking(call, IHi);
    after: Call-Invariant(call, err),
	   success => Call-Locking(call, VLo)*/
/*If not threaded:
  before: requests not enabled;
  after:  requests enabled iff protocol concurrent.*/
ilu_boolean
_ILU_C_FinishParameters(ilu_Call call, ILU_C_Object * obj,
			ILU_ERRS((bad_locks, broken_locks)) * err)
{
  ilu_Connection  conn;
  ilu_boolean     ans;
  conn = ilu_ConnectionOfCall(call);
  ans = ilu_RequestRead(call, err);
  if ((!threaded) && ilu_ThreadPerRequest(conn))
    EnableRequests(conn, call);
  return ans;
}

/**before: Main Invariant && Call-VLo(call);
    after: Call-Invariant(call, err),
	   success => call->ca_ms == ilu_cmsHi. */

/**If not threaded:
  before: requests enabled iff protocol concurrent;
  after:  requests not enabled.*/
ilu_boolean
_ILU_C_BeginReply(ilu_Call call, ilu_boolean exceptions,
		  ilu_cardinal argSize,
		  ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Connection  conn = ilu_ConnectionOfCall(call);
  ilu_boolean     ans;
  if ((!threaded) && ilu_ThreadPerRequest(conn))
    DisableRequests(conn, call);
  ans = ilu_BeginReply(call, exceptions, argSize, err);
  return ans;
}

/*If not threaded:
  before: requests enabled iff protocol concurrent;
  after:  requests not enabled.*/
ilu_boolean
_ILU_C_BeginException(ilu_Call call, ilu_cardinal evalue,
		      ilu_cardinal argSize,
		      ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_Connection  conn = ilu_ConnectionOfCall(call);
  ilu_boolean     ans;
  if ((!threaded) && ilu_ThreadPerRequest(conn))
    DisableRequests(conn, call);
  ans = ilu_BeginException(call, evalue, argSize, err);
  return ans;
}

/*Main Invariant; Call-Locking(call, OHi)*/

/*If not threaded: requests not enabled.*/
ilu_boolean
_ILU_C_FinishReply(ilu_Call call, ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_boolean     ans;
  ans = ilu_FinishReply(call, err);
  return ans;
}

/*If not threaded: requests not enabled.*/
ilu_boolean
_ILU_C_FinishException(ilu_Call call,
		       ILU_ERRS((bad_locks, IoErrs)) * err)
{
  ilu_boolean     ans;
  ans = ilu_FinishException(call, err);
  return ans;
}

/**Before: Call-Invariant(call, err);
    After: Main Invariant*/
void 
_ILU_C_FinishServingCall(ilu_Call call, ilu_Error * err)
{
  ilu_Connection  conn = ilu_ConnectionOfCall(call);
  ilu_FinishCall(call, err);
  if (!threaded && !call->ca_reqs_enabled) {
    ILU_ERR_SWITCH(*err) {
      ILU_ERR_CASE(comm_failure, v) {
	if (!ilu_ConnectionServingP(conn)) {
	  ILU_ERRS((bad_param, broken_locks,
		    bad_locks, internal)) lerr;
	  (void) ilu_DoneServingConnection(conn, &lerr);
	  DisposeErr(&lerr,
		     "_ILU_C_FinishServingCall's call on ilu_DoneServingConnection");
	}
      }
      ILU_ERR_ELSE
	EnableRequests(conn, call);
    } ILU_ERR_ENDSWITCH;
  }
  DisposeErr(err, "serving a call");
  /* Should hand off to app somehow instead. */
  return;
}

/*Main Invariant; Call-Locking(call, Lo)*/
/**If not threaded:
  before: requests enabled iff protocol concurrent;
  after:  requests enabled iff protocol concurrent.*/
ilu_boolean
_ILU_C_NoReply(ilu_Call call,
	       ILU_ERRS((bad_param, bad_locks, broken_locks)) * err)
{
  ilu_boolean     ans;

  ans = ilu_NoReply(call, err);
  return ans;
}

/*Main Invariant holds; L2 otherwise unconstrained*/

void ILU_C_Stoppable_Run(int* stop) {

    threadedSet = ilu_TRUE;
    ilu_RunMainLoop( stop );
}

void ILU_C_Run( )
{
    int		stop = 0;

    threadedSet = ilu_TRUE;
    if (threaded)
      while (1) OS_SLEEP(30);
    else
      ilu_RunMainLoop( &stop );
}

/* L1, L2 unrestricted */

unsigned int _ILU_C_SafeStrlen (ilu_CString str)
{
  if (str == ILU_NIL)
    return 0;
  else
    return (strlen((char *) str));
}

/* L1, L2 unrestricted */
void _ILU_C_ExtendString (CORBA_char **str, CORBA_char item, CORBA_boolean atend)
{
  CORBA_char *s2;
  ilu_cardinal size;

  if (*str == ILU_NIL)
    size = 0;
  else
    size = strlen((char *) (*str));
  s2 = ilu_must_malloc(size + 2);
  if (!atend)
    {
      s2[0] = item;
      if (*str != ILU_NIL)
	strncpy ((char *) (s2 + 1), (char *) (*str), size);
      s2[size + 1] = 0;
    }
  else
    {
      if (*str != ILU_NIL)
	strncpy ((char *) s2, (char *) (*str), size);
      s2[size] = item;
      s2[size + 1] = 0;
    }
  *str = s2;
}

/* L1, L2 unrestricted */
void _ILU_C_ExtendWString (ilu_character **str, ilu_character item, CORBA_boolean atend)
{
  ilu_character *s2;
  ilu_cardinal size;

  size = _ILU_C_SafeWStrlen(*str);
  s2 = ilu_must_malloc((size + 2) * sizeof(ilu_character));
  if (!atend)
    {
      s2[0] = item;
      if (*str != ILU_NIL)
	{
	  ilu_character *p, *q;
	  ilu_cardinal i;

	  for (p = s2 + 1, q = *str, i = 0;  i < size;  i++)
	    *p++ = *q++;
	}
      s2[size + 1] = 0;
    }
  else
    {
      if (*str != ILU_NIL)
	{
	  ilu_character *p, *q;
	  ilu_cardinal i;

	  for (p = s2, q = *str, i = 0;  i < size;  i++)
	    *p++ = *q++;
	}
      s2[size] = item;
      s2[size + 1] = 0;
    }
  *str = s2;
}

/*L1, L2 unrestricted*/
void _ILU_C_PopString (CORBA_char **s, CORBA_char *ret)
{
  if (s != ILU_NIL && *s != ILU_NIL && **s != 0)
    {
      *ret = **s;
      *s = *s + 1;      
    }
}

/*L1, L2 unrestricted*/
void _ILU_C_PopWString (ilu_character **s, ilu_character *ret)
{
  if (s != ILU_NIL && *s != ILU_NIL && **s != 0)
    {
      *ret = **s;
      *s = *s + 1;
    }
}

/*L1, L2 unrestricted*/
unsigned int _ILU_C_SafeWStrlen (ilu_character *str)
{
  register ilu_character *p = str;

  if (str == ILU_NIL)
    return 0;

  while (*p++ != 0)
    ;
  return (p-str);
}

/*Main Invariant, Call-IHi(call)*/
static          ilu_boolean
_ILU_C_CatchException(ilu_Call call, ilu_Method method,
		      _ILU_C_ExceptionDescription *evec,
		      ILU_C_ENVIRONMENT * status,
		      ilu_cardinal exceptionIndex,
		      ILU_ERRS((IoErrs)) * err)
{

  if ((exceptionIndex > method->me_exceptionCount) ||
      exceptionIndex == 0) {
    status->_major = ILU_C_SYSTEM_EXCEPTION;
    status->returnCode = ILU_C_STDEX(UNKNOWN);
    status->ptr = (void *) ilu_MallocE(sizeof(ilu_cardinal), err);
    if (ILU_ERRNOK(*err))
      return ilu_FALSE;
    *((ilu_cardinal *) (status->ptr)) = 0;
    status->freeRoutine = 0;
  } else {
    unsigned int    valsize;
    ilu_Class       *valclass;
    ILU_C_InputFn   valfn;

    status->_major = ILU_C_USER_EXCEPTION;
    status->returnCode = ((ILU_C_ExceptionCode)
		   method->me_exceptionVector[exceptionIndex - 1]);
    valsize = evec[exceptionIndex - 1].size;
    if (valsize > 0) {
      status->ptr = (void *) ilu_MallocE(valsize, err);
      if (ILU_ERRNOK(*err))
	return ilu_FALSE;
      if (evec[exceptionIndex - 1].class != ILU_NIL) {
	valclass = evec[exceptionIndex - 1].class;
	(status->ptr) = (void *) _ILU_C_InputObject(call, *valclass,
						    ilu_FALSE, err);
      } else {
	valfn = evec[exceptionIndex - 1].inFn;
	(void) (*valfn) (call, status->ptr, err);
      }
      if (ILU_ERRNOK(*err))
	return ilu_FALSE;
      status->freeRoutine = evec[exceptionIndex - 1].freeFn;
    }
  }
  return ilu_TRUE;
}

ilu_cardinal ILU_C_SizeOfBoolean (ilu_Call call, CORBA_boolean b, ilu_Error *err)
{
  return (ilu_SizeOfBoolean(call, (ilu_boolean) b, err));
}

void ILU_C_OutputBoolean (ilu_Call call, CORBA_boolean b, ilu_Error *err)
{
  ilu_OutputBoolean (call, (ilu_boolean) b, err);
}

void ILU_C_InputBoolean (ilu_Call call, CORBA_boolean *b, ilu_Error *err)
{
  ilu_boolean b2;

  if (ilu_InputBoolean (call, &b2, err), ILU_ERROK(*err))
    *b = (CORBA_boolean) b2;
}

static ilu_cardinal _ILU_C_WStrlen(ilu_wstring s)
{
  register int i;
  if (s == ILU_NIL)
    return 0;
  for (i = 0;  s[i] != 0;  i++)
    ;
  return (ilu_cardinal) i;
}

/*
  _ILU_C_GenericCall

  void _ILU_C_GenericCall (ilu_Class class, ilu_Method method,
                           _ILU_C_ExceptionDescription *evec,
                           ILU_C_Object *discriminant,
                           ILU_C_ENVIRONMENT *status,
                           char *argdesc, ...)

  Operates in varargs mode, off "argdesc".  "argdesc" provides a description
  of the arguments to the method, as follows:

  "argdesc" contains 2 fields, separated by ':'.

  The first field provides a list
  of the types of all the arguments which are either In or InOut arguments,
  each as a one-character code.  The codes are capital letters if the argument
  is an InOut argument, lower-case if an In argument.  Specific types are
  encoded as follows:

  a & A:  short integer
  b & B:  integer
  c & C:  long integer
  d & D:  short cardinal
  e & E:  cardinal
  f & F:  long cardinal
  g & G:  short real
  h & H:  real
  i & I:  long real
  j & J:  short character
  k & K:  character
  l & L:  long character
  m & m:  byte
  n & N:  boolean
  o & O:  enumeration
  p & P:  object
  q & Q:  C string
  r & R:  Unicode string
  y    :  Out variable-length type
  s & S:  optional types
  z & Z:  other

  t-x & T-Y:  reserved for future use

  Values of primitive ISL types are presented on the varargs
  list as appropriate for a value of this type, given the CORBA ANSI C calling
  conventions.  Values of the meta-type "other" are presented as void *
  pointers to the real values, followed by a function pointer to one of the
  three routines for sizing, output, and input of the constructed type,
  depending on its position in the varargs part of the arglist.

  The sequence of value specified by the first field of "argdesc" actually
  appears twice in the varargs part of the _ILU_C_GenericCall arglist, as
  there is no assurance that "va_start" can be called twice in the same
  function.

  The second field of "argdesc" contains a list 
  of the types of all the arguments which are either the return type of the
  method, or Out or InOut arguments to the method, again as one-character
  codes, using the same coding scheme, but this time with lower-case
  characters representing Out arguments.  For the purposes of this routine,
  the return value of the method is passed and coded as another Out argument.
  The return value argument appears before the actual Out and InOut arguments
  in the arglist.  The code 'y' is added to represent Out parameters of
  variable-length types.

  Example:

  INTERFACE FOO;
  TYPE R = RECORD a : INTEGER, b : REAL END;
  EXCEPTION E1;
  EXCEPTION E2 : R;
  TYPE O4 = OBJECT
    METHODS
     Real-R-to-Real-R (a1 : REAL, INOUT a2 : R): REAL RAISES E1, E2 END
    END;

  The marshalling code for FOO.R and the surrogate C stub for FOO.O4.R-to-R
  would look like:

void _FOO_Output_R (ilu_Call _call, FOO_R * _val, ilu_Error *err)
{
  if (_val == ILU_NIL)
    return;

  ilu_OutputRecord (_call, err);
  if (ILU_ERRNOK(*err)) return;

  ilu_OutputInteger (_call, (_val->a), err);
  if (ILU_ERRNOK(*err)) return;

  ilu_OutputReal (_call, (_val->b), err);
  if (ILU_ERRNOK(*err)) return;

  ilu_EndRecord (_call);
}

ilu_cardinal _FOO_SizeOf_R (ilu_Call _call, FOO_R * _val, ilu_Error *err)
{
  ilu_cardinal size = 0;
  ilu_cardinal s;

  size = ilu_SizeOfRecord (_call);

  s = ilu_SizeOfInteger(_call, (_val->a));
  if (s == 0) return 0;
  size += s;

  s = ilu_SizeOfReal(_call, (_val->b));
  if (s == 0) return 0;
  size += s;

  ilu_EndRecord (_call);
  return size;
}

FOO_R * _FOO_Input_R (ilu_Call _call, FOO_R * _ref, ilu_Error *err)
{
  FOO_R * _val;

  ilu_InputRecord (_call, err);
  if (ILU_ERRNOK(*err)) return _val;

  if (_ref != ILU_NIL)
    _val = _ref;
  else
    _val = (FOO_R *) ilu_must_malloc (sizeof (FOO_R));

  ilu_InputInteger (_call, &_val->a, err);
  if (ILU_ERRNOK(*err)) return _val;

  ilu_InputReal (_call, &_val->b, err);
  if (ILU_ERRNOK(*err)) return _val;

  ilu_EndRecord (_call);

  return (_val);
}

void FOO_R__Free (FOO_R * _val)
{
}

ilu_real _FOO_O4_Real_R_to_Real_R
  (FOO_O4 _handle, ILU_C_ENVIRONMENT *_status, ilu_real a1, FOO_R *a2)
{
  ilu_real _retvalue;
  static _ILU_C_ExceptionDescription _evec[] = {
    { 0, ILU_NIL },
    { sizeof(FOO_R), &_FOO_R__Input }};

  _ILU_C_GenericCall (&_FOO_O4__ILUClassRecord,
                      &_FOO_O4__ILUClassRecord.cl_methods[0],
		      _evec, _handle, _status, "hZ:hZ",
                      a1,
                      (void *) a2,
                      (void)(*)() _Foo_SizeOf_R,
                      a1,
                      (void *) a2,
                      (void)(*)() _Foo_Output_R,
                      &_retvalue,
                      (void *) a2,
                      (void)(*)() _Foo_Input_R,
                      );
  return _retvalue;
}

*/

void
  _ILU_C_GenericCall (ilu_Class class, ilu_Method method,
		      _ILU_C_ExceptionDescription *evec,
		      ILU_C_Object *discriminant,
		      ILU_C_ENVIRONMENT *status,
		      char *argdesc, ...)
{
  ilu_Call_s      call_s;
  ilu_Call        call = &call_s;
  ilu_Error       err = ILU_INIT_NO_ERR;
  ilu_cardinal    size = 0, s;
  ilu_boolean     vaing = ilu_FALSE, initted = FALSE;
  ilu_boolean	needs_sizing;
  ilu_Object      kobj;
  va_list         ap;
  char           *p;
  ilu_cardinal    exceptionIndex;
  ILU_C_COMPLETIONSTATUS completed = CORBA_COMPLETED_NO;
  ilu_ProtocolException protocolErr;
  ILU_C_InterruptHandle intH = ILU_NIL;
  ilu_Connection  newconn;
  enum foo {a, b, c};

  ILU_CLER(err);

  (void) ilu_StartCall(call, discriminant->server, class, method,
		     MyLangIdx(), status->callerPassport,
		     &newconn, &err);
  if (newconn != ILU_NIL)
    (*Fork) (MonitorOutgoingConnection, newconn);
  if (ILU_ERRNOK(err))
    goto kerrOhneCall;
  intH = discriminant->interruptH;
  if (intH != ILU_NIL)
    if (!AddCallToIH(call, intH)) {
      ILU_C_RAISE_SYSTEM(status, NO_MEMORY, 0, NO);
      goto cerr;
    }

  needs_sizing = ilu_CallNeedsSizing(call);

  /* figure size of discriminant */
  if ((!needs_sizing) || _ILU_C_IsSingleton(class))
    size = 0;
  else {
    if ((kobj = _ILU_C_KernelObjOfObj(discriminant)) == ILU_NIL) {
      ILU_C_RAISE_SYSTEM(status, INV_OBJREF, 0, NO);
      goto cerr;
    }
    size = ilu_SizeOfObjectID(call, kobj, ilu_TRUE, class, &err);
    ilu_ExitServer(discriminant->server, ilu_ClassOfObject(kobj));
    if (ILU_ERRNOK(err))
      goto kerr;
  }
  
  va_start (ap, argdesc);
  vaing = ilu_TRUE;

  /* add in sizes of other args */

#define GENERIC_SIZEI(type,fn) { \
                                 type *v = va_arg(ap, type *); \
				 if (needs_sizing) {s = fn (call, *v, &err); \
				 if (ILU_ERRNOK(err)) goto kerr; \
				 else size += s; }\
			       } \
			       break
#define GENERIC_SIZES(type,fn) { \
                                 type v = (type) va_arg(ap, int); \
				 if (needs_sizing) { s = fn (call, v, &err); \
				 if (ILU_ERRNOK(err)) goto kerr; \
				 else size += s; }\
			       } \
			       break
#define GENERIC_SIZER(type,fn) { \
                                 type v = (type) va_arg(ap, double); \
				 if (needs_sizing) { s = fn (call, v, &err); \
				 if (ILU_ERRNOK(err)) goto kerr; \
				 else size += s; }\
			       } \
			       break
#define GENERIC_SIZE(type,fn) { \
                                 type v = va_arg(ap, type); \
				 if (needs_sizing) { s = fn (call, v, &err); \
				 if (ILU_ERRNOK(err)) goto kerr; \
				 else size += s; }\
			       } \
			       break


  for (p = argdesc;  p != ILU_NIL && *p != 0 && *p != ':';  p++)
    {
      switch (*p)
	{

	  /* Integers */

	case 'a':
	  GENERIC_SIZES(ilu_shortinteger, ilu_SizeOfShortInteger);
	case 'A':
	  GENERIC_SIZEI(ilu_shortinteger, ilu_SizeOfShortInteger);
	case 'b':
	  GENERIC_SIZE(ilu_integer, ilu_SizeOfInteger);
	case 'B':
	  GENERIC_SIZEI(ilu_integer, ilu_SizeOfInteger);
	case 'c':
	  GENERIC_SIZE(ilu_longinteger, ilu_SizeOfLongInteger);
	case 'C':
	  GENERIC_SIZEI(ilu_longinteger, ilu_SizeOfLongInteger);

	  /* Cardinals */

	case 'd':
	  GENERIC_SIZES(ilu_shortcardinal, ilu_SizeOfShortCardinal);
	case 'D':
	  GENERIC_SIZEI(ilu_shortcardinal, ilu_SizeOfShortCardinal);
	case 'e':
	  GENERIC_SIZE(ilu_cardinal, ilu_SizeOfCardinal);
	case 'E':
	  GENERIC_SIZEI(ilu_cardinal, ilu_SizeOfCardinal);
	case 'f':
	  GENERIC_SIZE(ilu_longcardinal, ilu_SizeOfLongCardinal);
	case 'F':
	  GENERIC_SIZEI(ilu_longcardinal, ilu_SizeOfLongCardinal);

	  /* Reals */

	case 'g':
	  GENERIC_SIZER(ilu_shortreal, ilu_SizeOfShortReal);
	case 'G':
	  GENERIC_SIZEI(ilu_shortreal, ilu_SizeOfShortReal);
	case 'h':
	  GENERIC_SIZE(ilu_real, ilu_SizeOfReal);
	case 'H':
	  GENERIC_SIZEI(ilu_real, ilu_SizeOfReal);
	case 'i':
	  GENERIC_SIZE(ilu_longreal, ilu_SizeOfLongReal);
	case 'I':
	  GENERIC_SIZEI(ilu_longreal, ilu_SizeOfLongReal);

	  /* Characters */

	case 'j':
	  GENERIC_SIZES(ilu_shortcharacter, ilu_SizeOfShortCharacter);
	case 'J':
	  GENERIC_SIZEI(ilu_shortcharacter, ilu_SizeOfShortCharacter);
	case 'k':
	  GENERIC_SIZES(ilu_character, ilu_SizeOfCharacter);
	case 'K':
	  GENERIC_SIZEI(ilu_character, ilu_SizeOfCharacter);
/*
	case 'l':
	  GENERIC_SIZE(ilu_longcharacter, ilu_SizeOfLongCharacter);
	case 'L':
	  GENERIC_SIZEI(ilu_longcharacter, ilu_SizeOfLongCharacter);
*/
	  /* Byte */

	case 'm':
	  GENERIC_SIZES(ilu_byte, ilu_SizeOfByte);
	case 'M':
	  GENERIC_SIZEI(ilu_byte, ilu_SizeOfByte);


	  /* Boolean */

	case 'n':
	  GENERIC_SIZES(CORBA_boolean, ILU_C_SizeOfBoolean);
	case 'N':
	  GENERIC_SIZEI(CORBA_boolean, ILU_C_SizeOfBoolean);

	  /* Enumeration */
	case 'o':
	  {
	    enum foo v;
	    v = va_arg(ap, enum foo);
	    if (needs_sizing) {
	      size += ilu_SizeOfEnum (call, (ilu_shortcardinal) v, &err);
	      if (ILU_ERRNOK(err)) goto kerr; };
	    break;
	  }
	case 'O':
	  {
	    enum foo *v;
	    v = va_arg(ap, enum foo *);
	    if (needs_sizing) {
	      size += ilu_SizeOfEnum (call, (ilu_shortcardinal) (*v), &err);
	      if (ILU_ERRNOK(err)) goto kerr; };
	    break;
	  }
	  /* Objects */

	case 'p':
	  {
	    ILU_C_Object *v;
	    ilu_Class pclass;

	    v = va_arg(ap, ILU_C_Object *);
	    pclass = va_arg(ap, ilu_Class);
	    if (needs_sizing) {
	      s = _ILU_C_SizeOfObject (call, v, pclass, ilu_FALSE, &err);
	      if (ILU_ERRNOK(err)) goto kerr;
	      size += s; };
	    break;
	  }
	case 'P':
	  {
	    ILU_C_Object **v;
	    ilu_Class pclass;

	    v = va_arg(ap, ILU_C_Object **);
	    pclass = va_arg(ap, ilu_Class);
	    if (needs_sizing) {
	      s = _ILU_C_SizeOfObject (call, *v, pclass, ilu_FALSE, &err);
	      if (ILU_ERRNOK(err)) goto kerr;
	      size += s;
	    };
	    break;
	  }

	  /* C String */

	case 'q':
	  {
	    ilu_string str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_string);
	    limit = va_arg(ap, ilu_cardinal);
	    if (needs_sizing) {
	      s = ilu_SizeOfString (call, str, strlen(str), limit, &err);
	      if (ILU_ERRNOK(err)) goto kerr;
	      size += s;
	    };
	    break;
	  }
	case 'Q':
	  {
	    ilu_string *str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_string *);
	    limit = va_arg(ap, ilu_cardinal);
	    if (needs_sizing) {
	      s = ilu_SizeOfString (call, *str, strlen(*str), limit, &err);
	      if (ILU_ERRNOK(err)) goto kerr;
	      size += s;
	    };
	    break;
	  }

	  /* Unicode String */

	case 'r':
	  {
	    ilu_wstring str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_wstring);
	    limit = va_arg(ap, ilu_cardinal);
	    if (needs_sizing) {
	      s = ilu_SizeOfWString (call, str, _ILU_C_WStrlen(str), limit, &err);
	      if (ILU_ERRNOK(err)) goto kerr;
	      size += s;
	    };
	    break;
	  }
	case 'R':
	  {
	    ilu_wstring *str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_wstring *);
	    limit = va_arg(ap, ilu_cardinal);
	    if (needs_sizing) {
	      s = ilu_SizeOfWString (call, *str, _ILU_C_WStrlen(*str), limit, &err);
	      if (ILU_ERRNOK(err)) goto kerr;
	      size += s;
	    }
	    break;
	  }

	  /* InOut Optional types */

	case 'S':
	  {
	    ilu_refany		*v;
	    ILU_C_SizeFn	fp;

	    v = va_arg(ap, ilu_refany *);
	    fp = va_arg(ap, ILU_C_SizeFn);
	    if (needs_sizing) {
	      s = (*fp) (call, *v, &err);
	      if (ILU_ERRNOK(err)) goto kerr; else size += s;
	    };
	    break;
	  }

	  /* other constructed types */

	case 's':
	case 'z':
	case 'Z':
	  {
	    void           *v;
	    ILU_C_SizeFn    fp;

	    v = va_arg(ap, void *);
	    fp = va_arg(ap, ILU_C_SizeFn);
	    if (needs_sizing) {
	      s = (*fp) (call, v, &err);
	      if (ILU_ERRNOK(err)) goto kerr; else size += s;
	    };
	    break;
	  }

	default:
	  ILU_C_RAISE_SYSTEM(status, INTERNAL, 0, NO);
	  goto cerr;
	}
    }

  if (!ilu_StartRequest(call, size, &err))
    goto kerr;

  /* marshall discriminant */

  if (!_ILU_C_IsSingleton(class))
    {
      if ((kobj = _ILU_C_KernelObjOfObj (discriminant)) == ILU_NIL)
	{
	  ILU_C_RAISE_SYSTEM(status, INV_OBJREF, 0, NO);
	  goto cerr;
	}
      ilu_OutputObjectID(call, kobj, ilu_TRUE, class, &err);
      if (ILU_ERRNOK(err))
	goto kerr;
    }
  
  /* marshall other arguments */

#define GENERIC_OUTPUT(type,fn) { \
                                  type v = va_arg(ap, type); \
				  fn (call, v, &err); \
				  if (ILU_ERRNOK(err)) goto kerr; \
				} \
				break

#define GENERIC_OUTPUTS(type,fn) { \
                                  type v = va_arg(ap, int); \
				  fn (call, v, &err); \
				  if (ILU_ERRNOK(err)) goto kerr; \
				} \
				break

#define GENERIC_OUTPUTR(type,fn) { \
                                  type v = va_arg(ap, double); \
				  fn (call, v, &err); \
				  if (ILU_ERRNOK(err)) goto kerr; \
				} \
				break

#define GENERIC_OUTPUTI(type,fn) { \
                                   type *v = va_arg(ap, type *); \
				   fn (call, *v, &err); \
				   if (ILU_ERRNOK(err)) goto kerr; \
				 } \
				 break

  for (p = argdesc;  p != ILU_NIL && *p != 0 && *p != ':';  p++)
    {
      switch (*p)
	{

	  /* Integers */

	case 'a':
	  GENERIC_OUTPUTS(ilu_shortinteger, ilu_OutputShortInteger);
	case 'A':
	  GENERIC_OUTPUTI(ilu_shortinteger, ilu_OutputShortInteger);
	case 'b':
	  GENERIC_OUTPUT(ilu_integer, ilu_OutputInteger);
	case 'B':
	  GENERIC_OUTPUTI(ilu_integer, ilu_OutputInteger);
	case 'c':
	  GENERIC_OUTPUT(ilu_longinteger, ilu_OutputLongInteger);
	case 'C':
	  GENERIC_OUTPUTI(ilu_longinteger, ilu_OutputLongInteger);

	  /* Cardinals */

	case 'd':
	  GENERIC_OUTPUTS(ilu_shortcardinal, ilu_OutputShortCardinal);
	case 'D':
	  GENERIC_OUTPUTI(ilu_shortcardinal, ilu_OutputShortCardinal);
	case 'e':
	  GENERIC_OUTPUT(ilu_cardinal, ilu_OutputCardinal);
	case 'E':
	  GENERIC_OUTPUTI(ilu_cardinal, ilu_OutputCardinal);
	case 'f':
	  GENERIC_OUTPUT(ilu_longcardinal, ilu_OutputLongCardinal);
	case 'F':
	  GENERIC_OUTPUTI(ilu_longcardinal, ilu_OutputLongCardinal);

	  /* Reals */

	case 'g':
	  GENERIC_OUTPUTR(ilu_shortreal, ilu_OutputShortReal);
	case 'G':
	  GENERIC_OUTPUTI(ilu_shortreal, ilu_OutputShortReal);
	case 'h':
	  GENERIC_OUTPUT(ilu_real, ilu_OutputReal);
	case 'H':
	  GENERIC_OUTPUTI(ilu_real, ilu_OutputReal);
	case 'i':
	  GENERIC_OUTPUT(ilu_longreal, ilu_OutputLongReal);
	case 'I':
	  GENERIC_OUTPUTI(ilu_longreal, ilu_OutputLongReal);

	  /* Characters */

	case 'j':
	  GENERIC_OUTPUTS(ilu_shortcharacter, ilu_OutputShortCharacter);
	case 'J':
	  GENERIC_OUTPUTI(ilu_shortcharacter, ilu_OutputShortCharacter);
	case 'k':
	  GENERIC_OUTPUTS(ilu_character, ilu_OutputCharacter);
	case 'K':
	  GENERIC_OUTPUTI(ilu_character, ilu_OutputCharacter);
/*
	case 'l':
	  GENERIC_OUTPUT(ilu_longcharacter, ilu_OutputLongCharacter);
	case 'L':
	  GENERIC_OUTPUTI(ilu_longcharacter, ilu_OutputLongCharacter);
*/

	  /* Byte */

	case 'm':
	  GENERIC_OUTPUTS(ilu_byte, ilu_OutputByte);
	case 'M':
	  GENERIC_OUTPUTI(ilu_byte, ilu_OutputByte);


	  /* Boolean */

	case 'n':
	  GENERIC_OUTPUTS(CORBA_boolean, ILU_C_OutputBoolean);
	case 'N':
	  GENERIC_OUTPUTI(CORBA_boolean, ILU_C_OutputBoolean);

	  /* Enumeration */

	case 'o':
	  {
	    enum foo v;
	    v = va_arg(ap, enum foo);
	    ilu_OutputEnum (call, (ilu_shortcardinal) v, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }
	case 'O':
	  {
	    enum foo *v;
	    v = va_arg(ap, enum foo *);
	    ilu_OutputEnum (call, (ilu_shortcardinal) (*v), &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }

	case 'p':
	  {
	    ILU_C_Object *v;
	    ilu_Class pclass;

	    v = va_arg(ap, ILU_C_Object *);
	    pclass = va_arg(ap, ilu_Class);
	    _ILU_C_OutputObject (call, v, pclass, ilu_FALSE, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }
	case 'P':
	  {
	    ILU_C_Object **v;
	    ilu_Class pclass;

	    v = va_arg(ap, ILU_C_Object **);
	    pclass = va_arg(ap, ilu_Class);
	    _ILU_C_OutputObject (call, *v, pclass, ilu_FALSE, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }

	  /* C String */

	case 'q':
	  {
	    ilu_string str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_string);
	    limit = va_arg(ap, ilu_cardinal);
	    ilu_OutputString (call, str, strlen(str), limit, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }
	case 'Q':
	  {
	    ilu_string *str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_string *);
	    limit = va_arg(ap, ilu_cardinal);
	    ilu_OutputString (call, *str, strlen(*str), limit, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }

	  /* Unicode String */

	case 'r':
	  {
	    ilu_wstring str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_wstring);
	    limit = va_arg(ap, ilu_cardinal);
	    ilu_OutputWString (call, str, _ILU_C_WStrlen(str), limit, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }
	case 'R':
	  {
	    ilu_wstring *str;
	    ilu_cardinal limit;

	    str = va_arg(ap, ilu_wstring *);
	    limit = va_arg(ap, ilu_cardinal);
	    ilu_OutputWString (call, *str, _ILU_C_WStrlen(*str), limit, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }

	  /* Optional types */

	case 'S':
	  {
	    ilu_refany		*v;
	    ILU_C_OutputFn	fp;

	    v = va_arg(ap, ilu_refany *);
	    fp = va_arg(ap, ILU_C_OutputFn);

	    (*fp) (call, *v, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }

	case 's':
	case 'z':
	case 'Z':
	  {
	    void           *v;
	    ILU_C_OutputFn  fp;

	    v = va_arg(ap, void *);
	    fp = va_arg(ap, ILU_C_OutputFn);

	    (*fp) (call, v, &err);
	    if (ILU_ERRNOK(err)) goto kerr;
	    break;
	  }

	default:
	  ILU_C_RAISE_SYSTEM(status, INTERNAL, 0, NO);
	  goto cerr;
	}
    }

  /* finish the request */

  completed = CORBA_COMPLETED_MAYBE;
  if (!ilu_FinishRequest (call, &err)) goto kerr;

  /* is there a reply? */

  if (! method->me_asynchronous)
    {
      /* yes */

      protocolErr = ilu_GetReply (call, &exceptionIndex, &err);
      if (ILU_ERRNOK(err)) goto kerr;

      if (protocolErr == ilu_ProtocolException_Success)
	{
	  completed = CORBA_COMPLETED_YES;
	  /* check to see if the user signalled Success (exceptionIndex == 0) */
	  if (exceptionIndex == 0)
	    {
	      status->returnCode = ILU_NIL;
	      status->_major = ILU_C_NO_EXCEPTION;

	      /* read in any return results */

#define GENERIC_INPUT(type,fn) { \
                                 type * v = va_arg(ap, type *); \
                                 fn (call, v, &err); \
				 if (ILU_ERRNOK(err)) goto kerr; \
			       } \
			       break

	      /* we left "p" pointing at the separating colon char */
	      for (++p;  p != ILU_NIL && *p != 0 && *p != ':';  p++)
		{
		  switch (*p)
		    {

		      /* Integers */

		    case 'a':
		    case 'A':
		      GENERIC_INPUT(ilu_shortinteger, ilu_InputShortInteger);
		    case 'b':
		    case 'B':
		      GENERIC_INPUT(ilu_integer, ilu_InputInteger);
		    case 'c':
		    case 'C':
		      GENERIC_INPUT(ilu_longinteger, ilu_InputLongInteger);

		      /* Cardinals */

		    case 'd':
		    case 'D':
		      GENERIC_INPUT(ilu_shortcardinal, ilu_InputShortCardinal);
		    case 'e':
		    case 'E':
		      GENERIC_INPUT(ilu_cardinal, ilu_InputCardinal);
		    case 'f':
		    case 'F':
		      GENERIC_INPUT(ilu_longcardinal, ilu_InputLongCardinal);

		      /* Reals */

		    case 'g':
		    case 'G':
		      GENERIC_INPUT(ilu_shortreal, ilu_InputShortReal);
		    case 'h':
		    case 'H':
		      GENERIC_INPUT(ilu_real, ilu_InputReal);
		    case 'i':
		    case 'I':
		      GENERIC_INPUT(ilu_longreal, ilu_InputLongReal);

		      /* Characters */

		    case 'j':
		    case 'J':
		      GENERIC_INPUT(ilu_shortcharacter, ilu_InputShortCharacter);
		    case 'k':
		    case 'K':
		      GENERIC_INPUT(ilu_character, ilu_InputCharacter);
/*
		    case 'l':
		    case 'L':
		      GENERIC_INPUT(ilu_longcharacter, ilu_InputLongCharacter);
*/

		      /* Byte */

		    case 'm':
		    case 'M':
		      GENERIC_INPUT(ilu_byte, ilu_InputByte);


		      /* Boolean */

		    case 'n':
		    case 'N':
		      GENERIC_INPUT(CORBA_boolean, ILU_C_InputBoolean);

		      /* Enum */

		    case 'o':
		    case 'O':
		      {
			ilu_shortcardinal val;
			enum foo *v;
			v = (enum foo *) va_arg(ap, enum foo *);
			ilu_InputEnum (call, &val, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			*v = (enum foo) val;
			break;
		      }

		      /* Object */

		    case 'p':
		    case 'P':
		      {
			ILU_C_Object **v;
			ilu_Class pclass;

			v = va_arg(ap, ILU_C_Object **);
			pclass = va_arg(ap, ilu_Class);
			*v = _ILU_C_InputObject (call, pclass, ilu_FALSE, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			break;
		      }

		      /* C String */

		    case 'q':
		      {
			ilu_string *str;
			ilu_cardinal limit;
			ilu_cardinal len;

			str = va_arg(ap, ilu_string *);
			limit = va_arg(ap, ilu_cardinal);
			*str = ILU_NIL;
			ilu_InputString (call, str, &len, limit, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			break;
		      }
		    case 'Q':
		      {
			ilu_string *str;
			ilu_cardinal limit;
			ilu_cardinal len;

			str = va_arg(ap, ilu_string *);
			limit = va_arg(ap, ilu_cardinal);
			ilu_free(*str);
			*str = ILU_NIL;
			ilu_InputString (call, str, &len, limit, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			size += s;
			break;
		      }

		      /* Unicode String */

		    case 'r':
		      {
			ilu_wstring *str;
			ilu_cardinal limit;
			ilu_cardinal len;

			str = va_arg(ap, ilu_wstring *);
			limit = va_arg(ap, ilu_cardinal);
			*str = ILU_NIL;
			ilu_InputWString (call, str, &len, limit, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			break;
		      }
		    case 'R':
		      {
			ilu_wstring *str;
			ilu_cardinal limit;
			ilu_cardinal len;

			str = va_arg(ap, ilu_wstring *);
			limit = va_arg(ap, ilu_cardinal);
			ilu_free(*str);
			*str = ILU_NIL;
			ilu_InputWString (call, str, &len, limit, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			size += s;
			break;
		      }

		      /* other constructed type */

		    case 'y':
		      {
			void ** v;
			ILU_C_InputFn fp;

			v = va_arg(ap, void **);
			fp = va_arg(ap, ILU_C_InputFn);

			*v = (*fp) (call, ILU_NIL, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			break;
		      }

		    case 's':
		    case 'S':
		    case 'z':
		    case 'Z':
		      {
			void * v;
			ILU_C_InputFn fp;

			v = va_arg(ap, void *);
			fp = va_arg(ap, ILU_C_InputFn);

			(void) (*fp) (call, v, &err);
			if (ILU_ERRNOK(err)) goto kerr;
			break;
		      }

		    case '*':
		      break;

		    default:
		      ILU_C_RAISE_SYSTEM(status, INTERNAL, 0, MAYBE);
		      goto cerr;
		    }
		}
	    }
	  else
	    /* indicates user signalled an expected exception */
	    {
	      if (!_ILU_C_CatchException(call, method, evec, status,
					 exceptionIndex, &err))
		goto kerr;
	    }
	  if (!ilu_ReplyRead (call, &err))
	    goto kerr;
	}
      else
	{
	  _ILU_C_SetProtocolError (status, protocolErr);
	  goto cerr;
	}
    }
  else
    {
      /* no return values or exceptions allowed, just return */
    }

  va_end(ap);
  vaing = ilu_FALSE;

kerr:		/* used for errors signalled via "err" */

  ilu_FinishCall (call, &err);
  if (ILU_ERROK(err))
    goto the_end;
  goto kerrOhneCall;

cerr:	/* used for errors signalled via "status" */

  ilu_FinishCall(call, &err);
  goto abend2;

 kerrOhneCall:

  _ILU_C_ConvertError (status, &err, completed);

 abend2:

  if (vaing)
    va_end(ap);
  
 the_end:
  
  if (initted && intH != ILU_NIL)
    RemCallFromIH(call, intH);
  
  return;
}

#ifdef ADD_VARIANT_SUPPORT

static ilu_Pickle			/* OUT, PASS */
  _ILU_C_AnyToPickle (CORBA_any *any,	/* IN, RETAIN */
		      ilu_Error *err)
{
  ilu_Pickle p;

  p.pi_len = 0;
  p.pi_bytes = ILU_NIL;
  if (any->_type == ILU_NIL) {	/* actually a pickle, just copy it */
    p.pi_bytes = ilu_MallocE(any->_pickle.pi_len, err);
    if (ILU_ERRNOK(*err)) return p;
    memcpy((void *) p.pi_bytes, (void *) any->_pickle.pi_bytes,
	   any->_pickle.pi_len);
    p.pi_len = any->_pickle.pi_len;
  } else {			/* need to marshal a pickle */
    ilu_Call_s call;
    ilu_cardinal size;
    ilu_StartPickle (&call, err);
    if (ILU_ERRNOK(*err)) return p;
    size = _ILU_C_SizeOfValue (any->_type, &call, any->_value, err);
    if (ILU_ERRNOK(*err)) return p;
    ilu_WritePickle (&call, size, any->_type->type_id, err);
    if (ILU_ERRNOK(*err)) return p;
    _ILU_C_OutputValue (any->_type, &call, any->_value, err);
    if (ILU_ERRNOK(*err)) return p;
    ilu_EndPickle (&call, &p, err);
  }
  return p;
}

static CORBA_any *
  _ILU_C_PickleToAny (ilu_Pickle p,	/* IN, PASS */
		      CORBA_any *any,	/* INOUT, RETAIN, MODIFY */
		      ilu_Error *err)
{
  ilu_string type_id;
  CORBA_TypeCode tc;

  type_id = ilu_PickleType (p, err);
  if (ILU_ERRNOK(*err)) return ILU_NIL;
  tc = _ILU_C_LookupIoFns(type_id);
  if (tc == ILU_NIL) {	/* this type unknown in this LSR */
    ILU_CLER(*err);
    any->_type = ILU_NIL;
    any->_value = ILU_NIL;
    any->_pickle.pi_len = p.pi_len;	/* return input pickle, but in `any' */
    any->_pickle.pi_bytes = p.pi_bytes;
    return any;
  } else {
    ilu_Call_s call;
    void *val = ILU_NIL;
    ilu_Pickle p2;

    ilu_StartPickle (&call, err);
    if (ILU_ERRNOK(*err)) return ILU_NIL;
    if (!ilu_ReadPickle(&call, p, err)) return ILU_NIL;
    val = _ILU_C_InputValue (tc, &call, ILU_NIL, err);
    if (ILU_ERRNOK(*err)) return ILU_NIL;
    ilu_EndPickle(&call, &p2, err);
    if (ILU_ERRNOK(*err)) {
      if (val != ILU_NIL && tc->freeFn != 0)
	(tc->freeFn) (val);
      return ILU_NIL;
    }
    ILU_CLER(*err);
    ilu_free(p2.pi_bytes);	/* free input pickle */
    any->_type = tc;
    any->_value = val;
    any->_pickle.pi_len = 0;
    any->_pickle.pi_bytes = ILU_NIL;
    return any;
  }
}

ilu_string
  ILU_C_Any_TypeID (CORBA_any *val, ILU_C_ENVIRONMENT *env)
{
  ilu_string type_id;
  CORBA_TypeCode tc;
  ilu_Error lerr;

  type_id = ilu_PickleType (val->_pickle, &lerr);
  _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
  if (!ILU_C_SUCCESSFUL(env)) return ILU_NIL;
  type_id = ilu_StrdupE(type_id, &lerr);
  _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
  return type_id;
}

CORBA_TypeCode
  ILU_C_Any_TypeCode (CORBA_any *val, ILU_C_ENVIRONMENT *env)
{
  ilu_string type_id;
  CORBA_TypeCode tc;
  ilu_Error lerr;

  type_id = ilu_PickleType (val->_pickle, &lerr);
  _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
  if (!ILU_C_SUCCESSFUL(env)) return ILU_NIL;
  tc = _ILU_C_LookupIoFns(type_id);
  return tc;
}

void *
  ILU_C_Any_Value (CORBA_any *val, ILU_C_ENVIRONMENT *env)
{
  ilu_string type_id;
  CORBA_TypeCode tc;
  ilu_Error lerr;
  void *v = ILU_NIL;

  type_id = ilu_PickleType (val->_pickle, &lerr);
  _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
  if (!ILU_C_SUCCESSFUL(env)) return ILU_NIL;
  tc = _ILU_C_LookupIoFns(type_id);
  if (tc != ILU_NIL) {	/* this type unknown in this LSR */
    ilu_Call_s call;
    ilu_Pickle p2;

    ilu_StartPickle (&call, &lerr);
    if (ILU_ERRNOK(lerr)) {
      _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
      return ILU_NIL;
    };
    if (!ilu_ReadPickle(&call, val->_pickle, &lerr)) {
      _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
      return ILU_NIL;
    };
    v = _ILU_C_InputValue (tc, &call, ILU_NIL, &lerr);
    if (ILU_ERRNOK(lerr)) {
      _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
      return ILU_NIL;
    };
    ilu_EndPickle(&call, &p2, &lerr);
    if (ILU_ERRNOK(lerr)) {
      _ILU_C_ConvertError(env, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
      if (v != ILU_NIL && tc->freeFn != 0)
	(tc->freeFn) (val);
      return ILU_NIL;
    }
  }
  return v;
}

CORBA_any *
  ILU_C_Any_Create (CORBA_TypeCode tc, void *c_val, ILU_C_ENVIRONMENT *e)
{
  CORBA_any *n;
  ilu_Error lerr;
  ilu_Call_s call;
  ilu_cardinal size;

  n = ilu_MallocE(sizeof(CORBA_any), &lerr);
  _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
  if (!ILU_C_SUCCESSFUL(e))
    return ILU_NIL;
  n->_type = ILU_NIL;
  n->_value = ILU_NIL;
  n->_pickle.pi_len = 0;
  n->_pickle.pi_bytes = ILU_NIL;
  ilu_StartPickle (&call, &lerr);
  if (ILU_ERRNOK(lerr)) {
    _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
    ilu_free(n);
    return ILU_NIL;
  };
  size = _ILU_C_SizeOfValue (tc, &call, c_val, &lerr);
  if (ILU_ERRNOK(lerr)) {
    _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
    ilu_free(n);
    return ILU_NIL;
  };
  ilu_WritePickle (&call, size, tc->type_id, &lerr);
  if (ILU_ERRNOK(lerr)) {
    _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
    ilu_free(n);
    return ILU_NIL;
  };
  _ILU_C_OutputValue (tc, &call, c_val, &lerr);
  if (ILU_ERRNOK(lerr)) {
    _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
    ilu_free(n);
    return ILU_NIL;
  };
  ilu_EndPickle (&call, &n->_pickle, &lerr);
  if (ILU_ERRNOK(lerr)) {
    _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
    ilu_free(n);
    return ILU_NIL;
  };
  return n;
}

CORBA_any* ILU_C_Any_Duplicate (CORBA_any *a, CORBA_Environment *e)
{
  CORBA_any *r;
  ilu_Error lerr;

  r = ilu_MallocE(sizeof(CORBA_any), &lerr);
  _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
  if (!ILU_C_SUCCESSFUL(e)) return ILU_NIL;
  r->_type = ILU_NIL;
  r->_value = ILU_NIL;
  r->_pickle.pi_len = a->_pickle.pi_len;
  r->_pickle.pi_bytes = ilu_MallocE(a->_pickle.pi_len, &lerr);
  if (ILU_ERRNOK(lerr)) {
    _ILU_C_ConvertError(e, &lerr, (ILU_ERROK(lerr) ? CORBA_COMPLETED_YES : CORBA_COMPLETED_NO));
    ilu_free(r);
    return ILU_NIL;
  };
  memcpy((void *) r->_pickle.pi_bytes, (void *) a->_pickle.pi_bytes,
	 a->_pickle.pi_len);
  return r;
}

void _CORBA_any__Output (ilu_Call _call, CORBA_any* _val, ilu_Error *_err)
{
  ilu_Pickle p;

  if (_val == NULL)
    return;

  if (_val->_type == ILU_NIL) {
    ilu_OutputPickle (_call, _val->_pickle, _err);
  } else {
    p = _ILU_C_AnyToPickle (_val, _err);
    if (ILU_ERRNOK(*_err)) return;
    ilu_OutputPickle(_call, p, _err);
    ilu_free(p.pi_bytes);
  }
  return;
}

ilu_cardinal _CORBA_any__SizeOf (ilu_Call _call, CORBA_any* _val, ilu_Error *_err)
{
  ilu_Pickle p;
  ilu_cardinal size = 0;

  if (_val == NULL)
    return size;

  if (_val->_type == ILU_NIL) {
    size = ilu_SizeOfPickle (_call, _val->_pickle, _err);
  } else {
    p = _ILU_C_AnyToPickle (_val, _err);
    if (ILU_ERROK(*_err)) {
      size = ilu_SizeOfPickle(_call, p, _err);
      ilu_free(p.pi_bytes);
    }
  }

  return (ILU_ERROK(*_err) ? size : 0);
}

/* To be CORBA-compliant, when a pickled value is read from the wire,
   it must be unpickled into a C value.  However, we provide the following
   attribute which may be set to prevent this, so that high efficency
   apps which don't need to unpickle don't have to.  */

ilu_boolean ILU_C_AutomaticUnpickling = ilu_FALSE;

CORBA_any* _CORBA_any__Input (ilu_Call _call, CORBA_any* _ref, ilu_Error *_err)
{
  ilu_Pickle p;
  CORBA_any* _val;

  if (_ref != NULL)
    _val = _ref;
  else
    _val = (CORBA_any*) ilu_MallocE(sizeof(CORBA_any), _err);
  _val->_type = ILU_NIL;
  _val->_value = ILU_NIL;
  _val->_pickle.pi_len = 0;
  _val->_pickle.pi_bytes = ILU_NIL;
  ilu_InputPickle (_call, &_val->_pickle, _err);
  if (ILU_ERRNOK(*_err)) {
    if (_val != _ref) ilu_free(_val);
    return ILU_NIL;
  } else {
    if (ILU_C_AutomaticUnpickling) {
      _ILU_C_PickleToAny (_val->_pickle, _val, _err);
      if (ILU_ERRNOK(*_err)) {
	if (_val != _ref) ilu_free(_val);
	return ILU_NIL;
      }
    }
    return _val;
  }
}

void CORBA_any__Free (CORBA_any* _val)
{
  /* What you put in the freeRoutine member of a CORBA_Environment for an exception parameterized by a CORBA_any */
  /* frees allocated storage inside _val (if any), but does not free(_val) */

  if (_val->_type != ILU_NIL) {
    if (_val->_value != ILU_NIL) {
      if (_val->_type->freeFn != 0)
	(_val->_type->freeFn) (_val->_value);
      ilu_free(_val->_value);
    }
  } else {
    ilu_free(_val->_pickle.pi_bytes);
  }
}

CORBA_any *CORBA_sequence_CORBA_any_allocbuf (CORBA_unsigned_long _count)
{
  CORBA_any *_p;
  CORBA_unsigned_long _size = sizeof(CORBA_any) * _count;

  if ((_p = (CORBA_any *) ilu_malloc(_size)) == ILU_NIL)
    { _ILU_C_MallocFailure(_size); return 0; }
  else
    { memset((void *) _p, 0, _size);  return _p; }
}

CORBA_any* CORBA_any_alloc ()
{
  return ((CORBA_any*) CORBA_sequence_CORBA_any_allocbuf(1));
}

#endif /* ADD_VARIANT_SUPPORT */

void
_ILU_C_SendException(ilu_Call call, _ILU_C_ExceptionDescription *evec,
		     ILU_C_ENVIRONMENT * status, ilu_Error * err)
{
  ilu_cardinal argSize = 0;
  ilu_cardinal eindex = 0, limit;
  ilu_Method method = ilu_MethodOfCall(call);

  ILU_CLER(*err);

  if (status->returnCode == ILU_NIL)
    return;

  limit = method->me_exceptionCount;
  for (eindex = 0; eindex < limit;  eindex ++)
    if (method->me_exceptionVector[eindex] == status->returnCode) 
      goto found;
  ilu_DebugPrintf("ilu.c: %s raises unexpected exception %p!\n",
		  ilu_NameOfMethod(method), status->returnCode);
  (void) ILU_ERR_CONS1(bad_param, err, minor, ilu_bpm_bogus_raise, 6);
  return;
  
found:

#define GENERIC_ESIZE(type,fn) argSize = fn (call, *((type *) status->ptr), err); break

  if (ilu_CallNeedsSizing(call)) {
    argSize = ilu_BeginSizingException(call, eindex+1, err);
    if (ILU_ERRNOK(*err))
      return;
  
    switch (evec[eindex].size)
      {
	/* Integers */

      case 'a':
	GENERIC_ESIZE(ilu_shortinteger, ilu_SizeOfShortInteger);
      case 'b':
	GENERIC_ESIZE(ilu_integer, ilu_SizeOfInteger);
      case 'c':
	GENERIC_ESIZE(ilu_longinteger, ilu_SizeOfLongInteger);

	/* Cardinals */

      case 'd':
	GENERIC_ESIZE(ilu_shortcardinal, ilu_SizeOfShortCardinal);
      case 'e':
	GENERIC_ESIZE(ilu_cardinal, ilu_SizeOfCardinal);
      case 'f':
	GENERIC_ESIZE(ilu_longcardinal, ilu_SizeOfLongCardinal);

	/* Reals */

      case 'g':
	GENERIC_ESIZE(ilu_shortreal, ilu_SizeOfShortReal);
      case 'h':
	GENERIC_ESIZE(ilu_real, ilu_SizeOfReal);
      case 'i':
	GENERIC_ESIZE(ilu_longreal, ilu_SizeOfLongReal);

	/* Characters */

      case 'j':
	GENERIC_ESIZE(ilu_shortcharacter, ilu_SizeOfShortCharacter);
      case 'k':
	GENERIC_ESIZE(ilu_character, ilu_SizeOfCharacter);
	/*
	  case 'l':
	  GENERIC_ESIZE(ilu_longcharacter, ilu_SizeOfLongCharacter);
	  */
	/* Byte */

      case 'm':
	GENERIC_ESIZE(ilu_byte, ilu_SizeOfByte);

	/* Boolean */

      case 'n':
	GENERIC_ESIZE(CORBA_boolean, ILU_C_SizeOfBoolean);

	/* Enumeration */
      case 'o':
	GENERIC_ESIZE(ilu_shortcardinal, ilu_SizeOfEnum);

	/* Objects */

      case 'p':
	argSize = _ILU_C_SizeOfObject (call, (ILU_C_Object *) status->ptr, *(evec[eindex].class), ilu_FALSE, err);
	break;

	/* C String */
      case 'q':
	argSize = ilu_SizeOfString (call, *((ilu_string *)(status->ptr)),
				    strlen(*((ilu_string *)(status->ptr))),
				    0xFFFF, err);

      case 'z':
	argSize = (*evec[eindex].sizeFn)(call, status->ptr, err);
	break;

      default:
	argSize = 0;
      }
    if (ILU_ERRNOK(*err)) return;
  }

  _ILU_C_BeginException (call, eindex+1, argSize, err);
  if (ILU_ERRNOK(*err)) return;

#define GENERIC_EOUT(type,fn) fn (call, *((type *) status->ptr), err); break

  switch (evec[eindex].size)
    {
      /* Integers */

    case 'a':
      GENERIC_EOUT(ilu_shortinteger, ilu_OutputShortInteger);
    case 'b':
      GENERIC_EOUT(ilu_integer, ilu_OutputInteger);
    case 'c':
      GENERIC_EOUT(ilu_longinteger, ilu_OutputLongInteger);

      /* Cardinals */

    case 'd':
      GENERIC_EOUT(ilu_shortcardinal, ilu_OutputShortCardinal);
    case 'e':
      GENERIC_EOUT(ilu_cardinal, ilu_OutputCardinal);
    case 'f':
      GENERIC_EOUT(ilu_longcardinal, ilu_OutputLongCardinal);

      /* Reals */

    case 'g':
      GENERIC_EOUT(ilu_shortreal, ilu_OutputShortReal);
    case 'h':
      GENERIC_EOUT(ilu_real, ilu_OutputReal);
    case 'i':
      GENERIC_EOUT(ilu_longreal, ilu_OutputLongReal);

      /* Characters */

    case 'j':
      GENERIC_EOUT(ilu_shortcharacter, ilu_OutputShortCharacter);
    case 'k':
      GENERIC_EOUT(ilu_character, ilu_OutputCharacter);
      /*
	case 'l':
	GENERIC_EOUT(ilu_longcharacter, ilu_OutputLongCharacter);
	*/
      /* Byte */

    case 'm':
      GENERIC_EOUT(ilu_byte, ilu_OutputByte);

      /* Boolean */

    case 'n':
      GENERIC_EOUT(CORBA_boolean, ILU_C_OutputBoolean);

      /* Enumeration */
    case 'o':
      GENERIC_EOUT(ilu_shortcardinal, ilu_OutputEnum);

      /* Objects */

    case 'p':
      _ILU_C_OutputObject (call, (ILU_C_Object *) status->ptr, *(evec[eindex].class), ilu_FALSE, err);
      break;

      /* C string */
    case 'q':
      ilu_OutputString (call, *((ilu_string *) (status->ptr)),
			strlen(*((ilu_string *) (status->ptr))),
			0xFFFF, err);

    case 'z':
      (*evec[eindex].outFn)(call, status->ptr, err);
      break;

    default:
      break;
    }
  if (ILU_ERRNOK(*err)) return;

  _ILU_C_FinishException (call, err);
  if (ILU_ERRNOK(*err)) return;

  return;
}

ilu_cardinal _ilu_CString__SizeOf (ilu_Call call, ilu_string str, ilu_Error *err)
{
  return (ilu_SizeOfString(call, str, _ILU_C_SafeStrlen(str), 0, err));
}

void _ilu_CString__Output (ilu_Call call, ilu_string str, ilu_Error *err)
{
  ilu_OutputString(call, str, _ILU_C_SafeStrlen(str), 0, err);
}

ilu_CString *_ilu_CString__Input (ilu_Call call, ilu_CString *strp, ilu_Error *err)
{
  ilu_cardinal len = 0;
  ilu_string *p;

  if (strp == ILU_NIL) {
    p = ilu_MallocE(sizeof(*p), err);
    if (ILU_ERRNOK(*err)) return ILU_NIL;
    *p = ILU_NIL;
  } else {
    p = strp;
  };
  ilu_InputString(call, p, &len, 0, err);
  if (ILU_ERROK(*err))
    return (p);
  else if (p != strp)
    ilu_free(p);
  return (ILU_NIL);
}

void ilu_CString__Free (ilu_CString *s)
{
  if (*s != ILU_NIL)
    ilu_free(*s);
}

/*============================================================*/
/*============================================================*/
/*                 Sequence operations                        */
/*============================================================*/
/*============================================================*/

#define Alloc( n, t )   ( t * ) ilu_must_malloc( n * sizeof( t ))
#define Realloc( p, n, t ) ( t * ) ilu_realloc( p, n * sizeof( t ))

void _ILU_C_AppendGeneric (ILU_C_Sequence h, char *p, int sz)
{
  char *ptr;

  /*
   ** place the item pointed to by p
   ** at the end of the sequence
   */

  if (h->_length >= h->_maximum)
    {
      h->_maximum = (h->_maximum + SEQUENCE_INCREMENT);
      if (h->_buffer != ILU_NIL)
	h->_buffer = ilu_realloc(h->_buffer, h->_maximum * sz);
      else
	h->_buffer = ilu_must_malloc(h->_maximum*sz);
    }
  ptr = h->_buffer + (h->_length * sz);
  memcpy (ptr, p, sz);
  h->_length += 1;
}

void _ILU_C_EveryElement(ILU_C_Sequence h, void (*proc)(void *,void *), int sz, void *data)
{
  int i;
  char *p;

  if (!h || h->_length <= 0)
    return;
  for(p = h->_buffer, i = 0; ((unsigned)i) < h->_length; i++, p += sz)
    (*proc)((void *) p, data);
}

void _ILU_C_PopGeneric (ILU_C_Sequence h, char *p, int sz)
{
    char	*ptr;

    /*
    ** return the top element
    ** in the sequence in p then
    ** remove it from the list.
    */

    if ( !h || h->_length <= 0 )
	return;
    memcpy( p, h->_buffer, sz );
    h->_length--;
    ptr = h->_buffer + sz;
    memmove (h->_buffer, ptr, h->_length * sz);
}

void _ILU_C_PushGeneric (ILU_C_Sequence h, char *p, int sz)
{
    int		l = h->_length;
    int		n;
    char	*ptr;

    /*
    ** place the item pointed to by p
    ** at the beginning of the sequence
    */

    h->_length++;
    n = h->_length * sz;
    if ( h->_length > h->_maximum ) {
        if ( h->_buffer )
    	    h->_buffer = Realloc( h->_buffer, n, char );
        else
    	    h->_buffer = Alloc( n, char );
	h->_maximum = h->_length;
    }
    ptr = h->_buffer + sz;

    memmove (ptr, h->_buffer, l * sz);
    memcpy (h->_buffer, p, sz);
}


/* L1, L2, Main unconstrained (rely on correct calls) */

static _ILU_C_MethodBlock GcCallbackMethods = { ILU_NIL, { 0 } };

static _ILU_C_MethodBlock *GcCallback__TrueTypeVector[] = {
	&GcCallbackMethods,
	ILU_NIL
};

static int CallbackInited = 0;
static ilu_Server cbs = ILU_NIL;
static ILU_C_OBJECT cbo = ILU_NIL;

/*Main invariant holds*/
void _ILU_C_EnsureGcClient(void)
{
  if (CallbackInited || ilu_IsGcClientSet())
    /* we alread initialized, or some other runtime must have already set things up for this process */
    return;

  GcCallbackMethods.c = ilu_GetGcCallbackClass();
  cbs = ILU_C_InitializeServer(ILU_NIL, ILU_NIL, ILU_NIL, ILU_NIL, ILU_NIL, TRUE);
  cbo = _ILU_C_CreateTrueObject(GcCallback__TrueTypeVector,
				"the-gc-callback", cbs, ILU_NIL, ilu_FALSE);
  if (cbo != ILU_NIL) {
    ilu_EnterServer(cbs, MOST_SPECIFIC_ILU_CLASS(cbo));
    ilu_SetGcClient(cbo->instanceId);
    ilu_ExitServer(cbs, MOST_SPECIFIC_ILU_CLASS(cbo));
  }
  CallbackInited = 1;
  return;
}
