/* IluDebug.java */
/* Chris Jacobi, January 3, 1997 3:54 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluDebug.java,v 1.10 1997/01/04 00:02:51 jacobi Exp $ */
/* See native code in IluJava_IluDebug.c */

package xerox.ilu;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.lang.System;
import xerox.basics.NowhereStream;

/**
 * This class helps in debugging IluJava. <p>
 * Applications better don't rely on this as it is likely
 * to be changed frequently.
 */
public class IluDebug {

    /**
     * Constant values for diverse debug flags
     */
    public static final int dont = 0;
    public static final int basic = 1;
    public static final int detailed = 2; 
    
    private static int masterDefault = dont; 
    
    private static final int gip(String key, int deflt) {
        return IluEnvironment.getIntProp(key, deflt);
    }
    
    //default value for all flags
    private static int def = gip("ilu.debug.default", masterDefault);
    
    //generic debugging
    private static int genericDebuggingFlag = gip("ilu.debug.db", def);
    public static final int traceGeneric() {return genericDebuggingFlag;}
    
    //spare to allow access without recompiling
    private static int spare1Flag = gip("ilu.debug.sp1", def);
    public static final int traceSpare1() {return spare1Flag;}
    
    //spare to allow access without recompiling
    private static int spare2Flag = gip("ilu.debug.sp2", def);
    public static final int traceSpare2() {return spare2Flag;}
    
    //initialization
    /*friendly*/ static int initsFlag = gip("ilu.debug.init", dont);
    public static final int traceInits() {return initsFlag;}

    //creation and destruction of objects
    private static int objectsFlag = gip("ilu.debug.oe", def);
    public static final int traceObjects() {return objectsFlag;}

    //gc of objects (remote and local)
    //UNUSUAL DEFAULT OF -1
    //Collectable object won be collected unless this flag is >= 0.
    //At this date it is too buggy for users, but collection will be
    //introduced without warning when ready.
    /*friendly*/ static int gcFlag = gip("ilu.debug.gc", -1);
    public static final int traceGC() {return gcFlag;}

    //transport of objects
    private static int objectTransportFlag = gip("ilu.debug.ot", def);
    public static final int traceObjectTransport() {return objectTransportFlag;}
    
    //Calls, marshalling
    /*friendly*/ static int callsFlag = gip("ilu.debug.cl", def);
    public static final int traceCalls() {return callsFlag;}

    //classes, methods etc.
    private static int definitionsFlag = gip("ilu.debug.df", def);
    public static final int traceDefs() {return definitionsFlag;}

    //ports, connections, transports...
    private static int pctFlag = gip("ilu.debug.pct", def);
    public static final int tracePCT() {return pctFlag;}

    //Monitor locks. Possibly expensive...
    private static int lockFlag = gip("ilu.debug.mltrace", 0); 
    public static final int traceLocks() {return lockFlag;}
    
    //Alarm tech.  Possibly expensive...
    private static int alarmFlag = gip("ilu.debug.alarm", 0); 
    public static final int traceAlarms() {return alarmFlag;}
    
    
    /**
     * Debugging aid to capture the java stack. 
     */
    public static String captureStackTrace() {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(bos);
        try {
            throw new java.lang.Exception();
        } catch (java.lang.Exception e) {
            e.printStackTrace(ps);
        } 
        return bos.toString();
    }
    
    
    /*
     * "Constructor" for a debug/log stream.
     */
    private static final PrintStream makeLog() {
        if (genericDebuggingFlag > dont) {
            return IluDebugStream.getIluDebugPrintStream();
        }
        return new PrintStream(new xerox.basics.NowhereStream());
    }
    
    /**
     * A stream whose output can be redirected and is used for debugging
     * purposes. While this may be redirected, it is still visible somewhere.
     */
    public static java.io.PrintStream log = makeLog();
    
     
    
    /**
     * Notification that ilu is about to raise some client
     * error soon.
     * 
     * - In Ilu debugging mode this will stop the thread and
     *   might even try to stop other threads to ease debugging.  
     * - In Ilu production mode this is either a no-op. It relies 
     *    on the caller to raise necessary errors.  
     *
     * clientPreError's are normal production features to Ilu. 
     */
    public static final void clientPreError(String s) {
        if (genericDebuggingFlag>dont) {
            block("clientPreError: " + s);
        }
    }
    
    
    /**
     * Notification that ilu is about to raise some internal
     * error soon.   
     * 
     * - In Ilu debugging mode this will stop the thread and
     *   might even try to stop other threads to ease debugging.  
     * - In Ilu production mode this might cause a log entry
     *   but will proceed and rely on the caller to raise 
     *   necessary errors.  
     */
    public static final void iluPreError(String s) {
        System.err.println("** xerox.ilu.IluDebug iluPreError: " + s);
        if (genericDebuggingFlag>dont) {
            block("iluError: " + s);
        }
    }
    
    
    /**
     * Ilu internal panic.
     * 
     * Does not return; No recovery possible.  
     * Severe ilu problem is encountered which makes
     * any further progress meaningless. Could as well
     * exit the program, but this way more debugging
     * is possible.
     * Trusted callers only.  
     */
    public static final void panic(String s) {
        System.err.println("**** xerox.ilu.IluDebug panic");
        System.err.println("     severe internal ilu problem");
        System.err.println(s);
        System.err.println(captureStackTrace());
        for (;;) {}
    }
    
    /**
     * block's the thread and blocks as much as conveniant
     * from Ilu..
     * 
     * Does not return; No recovery possible.  
     * This is used to stop ilu in situations which warrant
     * serious debugging in debugging mode; this shouldn't
     * be called in production mode.
     * Trusted callers only.  
     */
    static final void block(String s) {
        System.err.println("**** xerox.ilu.IluDebug block");
        System.err.println("     blocking thread in ilu debugging mode");
        System.err.println(s);
        System.err.println(captureStackTrace());
        for (;;) {}
    }
    
    /**
     * halt's the thread and blocks as much as conveniant
     * from Ilu.
     * Does not return; no recovery possible.  <p>
     *
     * This is used by test programs only; the difference
     * compared to "panic" is the users perception that  
     * the test program and not Ilu is initiating the action.
     * Trusted callers only.
     */
    public static final void halt(String s) {
        System.err.println("**** xerox.ilu.IluDebug halt");
        System.err.println("     halt invoked by test program");
        System.err.println(s);
        System.err.println(captureStackTrace());
        for (;;) {}
    }
    
    
    /*
     * Reports value of actual named debug flags into the native world...
     */
    private static native void reportFlags(
    	int genericDebuggingFlag, int spare1Flag, int spare2Flag,
    	int objectsFlag, int objectTransportFlag, int callsFlag, 
    	int definitionsFlag, int pctFlag, int gcFlag, 
    	int lockFlag, int alarmFlag
    	);
    	
    	
    /**
     * Reports the current state of most debug flags down
     * to the native world...
     */
    public static void reportSomeFlags() {
        reportFlags(
            genericDebuggingFlag, spare1Flag, spare2Flag, 
            objectsFlag, objectTransportFlag, callsFlag, 
            definitionsFlag, pctFlag, gcFlag, 
            lockFlag, alarmFlag
            );
    }
    
    
    /**
     * Can be called to assert initialization.
     * Internally called; ilu clients do not have to.
     */
    public static void init() {
        reportSomeFlags();
        iluDebugInstance.reportIluDebugInstance();
    }
    
    
    /**
     * We need an instance on the native side
     * to be able to invoke methods.
     */
    private native void reportIluDebugInstance();
    private static IluDebug iluDebugInstance = new IluDebug();
    static {
       IluInit.init();
       init();
    }
    
    /**
     * Dynamic version of captureStackTrace for impoverished
     * calls from the native side 
     */
    public String dynamicCaptureStackTrace() {
        return IluDebug.captureStackTrace(); 
    }
    
    /**
     * Dynamic version of getIntProp for impoverished
     * calls from the native side 
     */
    public int dynamicGetIntProp(String key, int deflt) {
        return IluEnvironment.getIntProp(key, deflt); 
    }


} // IluDebug
