/* IluJava_LockTech.c */
/* Chris Jacobi, February 2, 1997 9:00 am PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluJava_LockTech.c,v 1.5 1997/02/02 17:07:17 jacobi Exp $$ */
 
/*
 * This file defines the locking primitives used by iluxport.h
 * in terms of what is available in java  (which is sys_api.h)
 *
 * This could of course also be implemented using the real
 * thread package used.  That would certainly be more efficient,
 * probably be easier to implement but NOT portable to different
 * java implementations.
 */


#include <StubPreamble.h>
#include <interpreter.h>
#include <javaString.h>
#include <iluxport.h>
#include <iluerror.h>
#include "IluJava.h"

#include <sys_api.h>
/*
 *  WARNING
 *  *******
 * 
 *  sys_api.h as we got it here at xerox, in
 *  /project/java/jdk-1.0.2/include/sys_api.h
 *  has a bug:  It imports sysmacros_md.h
 *  which is not available.  We have made a
 *  a private copy of sysmacros_md to fix this
 *  java bug.  Here at Xerox we do not use 
 *  Sun's version which might be available in a 
 *  source release.
 *
 *  See sysmacros_md.h
 */



/* 
 * Never wait with the META lock hold.
 * Enter META first.
 * Never hold two locks unless one is the META lock.
 *
 * The reason we use separate monitors instead of doing all the waiting
 * with the meta lock is so that a single notify doesn't wake all threads.
 */

typedef struct {
    ilu_string d1;
    ilu_string d2;
    sys_thread_t * holder;
    sys_mon_t * mQueue; 
    char * stackWhenAquired;  /*optional; for debugging only*/
} ltj_Mutex_s;


/* queue element */
typedef struct R {
    struct R * prev;
    struct R * next;
    ilu_boolean disabled;
    sys_mon_t * elQueue;
} ltj_QueueEl_s;


typedef struct {
    ilu_string d1;
    ilu_string d2;
    ltj_QueueEl_s * first;
    ltj_QueueEl_s * last;
} ltj_Condition_s;


static void PrintWereAquired(ltj_Mutex_s * m) 
/* prints where a mutex was aquired, if that information is available */
{
    if (m->stackWhenAquired) {
        ilu_DebugPrintf("$ JAVA STACK WHEN LOCK WAS AQUIRED [[%s]]\n\n",
        	m->stackWhenAquired);
    }
}


#define MUTEX_FREE 0

#define NO_TIMEOUT -1
    /* Stupid java runtime system uses a different encoding then the 
     * java language which uses 0...
     * Don't actually use this to avaoid trouble when the java
     * implementation changes.
     */
#define BIG_TIMEOUT 60000
    /* Use this instead of NO_TIMEOUT to make sure threads
     * can die gracefully when necessary.  Use a loop...
     * (expressed in milliseconds)
     */
#define VERY_BIG_TIMEOUT 300000
     /* Like BIG_TIMEOUT only bigger because we let conditions
      * awake spuriously without loop (Relying on caller to loop)...
      * (expressed in milliseconds)
      */
 
#define META_ENTER sysMonitorEnter(_ilujava_metaLock)
#define META_EXIT sysMonitorExit(_ilujava_metaLock)

static sys_mon_t * _ilujava_metaLock;
static ilu_LockTech _ilujava_lockTech;
static ltj_QueueEl_s * _ilujava_freeList = 0; /* of queue elements */


/* create and initialize a sys_mon_t *  */
INTERNALONLY sys_mon_t * 
_ilujava_newJavaMonitor()
{
	sys_mon_t * m = (sys_mon_t *) java_sysMalloc(sysMonitorSizeof());
	memset((char *)m, 0, sysMonitorSizeof());
#ifdef JDK_11
	sysMonitorInit(m);
#else
	sysMonitorInit(m, FALSE);
#endif       
	return m;
}


/* must hold meta lock */
INTERNALONLY void
_ilujava_free_QueueEl(ltj_QueueEl_s * el)
{
	el->prev = 0;
	el->next = _ilujava_freeList;
	_ilujava_freeList = el;
}


/* must hold meta lock */
INTERNALONLY ltj_QueueEl_s *
_ilujava_new_QueueEl()
{
	ltj_QueueEl_s * el = _ilujava_freeList;
	if (el) {
		_ilujava_freeList = el->next;
	} else {
		el = (ltj_QueueEl_s *) java_sysMalloc(sizeof(ltj_QueueEl_s));
		el->elQueue = _ilujava_newJavaMonitor();
	}
	el->next = 0;
	el->prev = 0;
	return el;
}


/* must hold meta lock */
INTERNALONLY void
_ilujava_incl_QueueEl(ltj_Condition_s * cond, ltj_QueueEl_s * el)
{
	el->next = 0;
	el->prev = cond->last;
	if (cond->last==0) {
		cond->last = el;
		cond->first = el;
		return;
	}
	cond->last->next = el; 
}


/* must hold meta lock; idempotent */
INTERNALONLY void
_ilujava_excl_QueueEl(ltj_Condition_s * cond, ltj_QueueEl_s * el)
{
	if (el->prev) el->prev->next = el->next;
	if (el->next) el->next->prev = el->prev;
	if (cond->last==el) cond->last = el->prev;
	if (cond->first==el) cond->first = el->next;
	el->prev = 0;
	el->next = 0;
}


/* See iluxport.h lt_mcreate for specification */
INTERNALONLY ilu_Mutex
_ilujava_mcreate (ilu_string d1, ilu_string d2)
{
	ltj_Mutex_s * mutex; 
	mutex = (ltj_Mutex_s *) java_sysMalloc(sizeof(ltj_Mutex_s));
	if (mutex == 0) {
	    _iluJava_PANIC("$ _ilujava_mcreate failed; FATAL!!!");
	    return ILU_NIL;
	}
	mutex->d1 = d1;
	mutex->d2 = d2;
	mutex->holder = MUTEX_FREE;
	mutex->stackWhenAquired = 0;
	mutex->mQueue = _ilujava_newJavaMonitor();
	return (ilu_Mutex) mutex;
}

INTERNALONLY void
_ilujava_mdestroy (ilu_Mutex m, ILU_ERRS((bad_param, bad_locks)) * err)
{
	ltj_Mutex_s * mutex = (ltj_Mutex_s *) m;
	ILU_CLER(*err);
	if (mutex->holder != MUTEX_FREE) {
	    PrintWereAquired(mutex);
	    _iluJava_BUG("$ LockTech acquire: bad locks\n");
	    ILU_ERR_CONS0(bad_locks, err, 0); 
	    return; /* and leak the memory ... */
	}
	java_free(mutex->mQueue);
	java_free(mutex);
}


/* See iluxport.h lt_ccreate for specification */
INTERNALONLY ilu_Condition
_ilujava_ccreate (ilu_string d1, ilu_string d2)
{
	ltj_Condition_s * cond; 
	cond = (ltj_Condition_s *)java_sysMalloc(sizeof(ltj_Condition_s));
	if (cond == 0) {
	    _iluJava_PANIC("$ _ilujava_ccreate failed; FATAL!!!");
	    return ILU_NIL;
	}
	cond->d1 = d1;
	cond->d2 = d2;
	cond->first = 0;
	cond->last = 0;
	return (ilu_Condition) cond;
}


/* See iluxport.h lt_muncons for specification */
INTERNALONLY void
_ilujava_muncons (ilu_Mutex m, ilu_string * d1,
		     ilu_string * d2, ILU_ERRS((bad_param)) * err)
{
	ltj_Mutex_s * mutex = (ltj_Mutex_s *) m; 
	ILU_CLER(*err);
	*d1 = mutex->d1;
	*d2 = mutex->d2;
	}


/* See iluxport.h lt_cuncons for specification */
INTERNALONLY void
_ilujava_cuncons (ilu_Condition c, ilu_string * d1,
		     ilu_string * d2, ILU_ERRS((bad_param)) * err)
{
	ltj_Condition_s * cond = (ltj_Condition_s *) c;
	ILU_CLER(*err);
	*d1 = cond->d1;
	*d2 = cond->d2;
}


/* See iluxport.h lt_cdestroy for specification */
/* Bill says we are guaranteed that no thread is using this anymore */
INTERNALONLY void
_ilujava_cdestroy (ilu_Condition c, ILU_ERRS((bad_param)) * err)
{
	ltj_Condition_s * cond = (ltj_Condition_s *) c;
	ILU_CLER(*err);
	if (cond->first) {
	    _ilu_Assert(0, "bad cdestroy"); 
	}
	java_free(cond);
}


/* See iluxport.h lt_mdestroy for specification */

/* m1 must not be hold */
INTERNALONLY void
_ilujava_acquire1 (ltj_Mutex_s * m1, sys_thread_t * thread, 
	ltj_Condition_s * cond, ltj_QueueEl_s * el)
{
	_ilu_Assert((m1->holder != thread), "self entering lock 1");
	while (1) {
		META_ENTER;
		/* Do this here so lock is entered only once;
		 * Looping is cheaper then entering the lock
		 */
		if (el) {
			_ilujava_excl_QueueEl(cond, el);
			_ilujava_free_QueueEl(el);
			el = 0;
		}
		if (m1->holder == MUTEX_FREE) {
			m1->holder = thread;
			if (_ilujava_lockFlag>1) {
			    m1->stackWhenAquired = _iluJava_captureJavaStack();
			}
			META_EXIT;
			return;
		}
		META_EXIT;
		sysMonitorEnter(m1->mQueue); 
		if (m1->holder != MUTEX_FREE) {
#ifdef JDK_11
			sysMonitorWait(m1->mQueue, BIG_TIMEOUT, FALSE);
#else
			sysMonitorWait(m1->mQueue, BIG_TIMEOUT);
#endif
		}
		sysMonitorExit(m1->mQueue);
	}
}


/* m1 and m2 must not be hold */
INTERNALONLY void
_ilujava_acquire2 (ltj_Mutex_s * m1, ltj_Mutex_s * m2, sys_thread_t * thread, 
	ltj_Condition_s * cond, ltj_QueueEl_s * el)
{
	_ilu_Assert((m1->holder != thread) && (m2->holder != thread), "self entering lock 2");
	while (1) {
		META_ENTER;
		/* do this here so lock is entered only once */
		if (el) {
			_ilujava_excl_QueueEl(cond, el);
			_ilujava_free_QueueEl(el);
			el = 0;
		}
		/* real stuff */
		if ((m1->holder == MUTEX_FREE) && (m2->holder == MUTEX_FREE)) {
			m1->holder = thread;
			m2->holder = thread;
			if (_ilujava_lockFlag>1) {
			    m1->stackWhenAquired = _iluJava_captureJavaStack();
			    m2->stackWhenAquired = _iluJava_captureJavaStack();
			}
			META_EXIT;
			return;
		}
		META_EXIT;
		/*
		 * We need both monitors
		 * it doesn't matter which mutex we are waiting for:
		 * either one will be notified on change. 
		 */
		if (m1->holder != MUTEX_FREE) {
			sysMonitorEnter(m1->mQueue);
			if (m1->holder != MUTEX_FREE) {
#ifdef JDK_11
				sysMonitorWait(m1->mQueue, BIG_TIMEOUT, FALSE);
#else
				sysMonitorWait(m1->mQueue, BIG_TIMEOUT);
#endif
			}
			sysMonitorExit(m1->mQueue);
		} else if (m2->holder != MUTEX_FREE) {
			sysMonitorEnter(m2->mQueue);
			if (m2->holder != MUTEX_FREE) {
#ifdef JDK_11
				sysMonitorWait(m2->mQueue, BIG_TIMEOUT, FALSE);
#else
				sysMonitorWait(m2->mQueue, BIG_TIMEOUT);
#endif
			}
			sysMonitorExit(m2->mQueue);
		}
	}
}


/* See iluxport.h lt_acquire for specification */
INTERNALONLY void
_ilujava_acquire (ilu_Mutex m, ILU_ERRS((bad_param, bad_locks)) * err)
{
	ltj_Mutex_s * mutex = (ltj_Mutex_s *) m;
	sys_thread_t * thread = sysThreadSelf();
	ILU_CLER(*err);
	if (mutex->holder == thread) {
		PrintWereAquired(mutex);
		_iluJava_BUG("$ LockTech acquire: bad locks\n");
		ILU_ERR_CONS0(bad_locks, err, 0);
		return;
	}
	_ilujava_acquire1(mutex, thread, 0, 0);
}


/* See iluxport.h lt_hold for specification */
INTERNALONLY void
_ilujava_hold (ilu_Mutex m, ILU_ERRS((bad_param, bad_locks)) * err)
{
	ltj_Mutex_s * mutex = (ltj_Mutex_s *) m;
	sys_thread_t * thread = sysThreadSelf();
	ILU_CLER(*err);
	if (mutex->holder != thread) {
		PrintWereAquired(mutex);
		_iluJava_BUG("$ LockTech hold: bad locks\n");
		ILU_ERR_CONS0(bad_locks, err, 0);
		return;
	}
}


/* See iluxport.h lt_release for specification */
INTERNALONLY void
_ilujava_release (ilu_Mutex m, ILU_ERRS((bad_param, bad_locks)) * err)
{
	ltj_Mutex_s * mutex = (ltj_Mutex_s *) m;
	sys_thread_t * thread = sysThreadSelf();
	ILU_CLER(*err);
	if (mutex->holder != thread) {
		PrintWereAquired(mutex);
		_iluJava_BUG("$ LockTech release: bad locks\n");
		ILU_ERR_CONS0(bad_locks, err, 0);
		return;
	}
	META_ENTER;
	sysMonitorEnter(mutex->mQueue);
	mutex->holder = MUTEX_FREE;
	if (mutex->stackWhenAquired) {
	    ilu_free(mutex->stackWhenAquired);
	    mutex->stackWhenAquired = 0;
	}
	sysMonitorNotifyAll(mutex->mQueue);
	sysMonitorExit(mutex->mQueue);
	META_EXIT;
}


/* See iluxport.h lt_notify for specification */
INTERNALONLY void
_ilujava_notify (ilu_Condition c,
		    ILU_ERRS((bad_param)) * err)
{
	ltj_Condition_s * cond = (ltj_Condition_s *) c;
	ltj_QueueEl_s * el;
	ILU_CLER(*err);
	META_ENTER;
	el = cond->first;
	if (el) _ilujava_excl_QueueEl(cond, el);
	META_EXIT;
	if (el) {
		sysMonitorEnter(el->elQueue);
		sysMonitorNotifyAll(el->elQueue);
		sysMonitorExit(el->elQueue);
	}
}


#define DEBUGGINGWAIT 0

/* See iluxport.h lt_wait for specification */
INTERNALONLY void
_ilujava_wait (ilu_Condition c, ilu_Mutex m, ilu_Mutex m2,
		  ILU_ERRS((bad_param, bad_locks)) * err)
{
	ilu_boolean mustFree2 = 0;
	ltj_Condition_s * cond = (ltj_Condition_s *) c;
	ltj_Mutex_s * mutex = (ltj_Mutex_s *) m;
	ltj_Mutex_s * mutex2 = (ltj_Mutex_s *) m2;
	sys_thread_t * thread = sysThreadSelf();
	ltj_QueueEl_s * el;
	
	if (DEBUGGINGWAIT) ilu_DebugPrintf("$LockTech wait entered\n");
	ILU_CLER(*err);
	if (mutex->holder != thread || mutex2->holder != thread ) {
		PrintWereAquired(mutex);
		PrintWereAquired(mutex2);
		_iluJava_BUG("$ LockTech wait: bad locks");
		ILU_ERR_CONS0(bad_locks, err, 0);
		return;
	}
	
	META_ENTER;
	if (mutex != mutex2) {
		sysMonitorEnter(mutex2->mQueue);
		mutex2->holder = MUTEX_FREE;
		if (mutex2->stackWhenAquired) {
			ilu_free(mutex2->stackWhenAquired);
			mutex2->stackWhenAquired = 0;
		}
		sysMonitorNotifyAll(mutex2->mQueue);
		sysMonitorExit(mutex2->mQueue);
	}
	sysMonitorEnter(mutex->mQueue);
	mutex->holder = MUTEX_FREE;
	if (mutex->stackWhenAquired) {
		ilu_free(mutex->stackWhenAquired);
		mutex->stackWhenAquired = 0;
	}
	sysMonitorNotifyAll(mutex->mQueue);
	sysMonitorExit(mutex->mQueue);
	el = _ilujava_new_QueueEl();
	/* Take this lock before releasing the meta lock */
	sysMonitorEnter(el->elQueue);
	_ilujava_incl_QueueEl(cond, el);
	META_EXIT;
	if (DEBUGGINGWAIT) ilu_DebugPrintf("$LockTech wait done  with META\n");
#ifdef JDK_11
	sysMonitorWait(el->elQueue, VERY_BIG_TIMEOUT, FALSE);
#else
	sysMonitorWait(el->elQueue, VERY_BIG_TIMEOUT);
#endif
	sysMonitorExit(el->elQueue);

	/* No lock is hold here */
	if (DEBUGGINGWAIT) ilu_DebugPrintf("$LockTech wait done waiting\n");
	
	if (mutex == mutex2) {
	    _ilujava_acquire1(m, thread, cond, el);
	} else {
	    _ilujava_acquire2(m, m2, thread, cond, el);
	}
	if (DEBUGGINGWAIT) ilu_DebugPrintf("$LockTech wait all done\n");
}



EXPORTLIMITED void 
_ilujava_LockTechInit () 
{
	ILU_ERRS((bad_param, no_memory)) err = ILU_INIT_NO_ERR;
	/* define meta lock */
	_ilujava_metaLock = _ilujava_newJavaMonitor();
	/* register procedures */
	_ilujava_lockTech.lt_mcreate	= _ilujava_mcreate;
	_ilujava_lockTech.lt_muncons	= _ilujava_muncons;
	_ilujava_lockTech.lt_acquire	= _ilujava_acquire;
	_ilujava_lockTech.lt_hold	= _ilujava_hold;
	_ilujava_lockTech.lt_release	= _ilujava_release;
	_ilujava_lockTech.lt_mdestroy	= _ilujava_mdestroy;
	_ilujava_lockTech.lt_ccreate	= _ilujava_ccreate;
	_ilujava_lockTech.lt_cuncons	= _ilujava_cuncons;
	_ilujava_lockTech.lt_notify	= _ilujava_notify;
	_ilujava_lockTech.lt_cdestroy	= _ilujava_cdestroy;
	_ilujava_lockTech.lt_wait	= _ilujava_wait;

	ilu_SetLockTech(&_ilujava_lockTech, &err);
	if ILU_ERRNOK(err) {
		_iluJava_PANIC("$ ** LockTech LockTechInit error\n");
	}
}

static char *
ThreadID(sys_thread_t * t)
/* Returns some identification of a thread
 * Debugging aid only !
 * Leaks memory
 */
{
	char * cs = 0;
	struct Hjava_lang_String * js;
	struct Hjava_lang_Object * jt; /*actually a thread*/
	if (t) {
	    jt = (Hjava_lang_Object *) /*sys_api.h:*/ sysThreadGetBackPtr(t);
	    js = (struct Hjava_lang_String *) execute_java_dynamic_method(EE(),
            	(struct Hjava_lang_Object *) jt,
            	"toString", 
            	"(V)Ljava/lang/String;"
            	);
	    cs = allocCString(js); /*malloc is used*/
	}
	return cs;
}
