/* IluOInt.java */
/* Chris Jacobi, January 26, 1997 11:17 am PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluOInt.java,v 1.17 1997/01/26 22:05:20 jacobi Exp $ */

/*
 * This class helps in the mapping from real Java objects to kernel object 
 * See also IluJava_IluOInt.c
 */

package xerox.ilu;
import java.util.Hashtable;
import xerox.basics.Consumer0;

/**
 * This class helps in the mapping from real Java objects to kernel object;
 * It resides in the middle.  It implements what the kernel considers a 
 * LSO. While applications might hold an IluOInt; none of the operations
 * here should be used by applications.
 */
public final class IluOInt {
    /* friends */ long yIluKernelObject = 0;
    /* friends */ long yServer = 0;//true AND surrogate
    /* friends */ long yOwnerKey = 0;
    /* friends */ xerox.ilu.IluServer jServer = null; //currently only if true
    /* friends */ boolean veryInterest = false; //by kernel
    /* friends */ boolean retained = false; //by application
    // friends boolean condemned = false; //no more stored explicitely
    /* friends */ boolean destroyed = false;
    /* friends */ boolean ghost = false; //Used unprotected. Delays garbage
        //collection a turn.  Used conceptionally to un-enqueue finalizations
        //when a disguised ref is undisguised but finalization is already
        //enqueued from before undisguisation.  Since used unprotected, must
        //not participate in proof of memory safety.
    /* friends */ xerox.ilu.IluClassRep jClassRep = null;
    /* friends */ xerox.ilu.IluObject jJavaObject = null;
    
    /** No random creations
     */
    private IluOInt() {
        if (IluDebug.traceGC()<0) {
            //Prevent any garbage collection.
            //Temporary courtesy to early users until
            //finalization is trusted to work correctly.
            IluForEver.remember(this);
        }
    }
    
    /** 
     * Reserved for LSR to generate new instances.
     * Instance is not yet created all the way.
     */
    private static final IluOInt allocateOI() {
        //RT will fill in fields !
        return new IluOInt();
    }
    
    /**
     *         
     */
    public final xerox.ilu.IluObject getRealObject() {
        return this.jJavaObject;
    }
    
    //tables to temporarily prevent gc
    private static Hashtable interestTable = new Hashtable(); 
    
    //Finalize real work on harmless thread
    private static IluOIntFinalizer finThread = new IluOIntFinalizer();

    /** 
     * To prevent garbage collection of true object 
     */
    public static
    void retain(xerox.ilu.IluObject0 realObject) {
        IluOInt oi = peek(realObject);
        if (oi == null ) throw new IluRuntimeError("can't retain");
        oi.retainOI();
    }
    
    /*
     * To prevent garbage collection of true object
     */
    /*friendly*/ void retainOI() {
        if (this.destroyed) {
            throw new IluRuntimeError("destroyed; too late to retain");
        }
        xerox.ilu.IluServer server = this.jServer;
        if (server != null) {
            if (IluDebug.traceGC()>2) {
                IluDebug.log.println("! IluOInt.retain " + this);
            }
            this.retained = true;
            server.retainTable.put(this, this);
        }
    }
    
    /** 
     * Allow garbage collection of true object 
     */
    public static
    void expose(xerox.ilu.IluObject0 realObject) {
        IluOInt oi = peek(realObject);
        if (oi != null) oi.exposeOI();
    }
    
    /*
     * Preventing garbage collection of true object 
     */
    /*friendly*/ void exposeOI() {
        xerox.ilu.IluServer server = this.jServer;
        if (server!=null) {
            if (IluDebug.traceGC()>2) {
                IluDebug.log.println("! IluOInt.expose " + this);
            }
            this.retained = false;
            server.retainTable.remove(this);
        }
    }
    
    /* 
     * Called from from kernel when it starts being "very interested". 
     * Server lock is/must-be held
     */
    /*friendly*/ void showInterest() {
        if (IluDebug.traceGC()>0) {
            IluDebug.log.println("! IluOInt.showInterest " + this);
        }
        this.veryInterest = true;
        interestTable.put(this, this);
        if (! this.retained) this.ghost = true;
    }
    
    /* 
     * Called from from kernel when it is no more "very interested". 
     * Server lock is/must-be held
     */
    /*friendly*/ void removeInterest() {
        if (IluDebug.traceGC()>0) {
            IluDebug.log.println("! IluOInt.removeInterest " + this);
        }
        this.veryInterest = false;
        interestTable.remove(this);
    }
    
    /** 
     * Given a prospective real TRUE object, create the internal 
     * object. A real object may have at most one internal.
     *
     * WARNING: Don't drop the return value before calling 
     * retain or something similar. To be ultra conservative
     * also expose or retain the object according to your
     * plans; don't rely on the default being what you want.
     *
     */
    public static 
    IluOInt newOI(
    		xerox.ilu.IluObject0 realObject, 
    		xerox.ilu.IluClassRep iluClassRep,
    		xerox.ilu.IluServer server
    		) throws xerox.ilu.IluSystemException {
        IluOInt oi = allocateOI();
        oi.jClassRep = iluClassRep;
        oi.jServer = server;
        switch (oi.setupForPeek((xerox.ilu.IluObject)realObject)) {
            case 0: break;
            case 1: 
                IluDebug.clientPreError("object already known to Ilu");
                throw new IluRuntimeError("object already known to Ilu");
        }
        return oi;
    }
    
    /** 
     * Makes the real object remember the IluOInt.
     * 0 is success.
     * If realObject is already known to Ilu this will
     * return 1 and no state is changed.
     */
    private native int
    setupForPeek(xerox.ilu.IluObject realObject);
    
    /** 
     * Given a prospective real TRUE object, find an internal if it exists.
     * Returns null if none exists or existence can not be proved.
     *
     * (If caller is outside the server lock, the Ilu'ness of the returned
     * IluOInt can get lost, but the memory is safe).
     */
    public static IluOInt
    peek(xerox.ilu.IluObject0 realObject) {
        //dance to expose IluOInt to garbage collector tracing 
        //already from within appropriate monitor locking.
        IluOIntContainer container = IluOIntContainer.newOIIntContainer();
        nPeek((xerox.ilu.IluObject)realObject, container);
        return container.getContentsAndDispose();
    }

    /** 
     * Given a prospective real TRUE object, find an internal if it exists.
     * Returns null if none exists or existence can not be proved.
     * Stuffs result into a container to make it traceable to the Java
     * garbage collector before releasing internal monitor lock.
     */
    static native void
    nPeek(xerox.ilu.IluObject realObject, IluOIntContainer container);
    
    /** 
     * Given a prospective real TRUE object, find an internal if it exists.
     * Returns null if none exists or existence can not be proved.
     * IluCall argument to make traceable atomic.
     */
    /*friendly*/ static native IluOInt
    nPeekWithCall(xerox.ilu.IluObject realObject, IluCall call);

    /** 
     * Called by LSR to finish generating new SURROGATE instances.
     * This is usefull for SURROGATE objects.
     * It can be called for true objects, but the getRealObject is not null.
     */
    /*friends*/ void createSurrogateForOI() 
               throws xerox.ilu.IluSystemException {
        synchronized (this) {
            if (this.getRealObject() != null) {
                //corresponding real object is already created 
                return;
            }
            xerox.ilu.IluObject surr = null;
            if (this.jClassRep.factory != null) {
               Object stupidCast = this;
               surr = (xerox.ilu.IluObject)
                   this.jClassRep.factory.createSurrogateObject(stupidCast);
            } else if (this.jClassRep.jSurrClass != null) {
                try {
                    surr = (xerox.ilu.IluObject)
                        this.jClassRep.jSurrClass.newInstance();
                } catch (IllegalAccessException e) {
                } catch (InstantiationException e) {
                }
            }
            if (surr == null) {
                IluDebug.clientPreError("failed creating surrogate");
                throw new IluRuntimeError("failed creating surrogate");
            }
            if (surr instanceof IluAnchorSupport) {
                 ((IluAnchorSupport)surr).rememberAnchor(this);
            }
            switch (this.setupForPeek(surr)) {
                case 0: break;
                case 1: 
                    IluDebug.clientPreError("object already exists");
                    throw new IluRuntimeError("object already exists");
            }
        }
    }
    
    /** 
     */
    public final void  
    finalize() throws Throwable {
        if (IluDebug.traceObjects() > 1) {
            IluDebug.log.println("! IluOInt: finalize "
            	+ this.destroyed);
        }
        if (this.destroyed) {
            //we are not interested in this in anymore
            return;
        }
        //use separate thread so we won't lock up the real finalizer
        finThread.provide(this);
    }
    
    /* Disabled because of non-use.
     * Java calls into native too propagate finalization.
     * Locking constraints specified in native code. 
     *
     * private native void disablePeek() throws Throwable;
     */
    
    /** 
     * "Destroys object" if it is neither used or appears to be used.
     * Might or might not succeed.
     * Returns whether there is a need to re-finalize
     */ 
    /*friendly*/ native boolean destroyGentleOI();
    
    /** 
     * "Destroys object" and ignores whether the application keeps using it.
     * Further use of the object might raise whatever errors in the 
     * application; but the kernel will keep its piece of the object 
     * until it is safe. You get what you deserve.
     */ 
    /*friendly*/ native void destroyRudeOI();

    /** 
     * Publishes the SBH of the object in the local object domain
     */
    /*friendly*/ native void publishOI();
    
    /** 
     * withdrawOI is undo of publishOI
     */
    /*friendly*/ native void withdrawOI();
    
    /** 
     * See comments in IluRT0.sbhOfObject
     */
    /*friendly*/ native String sbhOfOI();
    
} //IluOInt


/*friendly*/ final class IluOIntFinalizer 
    extends xerox.basics.Consumer0 {
     private static IluFinalizeReporter reporter = 
         new IluFinalizeReporter();
        
    /*friendly*/ IluOIntFinalizer() {
        super(IluEnvironment.threadPriority(), "IluOIntFinalizer");
    } //IluOIntFinalizer
    
    protected void consume(Object obj) {
        IluOInt oi = (IluOInt) obj;
        if (oi.destroyed) {
            //the ilu object has already been destroyed
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluOInt: already destroyed " + this);
            }
        } else if (oi.ghost | oi.destroyGentleOI()) {
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluOInt: re-enable " + this);
            }
            oi.ghost = false;
        } else {
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluOInt: forget " + this);
            }
            xerox.ilu.IluObject robj = oi.getRealObject();
            if (robj instanceof omg.corba.corba_ServantFinalizable) {
                reporter.provide(robj);
            }
        }
    } //consume

} //IluOIntFinalizer


/*friendly*/ final class IluOIntContainer {
    IluOInt oi = null; //accessed by native code
    IluOIntContainer () {
    }
    static final IluOIntContainer newOIIntContainer() {
        return new IluOIntContainer();
    }
    final IluOInt getContentsAndDispose() {
        IluOInt oi = this.oi;
        this.oi = null;
        return oi;
    }
} //IluOIntContainer


/*friendly*/ final class IluFinalizeReporter 
    extends xerox.basics.Consumer0 {
        
    /*friendly*/ IluFinalizeReporter() {
        super(IluEnvironment.threadPriority(), "IluFinalizeReporter");
    } //IluFinalizeReporter
    
    protected void consume(Object obj) {
        omg.corba.corba_ServantFinalizable fobj =
            (omg.corba.corba_ServantFinalizable) obj;
        try {
            fobj.servantFinalize(); 
        } catch (Throwable t) {
        }
    } //consume

} //IluFinalizeReporter

