/* IluObjectTable.java */
/* Chris Jacobi, December 17, 1996 4:41 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluObjectTable.java,v 1.11 1996/12/18 00:58:46 jacobi Exp $ */
 
/* 
 * The native implementation is in IluJava_IluServer.c
 */
 
package xerox.ilu;
    import java.util.Hashtable;
    //import xerox.ilu.IluAnchorSupport;
    //import xerox.ilu.IluClassRep;
    //import xerox.ilu.IluDebug;
    //import xerox.ilu.IluOInt;
    //import xerox.ilu.IluRuntimeError;
    //import xerox.ilu.IluServer;
       
/**
 * An object table gives the application the ability to create true
 * objects upon presentation of an instance handle.  This is used
 * by the application to pass in to the creation of a true server.<p>
 *
 * This must be subclassed to be useful. <p>
 * WARNING: A ilu server monitor lock is hold while calls to the
 * abstract procedures; don't do random ilu calls. <p>
 * Restriction: An IluObjectTable can serve at most one server.
 */

public class IluObjectTable {
   
    private xerox.ilu.IluServer jServer = null; //native access
    
    /** 
     * Constructor client accessible as IluObjectTable's 
     * must be subclassed to be usefull.
     */
    protected IluObjectTable() {
    }
   
    /** 
     * Literals for the lifetime argument of returnTrueObject.
     * Remember obj to prevent java garbage collection.
     * Forget obj to subject it to java garbage collection. 
     */
    public static final int unspec = 0;
    public static final int remember = 1;
    public static final int forget = 2;
    
    /** 
     * Call-back to return resulting true object.
     * Do not use except from within "createTrueObject".
     */
    protected void returnTrueObject(
    		xerox.ilu.IluObject0 tobj, 
    		xerox.ilu.IluClassRep iluClass, 
    		int lifetime, //0 gives default behaviour,
    		xerox.ilu.IluAnchorSupport extraAnchor //null is ok.
    		) {
    	theTrueObject = tobj;
    	theClassRep = iluClass;
    	theLifetime = lifetime;
    	theAnchor = extraAnchor;
    }
        
    /**
     * Procedure called to require creation of the object associated  
     * with the given instance handle (or deny creation if desired).  
     * Called by server for unknown true objects.
     * Use returnTrueObject to actually return the created object.
     *
     * Called with a server monitor lock held.
     *
     * An object may be "made true" only once.
     * "Abstract": called by LSR, implemeneted in application
     * 
     */ 
    protected /*abstract*/ void createTrueObject(String ih) {
    }
   
    /** 
     * Notification that the server using this object table is 
     * being closed. Called with a server monitor lock hold.  
     * Most subclasses ignore this.
     */
    protected /*abstract*/ void objectTableFreed(){
    }
   
    /**  Accessor function */
    public final IluServer serving() {
        return this.jServer;
    }


    private xerox.ilu.IluObject0 theTrueObject = null;
    private xerox.ilu.IluClassRep theClassRep = null;
    private int theLifetime = 0;
    private xerox.ilu.IluAnchorSupport theAnchor = null;
       
   /* 
    * Calling mechanism for the createTrueObject method.
    * Called from native code.  Caller will register object with server.
    */
    private final xerox.ilu.IluOInt doCreateTrueObject(String ih) {
        try {
            IluOInt joi = null;
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluObjectTable.doCreateTrueObject requests:" + ih);
            }
            this.theTrueObject = null;
            this.theClassRep = null;
            createTrueObject(ih);
            xerox.ilu.IluObject0 ob = this.theTrueObject;
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluObjectTable.doCreateTrueObject got:" + ob);
            }
            if (ob == null) {
                //application didn't create an object: this is ok
                return null;
            }
            xerox.ilu.IluClassRep classRep = this.theClassRep;
            if (classRep==null) {
                IluDebug.clientPreError("** xerox.ilu.IluObjectTable class not specified " + ih + " " + ob);
                return null;
            }
            joi = IluOInt.newOI(ob, classRep, this.jServer); 
            	//raises IluSystemException if object already exists
            xerox.ilu.IluAnchorSupport anchor = theAnchor;
            int lifetime = theLifetime;
            if (anchor != null) {
                anchor.rememberAnchor(joi);
                if (lifetime==unspec) {lifetime = forget;}
            }
            if (ob != anchor & ob instanceof IluAnchorSupport) {
                ((IluAnchorSupport) ob).rememberAnchor(joi);
                if (lifetime==unspec) {lifetime = forget;}
            } 
            switch (lifetime) {
               case remember:
               case forget:
                   break;
               default:
                   //Since this is in object table, default is forget
                   //otherwise default would depend on classRep.isCollectible()
                   lifetime = forget;
            }
            switch (lifetime) {
                case remember:
                    joi.retainOI();
                    break;
                case forget:
                    joi.exposeOI();
                    break;
            }
            if (IluDebug.traceObjects() > 0) {
                IluDebug.log.println("! IluObjectTable.doCreateTrueObject joi: " + joi);
            }
            return joi;
        } catch (ThreadDeath td) {
            throw new ThreadDeath();
        } catch (Exception e) {     
            //Don't propagate exceptions: this is called by the ilu
            //kernel which wouldn't know what to do.
            IluDebug.clientPreError("** xerox.ilu.IluObjectTable client failure " + e);
            e.printStackTrace();
            return null;
        }
    }
   
   /* 
    * Calling mechanism for the objectTableFreed method.
    * Called from native code
    */
    private void doObjectTableFreed(){
        try {
            this.jServer.retainTable.clear();
            objectTableFreed();
        } catch (ThreadDeath td) {
            throw new ThreadDeath();
        } catch (Exception e) {     
            //Don't propagate exceptions: this is called by the ilu
            //server which wouldn't know what to do.
            IluDebug.clientPreError("** xerox.ilu.IluObjectTable failure " + e);
            e.printStackTrace();
        }
    } 
    
           
    /*  
     * Used by server creation code to tie IluObjectTable to server.
     */
    /*friendly*/ synchronized void 
    setServer(xerox.ilu.IluServer server) {
        if (this.jServer!=null && this.jServer!=server) {
            IluDebug.clientPreError("Object table can serve only 1 server");
            throw new xerox.ilu.IluRuntimeError(
            	"Object table can serve only 1 server");
        }
        this.jServer = server;
    }       
   
} //IluObjectTable

