/* IluPort.java */
/* Chris Jacobi, January 3, 1997 3:43 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluPort.java,v 1.14 1997/01/04 00:02:51 jacobi Exp $ */
 
/*
 * See IluJava_IluPort.c
 */

package xerox.ilu;

/**
 * This is the Java representation of the concept ilu_Port.<p>
 * Server applications should create ports with the IluServer interface.  
 * However this interface provides additional methods to destroy ports.
 * @see IluServer
 */
public class
IluPort {
    /*private; friendly for debugging*/ long yPort = 0;
    /*private; friendly for debugging*/ long yConn = 0;
    /*friendly*/ xerox.ilu.IluServer jServer = null;
    private String jProtocolInfo = null;
    private xerox.ilu.IluTransportInfo jTransportInfo = null;
    private xerox.ilu.IluPassport jPassport = null;
    private int pstate = portVirgin; // use the server lock to protect
    
    //State codes
    private static final int portVirgin = 0;
    private static final int portActive = 1;
    private static final int portClosed = 2;
    private static final int portDone = 3;

    //Codes to communicate with the native side
    private static final int watcherLoop = 0;
    private static final int watcherFork = 1;
    private static final int watcherClose = 2;
    private static final int watcherDone = 3;;

    private IluPort(){
    }
    
    /** constructor */
    public static IluPort createPort(
        	xerox.ilu.IluServer server,      		//required
        	String protocolInfo,            		//null is ok
        	xerox.ilu.IluTransportInfo transportInfo,	//null is ok
        	xerox.ilu.IluPassport passport  		//null is ok
        	) throws xerox.ilu.IluSystemException {
        if (IluDebug.tracePCT()>0) {
            IluDebug.log.println("! IluPort: createPort");
        }
        if (transportInfo == null) {
            String[] sa = {"sunrpcrm", "tcp_0_0"};
            transportInfo = new xerox.ilu.IluTransportInfo(sa);
        }
        if (protocolInfo==null) {
            protocolInfo = "sunrpc";
        }
        server.promisedPort = true; 
        server.finishInitialization();
        IluPort port = new IluPort();
        port.jServer = server;
        port.jProtocolInfo = protocolInfo;
        port.jTransportInfo = transportInfo;
        port.jPassport = passport;
        port.nativeInitPort();
        if (IluDebug.tracePCT()>1) {
            IluDebug.log.println("! IluPort: half way " + port.yPort);
        }
        synchronized (server.lock) {
            if (server.state>1) {
               throw new xerox.ilu.IluSomeSystemException("Bad server");
            }
            server.portTable.put(port, port);
            port.pstate = portActive;
        }
        if (IluDebug.tracePCT()>0) {
            IluDebug.log.println("! IluPort: init " + port.yPort);
        }
        if (port.yPort == 0) {
           IluDebug.clientPreError("! IluPort failure: yPort==0");
           throw new xerox.ilu.IluSomeSystemException("Bad port creation");
        }
        IluPortWatcherThread.forkPortWatcher(port);
        if (IluDebug.tracePCT()>0) {
            IluDebug.log.println("! IluPort: done " + port.yPort);
        }
        return port;
    } //createPort
    
    private native void nativeInitPort();
    private native int nativeWatchPort(); //side effect: sets yConn
    private native void nativeClosePort();
    private native void nativeDonePort();
        
    /*friendly*/ void portWatcherLoop() {
        int code;
        for (;;) {
            if (IluDebug.tracePCT()>0) {
                IluDebug.log.println("! IluPort: loop " + this.yPort);
            }
            code = nativeWatchPort();//sets yConn if necessary
            switch (code) {
               case watcherLoop: 
                   break; //spurious wakeup
               case watcherFork:
                   IluServerConnection.forkConnectionReader(yConn); 
                   yConn = 0;
                   break;
               case watcherClose:
                   //rare error situation
                   closePort(); 
                   break;
               case watcherDone: 
                   doneWithPort(); 
                   return;
            }
            Thread.yield(); //serve existing connections first
        }
    } //portWatcherLoop
    
    public void closePort() {
        xerox.ilu.IluServer server = this.jServer;
        IluPort that = null;
        if (server != null) {
            Object ob = server.portTable.remove(this);
            if (ob == this) {
                synchronized (server.lock) {
                    if (this.pstate == portActive) {
                        this.pstate = portClosed; 
                        that = this;
                    }
                }
            }
            if (that!=null) {
                //if called from portwatcher loop 
                //    this is same thread where doneWithPort will be called
                //if called from random other thread 
                //    excludes other random threads
                //    but doesn't exclude spurios io errors
                that.nativeClosePort();
            }
        }
    } //closePort
    
    private void doneWithPort() {
        closePort();
        xerox.ilu.IluServer server = this.jServer;
        IluPort that = null;
        if (server != null) {
            synchronized (server.lock) {
                if (this.pstate == portClosed) {
                    this.pstate = portDone; 
                    this.jServer = null;
                    that = this;
                }
            }
            if (that!=null) that.nativeDonePort();
        }
    } //doneWithPort
    
}//IluPort


/**
 * IluPortWatcherThread is a helper class implementing fork for IluPort.
 *
 * If you wonder why this is a separate class, congratulations.
 * The reason is that IluPort has member variables accessed from
 * native code.  If IluPort would extend Thread, those variables
 * would be aligned erronously by the JDK1.0.2 native stubber.  
 */
final class IluPortWatcherThread extends Thread {
    
    private IluPort jPort; 
    private static int priority =
        IluEnvironment.threadPriority();

    private IluPortWatcherThread() {
        super("IluPortWatcherThread");
    } //IluPortWatcherThread

    static void forkPortWatcher(xerox.ilu.IluPort port) {
        IluPortWatcherThread pw = new IluPortWatcherThread();
        pw.jPort = port; 
        pw.setDaemon(port.jServer.deamon);
        pw.setPriority(priority);  
        pw.start();
    } //forkPortWatcher

    /** shouldn't be public */
    public void run() {
        try {
            this.jPort.portWatcherLoop();
        } catch (Exception e ) {
            IluDebug.log.println(e.toString());
            IluDebug.iluPreError("** caught in connection forker..");
            //make message, but let Ilu continue...
        }
    } //run

} //IluPortWatcherThread


