/* IluServer.java */
/* Chris Jacobi, January 11, 1997 5:00 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */
 
/* $Id: IluServer.java,v 1.17 1997/01/12 01:06:15 jacobi Exp $ */
 
/* 
 * See IluJava_IluServer.c for native implementation.
 */

package xerox.ilu;
    import java.util.Hashtable;

/**
 * This is the Java representation of the an ilu_Server. <p>
 *
 * Since servers have associated threads running, don't expect
 * servers to be garbage collected.
 */
public class IluServer {
    
    private long yIluServer = 0;
    private long yPtr = 0; //just in case the c side needs a hook
    private String id; 
    private xerox.ilu.IluObjectTable jot = null;
    /*friendly*/ Hashtable portTable; 
    /*friendly*/ Hashtable retainTable = null; //prevent gc of IluOInt's
    /*friendly*/ int state = 0;
    /*friendly*/ Object lock;
    /*friendly*/ boolean deamon = false;
    private boolean hasBeenAsserted = false; /*synchronized*/
    /*friendly*/ boolean promisedPort = false;  
    private static IluServer defaultServer0 = null;
      
         
    /**
     * Application visible "constructor". <p>
     *
     * Excpect the type of objFactory to change; null objFactory is legal 
     * and results in a server not automatically generating true objects.
     * The application must create at least one port if the server should
     * communicate.
     */
    public static IluServer createServer(
                String id //null -> make one up
                ) throws xerox.ilu.IluSystemException {
        IluServer serv = new IluServer();
        if (id == null) id = xerox.ilu.IluRT0.inventID();
        serv.id = id;
        if (IluDebug.tracePCT()>0) {
            IluDebug.log.println("! IluServer: create server for " + id);
        }
        return serv;
    } //createServer
    
    
    /**
     * "accessor". <p>
     */
    /*friendly*/ static IluServer defaultServer() 
            throws xerox.ilu.IluSystemException {
        IluServer serv = defaultServer0;
        if (serv == null) {
            serv = createServer(null);
            serv.finishInitialization();
            defaultServer0 = serv;
        }
        return serv;
    } //defaultServer
    

    /**
     * Creates a port and adds it to the server.
     */
    public IluPort createPort(
            String protocolInfo,                	//null -> default
            xerox.ilu.IluTransportInfo transportInfo,	//null -> default
            IluPassport passport                 	//null -> default
            ) throws xerox.ilu.IluSystemException {
        IluPort port;
        this.promisedPort = true; 
        this.finishInitialization();
        port = IluPort.createPort(this, protocolInfo, transportInfo, passport);
        return port;
    } //createPort


    /**
     * Sets the "deamon" flag for ports.
     * Must be called before ports are created.
     * Not synchronized; this is for the server creator only.
     * Returns "this" for conveniance.
     */
    public IluServer setDeamonFlag(boolean deamon)
            throws xerox.ilu.IluSystemException {
        this.deamon = deamon;
        return this;
    } //setDeamonFlag
    
    
    /**
     * Sets the object table for the server. <p>
     *
     * Must be called before any ports are created.  Will not
     * be advertized unless ports a created or finishInitialization
     * is called directly or indirectly.<p>
     *
     * Not synchronized; this is for the server creator only.
     * An object table can have exactly one server.
     * A server can have at most one object table.
     * Returns "this" for conveniance.<p>
     *
     * Excpect the type of objFactory to change; null objFactory is legal 
     * and results in a server not automatically generating true objects.
     */
    public IluServer setObjectTable(xerox.ilu.IluObjectTable ot)
            throws xerox.ilu.IluSystemException {
        if (this.hasBeenAsserted) {
            throw new xerox.ilu.IluSystemException(
                "must be called before initialization is asserted"
                );
        }
        if (this.jot != null) {
            throw new xerox.ilu.IluSystemException(
                "server already has an object table"
                );
        }
        this.jot = ot;
        ot.setServer(this);
        return this;
    } //setObjectTable
    
    
    /**
     * Makes sure enough of the server has been initialized
     * so it can really be used.
     * Returns "this" for conveniance.
     */
    public IluServer finishInitialization(
            ) throws xerox.ilu.IluSystemException {
        if (! this.hasBeenAsserted) {
            boolean promisedPort = this.promisedPort;
            if (this.state == 29) {
                throw new xerox.ilu.IluSystemException(
                    "use of destroyed server"
                );
            }
            synchronized (this) {
                if (! this.hasBeenAsserted) {
                    if (IluDebug.tracePCT()>0) {
                        IluDebug.log.println("! IluServer: finish init " 
                            + this.id);
                    }
                    this.nativeOfCreateServer(this.id);
                    this.hasBeenAsserted = true;
                }
            }
            if (!this.promisedPort) { 
                this.createPort(null, null, null); 
            }
        }
        if (defaultServer0 == null) {
            defaultServer0 = this;
        }
        return this;
    } //finishInitialization
    
    
    /**
     * Accessor function 
     */
    public final String serverId() {
        return this.id;
    }
    
    
    /*
     * Private<br>
     * Implementation for createServer.
     */
    private native void nativeOfCreateServer(String id);
    
    
    /*
     * Constructor disabled: use createServer.<br>
     * Prevents creation of random servers.
     */
    private IluServer() {
        this.retainTable = new Hashtable();
        this.portTable = new Hashtable();
        this.lock = retainTable;
    } 
    
    /**
     * Destroys this server.<p>
     * Should check classloader for permission !?.
     */
    public void destroyServer() {
        if (this == defaultServer0) {
            defaultServer0 = null;
        }
        synchronized (this.lock) {
           if (this.state>=9) return;
           this.state = 19;
        };
        Hashtable safeCopy = (Hashtable) this.portTable.clone();
        java.util.Enumeration enum = safeCopy.keys();
        while(enum.hasMoreElements()) {
            IluPort p = (IluPort) enum.nextElement();
            p.closePort();
        }
        this.nativeOfDestroyServer();
        this.state = 29;
         this.hasBeenAsserted = false;
    } //destroyServer
    
    
    private native void nativeOfDestroyServer();
    private native void nativeOfFreeServer();
    
    
    protected void finalize() throws Throwable {
        this.destroyServer();
        this.state = 99;
        this.nativeOfFreeServer();
    } //finalize
    
    
    static {
        IluInit.init();
    }
    
} //IluServer


