;;; -*- Mode:Lisp; Package: ILU; Syntax:COMMON-LISP; Base:10 -*-
#|
Copyright (c) 1991-1995 Xerox Corporation.  All Rights Reserved.  

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: ilu.lisp,v 1.86 1997/01/29 03:24:18 janssen Exp $
|#

(cl:in-package :ilu)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Constants
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;; packet types ;;;;;;;;;;;;;;;;;;;;

(defconstant +packet-type-request+	0	"type value for request packet")
(defconstant +packet-type-reply+	1	"type value for reply packet")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Objects
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *language-id* nil "will be set to id for LISP language mapping")

(defun my-language-index () *language-id*)

(defun class-or-class-name->type (c-or-c-n)
  (if (or (stringp c-or-c-n)
	  (symbolp c-or-c-n))
      (find-class c-or-c-n)
    c-or-c-n))

(defun class-or-class-name->name (c-or-c-n)
  (if (or (stringp c-or-c-n)
	  (symbolp c-or-c-n))
      c-or-c-n
    (class-name c-or-c-n)))

(defmethod ilu-class-id (something)
  (declare (ignore something))
  nil)

(defmethod ilu-class-name (classname)
  (declare (ignore classname))
  nil)

(defmethod ilu-class-record (classname)
  (declare (ignore classname))
  nil)

(defgeneric ilu-class-info (classname &optional info-type))

(defun get-class-name (class)
  (ilu-class-name (class-or-class-name->name class)))

(defclass ilu-object ()
  ((ilu-cached-kernel-obj :initform nil :accessor ilu-cached-kernel-obj)	; C-runtime handle
   (ilu-server     	:initform nil :accessor ilu-server)
   (ilu-class	   	:initform nil :accessor ilu-class)
   (ilu-publish-proof	:initform nil :accessor ilu-publish-proof)
   (ilu-instance-id	:initform nil)
   (ilu-server-id	:initform nil)
   (ilu-string-binding-handle :initform nil)))

(defmethod initialize-instance :after ((self ilu-object)
				       &key &allow-other-keys)
  (optional-finalization-hook self))

;; before:  NOT Inside (server, class)
;; after:  if result non-nil, Inside(server, class) of result
(defmethod kernel-obj ((self ilu-object))
  (ilu_enter-server (ilu-server self) (ilu-class self))
  (unless (ilu-cached-kernel-obj self)
    (ilu_exit-server (ilu-server self) (ilu-class self)))
  (ilu-cached-kernel-obj self))

(defmethod instance-id ((self ilu-object))
  (or (slot-value self 'ilu-instance-id)
      (if (null (ilu-cached-kernel-obj self))
	  "(no kernel object)"
	(let ((kernel-object (kernel-obj self)))
	  (if (null kernel-object)
	      nil
	    ;; now Inside(SERVER,CLASS)
	    (let ((id-string (ilu_instance-id kernel-object)))
	      (ilu_exit-server (ilu-server self) (ilu-class self))
	      ;; now out of Server
	      (setf (slot-value self 'ilu-instance-id) id-string)
	      id-string))))))

(defmethod server-id ((self ilu-object))
  (or (slot-value self 'ilu-server-id)
      (if (ilu-server self)
	  (setf (slot-value self 'ilu-server-id) (id-of-kernel-server
						  (ilu-server self)))
	"(no kernel server)")))

(defmethod print-object ((self ilu-object) stream)
  (format stream "#<~s ~a@~a>"
	  (type-of self)
	  (instance-id self)
	  (server-id self)))

(defclass ilu-singleton-object (ilu-object) ())

(defmethod ilu-class-name ((self ilu-object))
  (ilu-class-info self :name))

(defmethod find-ilu-class-name (object)
  (declare (ignore object))
  nil)

(defmethod find-ilu-class-name ((object ilu-object))
  (dolist (class
	   #+ACLPC (class-precedence-list (class-of object))
	   #-ACLPC (clos:class-precedence-list (class-of object))
	   )
    (let ((name (ilu-class-name (class-name class))))
      (when name
	(return name)))))

(defmethod find-ilu-class-id ((object ilu-object))
  (dolist (class
	   #+ACLPC (class-precedence-list (class-of object))
	   #-ACLPC (clos:class-precedence-list (class-of object))
	   )
    (let ((id-pointer (ilu-class-id class)))
      (when id-pointer
	(return id-pointer)))))

(defmethod find-ilu-class-record ((object ilu-object))
  (dolist (class
	   #+ACLPC (class-precedence-list (class-of object))
	   #-ACLPC (clos:class-precedence-list (class-of object))
	   )
    (let ((record (ilu-class-record (class-name class))))
      (when (and record (not (c-null-pointer-p record)))
	(return record)))))

(defmethod rpc-method ((self ilu-object) putative-class method-id)
  (find-method-by-id putative-class method-id))

(defmethod ilu-class-id ((self ilu-object))
  (ilu-class-info self :id))

(defvar *ilu-to-lisp-class-table* (make-hash-table) "KEY is kernel class record, VALUE is CLOS class")
(defvar *lisp-to-ilu-class-table* (make-hash-table) "KEY is CLOS class, VALUE is ILU kernel class record")
(defvar *exception-repository-ids* (make-hash-table) "KEY is lisp condition name, VALUE is CORBA repository ID string")

(defvar *ilu-initialized* nil "t if the ILU kernel has been initialized")

(defvar *version* nil "set to string determining the current version of ILU")

(defun make-exception-vector (interface exceptions)
  (let ((vec (ilulisp_cons-exception-vector (length exceptions))))
    (do ((i 0 (1+ i))
	 (e exceptions (cdr e)))
	((null e) nil)
      (ilulisp_set-exception-vector
       vec i (ilu_define-exception interface (symbol-name (car e)))))
    vec))

(defun ilu-name->interface (name)
  "strips off the last period-separated component from a string"
  (do ((i (1- (length name)) (1- i)))
      ((< i 0) name)
      (if (eql (char name i) #\.)
	  (return (subseq name 0 i)))))

(defun make-ilu-class-record (class-info)
  (let* ((superclasses (dig-out-class-info class-info :superclass-ids))
	 (superclass-vec (ilulisp_start-superclass-vector
			  (length superclasses) (car superclasses)))
	 (superclass-index 0))
    (dolist (superclass (cdr superclasses))
      (incf superclass-index)
      (ilulisp_add-superclass-to-vector
       superclass-vec
       (length superclasses)
       superclass-index
       superclass))
    (let ((class-record (ilu_define-object-type
			 (dig-out-class-info class-info :name)
			 (dig-out-class-info class-info :brand)
			 (dig-out-class-info class-info :id)
			 (dig-out-class-info class-info :singleton)
			 (dig-out-class-info class-info :optional-p)
			 (dig-out-class-info class-info :collectible-p)
			 (dig-out-class-info class-info :authentication)
			 (length (dig-out-class-info class-info :methods))
			 (length superclasses)
			 superclass-vec))
	  (methods (dig-out-class-info class-info :methods))
	  (method-index 0)
	  (intfc (ilu-name->interface (dig-out-class-info class-info :name))))
      (dolist (method methods)
	(ilu_define-method
	 class-record
	 method-index
	 (dig-out-method-info method :c-name)
	 (dig-out-method-info method :index)
	 (dig-out-method-info method :functional-p)
	 (dig-out-method-info method :asynchronous-p)
	 (length (dig-out-method-info method :exceptions))
	 (make-exception-vector intfc (dig-out-method-info method :exceptions)))
	(incf method-index))
      (unless (ilu_object-type-defined class-record)
	      (error "ilu_object-type-defined failed"))
      class-record)))

(defun initialize-ilu-class (class-name class-info)
  (let ((class (find-class class-name)))
    (if *ilu-initialized*
	(let ((record (make-ilu-class-record class-info)))
	  ; (ilu_register-class record)
	  (setf (gethash class *lisp-to-ilu-class-table*) record)
	  (setf (gethash record *ilu-to-lisp-class-table*) class))
      (setf (gethash class *lisp-to-ilu-class-table*) (c-null-pointer)))))

(defun initialize-ilu ()
  ;; set up LockTech and WaitTech
  (unless *ilu-initialized*
    (setf *ilu-initialized* t)
    (setf *version* (get-ilu-version))
    (setf *language-id* (ilu_register-language "LISP"))
    (when (not (string-equal *version* (get-lisp-runtime-version)))
      (error "Lisp runtime compiled against ILU version ~s, ILU kernel library compiled against ILU version ~s"
	     (get-lisp-runtime-version) *version*))
    ;;
    (initialize-locking)	;; defined in "ilu-<implementation>"
    ;;
    (maphash #'(lambda (v k)
		 (let ((classname (class-name v))
		       (r (ilu_find-class-from-id (ilu-class-info (class-name v) :id))))
		   (unless
		       (string-equal *version*
				     (ilu-class-info classname :ilu-version))
		     (warn "Stubs for ~s compiled with ILU stubber for version ~s, but current kernel is at ILU version ~s"
			   v (ilu-class-info classname :ilu-version)
			   *version*))
		   (cond
		    ((c-null-pointer-p r)
		     (initialize-ilu-class classname (ilu-class-info classname)))
		    ((not (= k r))
		     (setf (gethash v *lisp-to-ilu-class-table*) r)
		     (setf (gethash r *ilu-to-lisp-class-table*) v)))))
	     *lisp-to-ilu-class-table*)
    ;;
    (initialize-gc-callback)	;; in ilu-server.lisp
    ))

(defmacro ensure-initialized ()
    `(unless *ilu-initialized* (initialize-ilu)))

(defmethod singleton-p ((self ilu-object))
  nil)

(defmethod singleton-p ((self ilu-singleton-object))
  t)

(defmethod initial-pointer-reftype ((self ilu-object))
  :weak)

;; Inside(server,class)
(defun language-specific-object (obj)
  (let ((index (get-language-specific-object obj (my-language-index))))
    (unless (zerop index)
      (lookup-registered-lisp-object index))))

;; Inside(server,class)
(defsetf language-specific-object (obj) (lspo)
  `(progn
     (register-language-specific-object
      ,obj
      (register-lisp-object ,lspo :reftype (initial-pointer-reftype ,lspo))
      (my-language-index))
     ,lspo))

;; before: Inside(server,class)
;; after:  Main invariant holds
(defun ilu-object->instance (class kernel-obj class-record)
  (let ((lisp-obj (language-specific-object kernel-obj)))
    (unless lisp-obj
      (setq lisp-obj (make-instance class)) 
      (setf (language-specific-object kernel-obj) lisp-obj)
      (setf (ilu-server lisp-obj) (ilu_ilu-server kernel-obj))
      (setf (ilu-class lisp-obj) (ilu_ilu-class kernel-obj))
      (setf (ilu-cached-kernel-obj lisp-obj) kernel-obj))
    (ilu_exit-server (ilu-server lisp-obj) class-record)
    lisp-obj))

;; L1_sup < smu; L2, Main unconstrained
(defun consider-sbh (newsbh)
  (multiple-value-bind (result new-server errorstruct)
      (ilu_consider-sbh newsbh)
    (declare (ignore new-server))
    (ecase result
      (0 (error "ILU signals error ~s" errorstruct))
      (1 (warn "non-reified server ~s" newsbh))
      (2 t)
      (3 (warn "~s is a true instance"))
      (4 t)
      (5 (progn
	   (format t "the server has been switched to the new contact info~%") t))
      )))

;; Main invariant holds
(defun sbh->instance (class-name sbh &key (mstid nil) (new-ih nil) (new-contact-info nil))
  "Given the CLASS-NAME and STRING-BINDING-HANDLE, return an ilu-object"
  (declare (string sbh mstid))
  (ensure-initialized)
  (let ((the-sbh (if new-ih
		     (concatenate 'string new-ih (subseq sbh (position #\@ sbh)))
		   sbh)))
    (when new-contact-info
      (consider-sbh the-sbh))
    (let ((ilu-obj (ilu_object-of-sbh
		    the-sbh
		    (ilu-class-record class-name))))
      (if (c-null-pointer-p ilu-obj)
	  nil
	;; now Inside (server(ilu-obj), class(ilu-obj))
	(ilu-object->instance class-name ilu-obj (ilu_ilu-class ilu-obj))
	;; now Main holds
	))))

;; Main invariant holds
(defmethod string-binding-handle ((self ilu-object))
  (or (slot-value self 'ilu-string-binding-handle)
      (let ((kobj (kernel-obj self)))
	(when kobj
	  (setf (slot-value self 'ilu-string-binding-handle)
	    (ilu_sbh-of-object kobj))
	  (ilu_exit-server (ilu-server self) (ilu-class self))
	  (slot-value self 'ilu-string-binding-handle)))))

(defun object-read (class call static-type-name)
  (multiple-value-bind (ilu-obj class-record)
      (ilulisp_input-object-id call (ilu-class-record
				     static-type-name) 0)
    (if (and ilu-obj (not (c-null-pointer-p ilu-obj)))
	(let ((lisp-obj (ilu-object->instance
			 (gethash class-record *ilu-to-lisp-class-table* class)
			 ilu-obj
			 (ilu-class-record static-type-name)
			 )))
	  lisp-obj)
      (unless (ilu-class-info static-type-name :optional-p)
	(error "Couldn't read kernel object of class ~s with pclass ~s~%" class static-type-name)))))

(defmethod object-write-prep (call self pclass)
  (declare (ignore call self pclass))
  (values))

(defun object-write (call self pclass)
  (unless self
    (unless (ilu-class-info pclass :optional-p)
      (error "NIL object of non-collectible class ~s passed" pclass)))
  (object-write-prep call self pclass)
  (object-id-write call
		   (if self (kernel-obj self) (c-null-pointer))
		   0 (ilu-class-record pclass))
  self)

(defun object-id-size (call kobj discriminator-p static-type exit-server-p)
  (let ((size (ilu_object-id-size call kobj discriminator-p static-type)))
    (when (and (not (c-null-pointer-p kobj)) exit-server-p)
      (ilu_exit-server (car exit-server-p) (cdr exit-server-p)))
    size))

(defun object-size (call self pclass)
  (unless self
    (unless (ilu-class-info pclass :optional-p)
      (error "NIL object of non-collectible class ~s passed" pclass)))
  (object-write-prep call self pclass)
  (object-id-size call (if self (kernel-obj self) (c-null-pointer))
		  0 (ilu-class-record pclass)
		  (cons (ilu-server self)
			(ilu-class self))))

(defmethod publish ((self ilu-object))
  (let ((kobj (kernel-obj self)))
    (when kobj
      (setf (ilu-publish-proof self) (ilu_publish-object kobj)))))

(defmethod withdraw ((self ilu-object))
  (when (stringp (ilu-publish-proof self))
    (let ((kobj (kernel-obj self)))
      (when kobj
	(let ((stat (ilu_withdraw-published-object kobj (ilu-publish-proof self))))
	  (if stat
	    (setf (ilu-publish-proof self) nil))
	  stat)))))

(defun lookup (sid ih type)
  (ensure-initialized)
  (let* ((class-record (ilu-class-record (class-or-class-name->name type)))
	 (kobj (ilu_lookup-object-by-oid sid ih class-record)))
    (when (not (c-null-pointer-p kobj))
      (ilu-object->instance (class-or-class-name->type type) kobj class-record))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;;  Exceptions
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(eval-when (eval load compile)
   (define-condition rpc-exception (serious-condition)
     ((exception-value :accessor exception-value :initarg :exception-value))
     (:documentation
      "All exceptions signalled by RPC servers are subtypes of this exception")
     (:report (lambda (c s)
		(write-string
		 (documentation (type-of c) 'type) s)))))

(defgeneric exception-value-read (call exception))
(defmethod exception-value-read (call (exception rpc-exception))
  (declare (ignore call)))

(defgeneric exception-value-size (call exception))
(defmethod exception-value-size (call (exception rpc-exception))
  (declare (ignore call))
  0)

(defgeneric exception-value-write (call exception))
(defmethod exception-value-write (call (exception rpc-exception))
  (declare (ignore call)))

(defmethod initialize-instance :after ((exception rpc-exception)
				       &key (call nil call-p)
				       &allow-other-keys)
  (when call-p
    (setf (exception-value exception)
      (exception-value-read call exception))))

(defun signal-exception (call position e)
  (begin-exception call position (exception-value-size call e))
  (exception-value-write call e)
  (finish-exception call))

(define-condition protocol-error (rpc-exception) ()
  (:documentation "RPC Error"))

;;;;;;;;;;;;;;;;;;;; protocol error values ;;;;;;;;;;;;;;;;;;;;

(defconstant +protocol-exception-success+           0
  "value returned when RPC protocol has succeeded")
(defconstant +protocol-exception-no-such-class+     1
  "value for bad class ID at server")
(defconstant +protocol-exception-version-mismatch+  2
  "value for bad version at server")
(defconstant +protocol-exception-no-such-method+    3
  "value for bad method ID at server")
(defconstant +protocol-exception-bad-arguments+     4
  "value for invalid argument formatting")
(defconstant +protocol-exception-unknown+           5
  "catchall for protocol errors")
(defconstant +protocol-exception-no-connection+     6
  "value for bad or lost connection")
(defconstant +protocol-exception-request-rejected+  7
  "value for server refusing to look at packets")
(defconstant +protocol-exception-request-timeout+   8
  "no answer within timeout period")

;;;;;;;;;;;; CORBA system exceptions ;;;;;;;;;;;;;;;;;;;;;;;;;

(defconstant +completed-yes+   0)
(defconstant +completed-no+    1)
(defconstant +completed-maybe+ 2)

(defun default-completion-status () +completed-maybe+)

(define-condition corba-system-exception (rpc-exception)
  ((completion-status :accessor completion-status :initarg :completion-status))
  (:documentation "Standard CORBA system exception"))

(define-condition corba-unknown (corba-system-exception) ()
  (:documentation "the unknown exception"))

(define-condition corba-bad-param (corba-system-exception) ()
  (:documentation "an invalid parameter was passed"))

(define-condition corba-no-memory (corba-system-exception) ()
  (:documentation "dynamic memory allocation failure"))

(define-condition corba-imp-limit (corba-system-exception) ()
  (:documentation "violated implementation limit"))

(define-condition corba-comm-failure (corba-system-exception) ()
  (:documentation "communication failure"))

(define-condition corba-inv-objref (corba-system-exception) ()
  (:documentation "invalid object reference"))

(define-condition corba-no-permission (corba-system-exception) ()
  (:documentation "no permission for attempted operation"))

(define-condition corba-internal (corba-system-exception) ()
  (:documentation "ORB internal error"))

(define-condition corba-marshal (corba-system-exception) ()
  (:documentation "error marshalling parameter/result"))

(define-condition corba-initialize (corba-system-exception) ()
  (:documentation "ORB initialisation failure"))

(define-condition corba-no-implement (corba-system-exception) ()
  (:documentation "operation implementation unavailable"))

(define-condition corba-bad-typecode (corba-system-exception) ()
  (:documentation "bad typecode"))

(define-condition corba-bad-operation (corba-system-exception) ()
  (:documentation "invalid operation"))

(define-condition corba-no-resources (corba-system-exception) ()
  (:documentation "insufficient resources for request"))

(define-condition corba-no-response (corba-system-exception) ()
  (:documentation "response to request not yet available"))

(define-condition corba-persist-store (corba-system-exception) ()
  (:documentation "persistent storage failure"))

(define-condition corba-bad-inv-order (corba-system-exception) ()
  (:documentation "routine invocations out of order"))

(define-condition corba-transient (corba-system-exception) ()
  (:documentation "transient failure - reissue request"))

(define-condition corba-free-mem (corba-system-exception) ()
  (:documentation "cannot free memory"))

(define-condition corba-inv-ident (corba-system-exception) ()
  (:documentation "invalid identifier syntax"))

(define-condition corba-inv-flag (corba-system-exception) ()
  (:documentation "invalid flag was specified"))

(define-condition corba-intf-repos (corba-system-exception) ()
  (:documentation "error accessing interface repository"))

(define-condition corba-context (corba-system-exception) ()
  (:documentation "error processing context object"))

(define-condition corba-obj-adapter (corba-system-exception) ()
  (:documentation "failure detected by object adapter"))

(define-condition corba-data-conversion (corba-system-exception) ()
  (:documentation "data conversion error"))

(define-condition corba-object-not-exist (corba-system-exception) ()
  (:documentation "non-existent object - delete reference"))
