;;  -*- Mode: Scheme; -*-
;;
;; Copyright (c) 1996 Siemens Corporation.  All Rights Reserved.  
;;
;; Unlimited use, reproduction, and distribution of this software is
;; permitted.  Any copy of this software must include both the above
;; copyright notice of Siemens Corporation and this paragraph.  Any
;; distribution of this software must comply with all applicable United
;; States export control laws.  This software is made available AS IS,
;; and SIEMENS CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
;; INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
;; AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
;; PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
;; THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
;; CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
;; SIEMENS CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
;;
;; $Author: bnell $
;; $Date: 1996/11/08 19:08:56 $
;; $Source: /home/yugo/kwuftn/CVS-repository/ilu-scheme-runtime/ilu.scm,v $
;; $Revision: 1.1.1.1 $
;;
(require 'ilu-object)
(require 'ilu-server)
(require 'format)

;; some standard ilu errors
(define ilu:success #t)
(define ilu:protocol-error "ilu.ProtocolError")

;; ilu protocol exception types
(define ilu:protocol-exception-success                 0)
(define ilu:protocol-exception-no-such-class-at-server 1)
(define ilu:protocol-exception-class-version-mismatch  2)
(define ilu:protocol-exception-no-such-method-on-class 3)
(define ilu:protocol-exception-garbage-arguments       4)
(define ilu:protocol-exception-unknown                 5)
(define ilu:protocol-exception-lost-connection         6)
(define ilu:protocol-exception-request-rejected        7)
(define ilu:protocol-exception-request-timeout         8)
(define ilu:protocol-exception-not                  1000)

;; internal types for ipc
(define ilu:rcv-req-stat-noop    0)
(define ilu:rcv-req-stat-quit    1)
(define ilu:rcv-req-stat-request 2)

;; should generics be in ilu package? or in any package?
(define-operation (run obj stop))
(define-operation (exit obj stop))
(define-operation (register-input-handler obj fd handler))
(define-operation (register-output-handler obj fd handler))
(define-operation (unregister-input-handler obj fd))
(define-operation (unregister-output-handler obj fd))
(define-operation (create-alarm obj))
(define-operation (set-alarm obj alarm time proc))
(define-operation (clear-alarm obj alarm))

;; main loop class
(define (make-ilu-main-loop)
  (object
   ;; run main loop
   ((run this stop)
    (error "ilu-main-loop is an abstract class"))

   ;; exit main loop
   ((exit this stop)
    (error "ilu-main-loop is an abstract class"))

   ;; register input handle for a file descriptor
   ;; (i dunno if scheme has fd's like unix....)
   ;; in any case handler can be a lambda fn or a closure or anything
   ((register-input-handler this fd handler)
    (error "ilu-main-loop is an abstract class"))

   ;; register output handle for a file descriptor
   ;; (i dunno if scheme has fd's like unix....)
   ;; in any case handler can be a lambda fn or a closure or anything
   ((register-output-handler this fd handler)
    (error "ilu-main-loop is an abstract class"))

   ;; unregister input handler for file descriptor
   ((unregister-input-handler this fd)
    (error "<ilu-main-loop> is an abstract class"))

   ;; unregister output handler for file descriptor
   ((register-output-handler this fd)
    (error "ilu-main-loop is an abstract class"))

   ;; create an alarm object
   ((create-alarm this)
    (error "ilu-main-loop is an abstract class"))

   ;; set an alarm object with a time and a handler
   ;; time is value in seconds represented by a double
   ;; proc can be any proc, lambda, closure, etc.
   ((set-alarm this alarm time proc)
    (error "ilu-main-loop is an abstract class"))

   ;; reset an alarm to do nothing
   ((clear-alarm this alarm)
    (error "ilu-main-loop is an abstract class"))
   ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; default ok return code
(define ilu:reply-success #f)

(define-operation (get-return-code obj))
(define-operation (set-return-code obj code))

(define-operation (get-caller-passport obj))
(define-operation (set-caller-passport obj passp))

(define-operation (get-status-value obj))
(define-operation (set-status-value obj value))

(define (make-ilu-status args)
  (let ((return-code (if args (list-ref args 0) #f))
	(caller-passport (if args (list-ref args 1) #f))
	(value (if args (list-ref args 2) #f)))
    (object
     ((get-return-code this) return-code)
     ((set-return-code this rc) (set! return-code rc))

     ((get-caller-passport this) caller-passport)
     ((set-caller-passport this cp) (set! caller-passport cp))

     ((get-status-value this) value)
     ((set-status-value this v) (set! value v))
     )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; this seems to be ignored in ilu library code so for now...
(define ilu:size-of-discriminator 4)

;; map recursively to nested lists or vectors
;; sequences are lists
;; arrays are vectors
(define (ilu:map-rec proc seq)
  (if seq
      (if (list? seq)
	  (if (list? (car seq))
	      (ilu:map-rec proc (car seq))
	      (map proc seq))
	  (if (vector? (vector-ref seq 0))
	      (ilu:map-rec proc (vector-ref seq 0))
	      (map proc (vector->list seq)))
	  )
      #f)
  )

;; sequences are nested lists
(define (ilu:input-sequence-rec proc diml)
  (if diml
      (do ((i 0 (+ i 1))
	   (l #f (append l (list (ilu:input-sequence-rec proc (cdr diml))))))
	  ((= i (car diml)) l))
      (apply proc #f)
      )
  )

;; arrays are nested vectors
(define (ilu:input-array-rec proc diml)
  (if diml
      (do ((i 0 (+ i 1))
	   (v (make-vector (car diml)) v))
	  ((= i (car diml)) v)
	(vector-set! v i (ilu:input-array-rec proc (cdr diml))))
      (apply proc #f)
      )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; GC callback object
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(if (not (defined? ilu:gc-did-init)) (define ilu:gc-did-init #f))
(define ilu:gc-callback-object #f)
(define ilu-gc-callback:class-record #f)

(define (ilu-gc-callback:create obj)
  (let ((nobj (make <ilu-gc-callback>)))
    (set-rpc-object nobj obj)
    (ilu:set-language-specific-object obj nobj)
    nobj))

(define (make-ilu-gc-callback arg)
  (let* ((our-server arg)
	 (obj (object-with-ancestors
	       ((ilu-object (make-ilu-object)))
	       ((get-server this) our-server)
;;	       ((set-server this server) (set! our-server server))
	       )))
    (set-instance-class-record obj (ilu:get-gc-callback-class))
    obj))

(define (ilu:gc-init)
  (if (not ilu:gc-did-init)
      (let ((server #f))
	(set! ilu:gc-did-init #t)

	(set! ilu-gc-callback:class-record (ilu:get-gc-callback-class))
	(ilu-object:register-surrogate-creator
	 ilu-gc-callback:class-record ilu-gc-callback:create)

	(set! server (ilu-server:create #f #f))
	(ilu-server:add-port server #f #f #t)
	(set! ilu:gc-callback-object (make-ilu-gc-callback server))
	(ilu-object:register-as-gc-callback ilu:gc-callback-object)
	))
  )

(ilu:gc-init)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; ILU stand-alone-repl (input loop for interactive use)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (ilu:init-stand-alone-repl argc argv)
  (ilu:register-input-handler *stdin* ilu:synthetic-repl)
  )

;; clean this up
(define (ilu:synthetic-repl)
  (let ((repl (lambda ()
		(pretty-print (eval (read)))
		(format #t "~!ilu> ")
		)))
    (with-dynamic-root
     (lambda () (with-input-from-port *stdin* repl))
     (lambda (errcode)
       (with-input-from-port *stdin*
	 (lambda ()
	   (cond
	    ((= errcode repl-quit) #t)
	    (#t (repl)))))))))
