/* isl2java.c */
/* Chris Jacobi, January 31, 1997 5:30 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

 
/* $Id: isl2java.c,v 1.32 1997/02/01 01:37:53 jacobi Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "name.h"
#include "util.h"
#include "genstub.h"
#include "prttree.h"
#include "context.h"


#ifdef _IS_POSIX
#include <unistd.h>
#include <errno.h>
#endif				/* _IS_POSIX */

static boolean
matchImportedWithString(void *p1, void *p2)
{
    return strcmp(((Imported) p1)->name, (char *) p2) == 0;
}


static void
addInterfaceToImports(list imports, Interface ifc)
{
    char *name = ifc->name->base_name;

    if (list_find(imports, matchImportedWithString, name) == 0) {
	Imported i = (Imported) iluparser_Malloc(sizeof *i);

	i->name = name;
	i->filename = 0;
	list_insert(imports, i);
    }
}


typedef struct {
    list imports;
    list sortedTypes;
} PtRock;


static void
processType(Type t, PtRock * r)
{
    static list pendingTypes;

    if (t->importInterfaceName != 0) {
	Type urt = myUrType(t);
	if (urt->interface) {
            addInterfaceToImports(r->imports, urt->interface);
	}
	return;
    }

    if (list_find(r->sortedTypes, matchPointer, t) != 0)
	return;
    if (pendingTypes == 0)
	pendingTypes = new_list();
    if (list_find(pendingTypes, matchPointer, t) != 0)
	fatal("internal error: circularity in types");
    list_insert(pendingTypes, t);

    if (t->description) {
	switch (t->description->type) {
            case object_Type:
		{
                    Class c = t->description->structuredDes.object;

                    list_enumerate(c->superclasses, (EnumProc) processType,
				   r);
		}
		break;

            default:
		/* null */
		break;
	}
    }
    else {
	/* alias */
	processType(t->supertype, r);
    }

    list_remove(pendingTypes, t);
    list_insert(r->sortedTypes, t);
}


static void
normalizeInterface(Interface ifc)
{
    PtRock rock;
    rock.imports = ifc->imports;
    rock.sortedTypes = new_list();
    list_enumerate(ifc->types, (EnumProc) processType, &rock);
    list_clear(ifc->types, FALSE);
    ifc->types = rock.sortedTypes;
}



static void
generate(Interface ifc)
{
    currentIfc = ifc;
    normalizeInterface(ifc);
    generateStub(ifc);
}


typedef enum {
    OptDebug,
    OptDir, OptDir1, OptXrx, OptXrx1, 
    OptSun, OptSun1, OptFlat, OptFlat1, 
    OptPrefix, OptPrefix1, OptNoPrefix, OptNoPrefix1,
    OptNoServ, OptNoServ1, 
    OptDel, OptDel1, OptNull
} Option;

typedef struct {
    char *optName;
    Option opt;
    boolean ifaceArg;
} OptTuple;

static OptTuple optionTable[] =
{
    {"debug", OptDebug, 0},
    {"dir", OptDir, 0},
    {"dir1", OptDir1, 1},
    {"xrx", OptXrx, 0},
    {"xrx1", OptXrx1, 1},
    {"sun", OptSun, 0},
    {"sun1", OptSun1, 1},
    {"flat", OptFlat, 0},
    {"flat", OptFlat1, 1},
    {"prefix", OptPrefix, 0},
    {"prefix1", OptPrefix1, 1},
    {"noprefix", OptNoPrefix, 0},
    {"noprefix1", OptNoPrefix1, 1},
    {"noserv", OptNoServ, 0},
    {"noserv1", OptNoServ1, 1},
    {"delegation", OptDel, 0},
    {"delegation1", OptDel1, 1},
    {"null", OptNull, 0}
};


static Option
getOption(char *name)
{
    int i;
    for (i = 0; i < sizeof optionTable / sizeof optionTable[0]; i++) {
	if (strcmp(name, optionTable[i].optName) == 0)
            return optionTable[i].opt;
    }
    return OptNull;
}


static boolean
optionNeedsInterface(Option o)
{
    return optionTable[o].ifaceArg;
}


static void
usage(void)
{
    fprintf(stderr, "usage: %s [-dir dir] [-flat] [-...] files\n", 
        programName
        );
    exit(0);
}


static IHandle defaultIHandle = 0;


static IHandle allocIHandle(char* islifname) 
{
    IHandle ih = newContext();
    if (defaultIHandle) {
       *ih = *defaultIHandle;
    }
    ih->islIfName = islifname;
    return ih;
}


static void contextAdd_xrx(IHandle ih)
{
    ih->stubSuffix = "_stub";
    ih->refSuffix = "";
    ih->visiHlp = 0;
    ih->sunHld = 0;
    ih->holderSuffix = "_var";
    ih->helperSuffix = "_help";
    ih->delClassSuffix = "_delegation";
}


static void contextAdd_Sun(IHandle ih)
{
    ih->stubSuffix = "Stub";
    ih->refSuffix = "";
    ih->visiHlp = 1;
    ih->sunHld = 1;
    ih->holderSuffix = "Holder";
    ih->helperSuffix = "Helper";
    ih->delClassSuffix = "_delegation";
}


static void contextAdd_flatness(IHandle ih, boolean flatness)
{
    ih->flatDir = flatness;
}


static void contextAdd_prefix(IHandle ih, char * prefix)
{
    if (strlen(prefix)==0) prefix = 0;
    ih->prefix = prefix;
}


static void contextAdd_dirName(IHandle ih, char * dirName)
{
    if (strlen(dirName)==0) dirName = 0;
    ih->dirName = dirName;
}


int main(int argc, char **argv)
{
    int optind = 1;
    char * ifname;
    IHandle ctx;
    initName();
    defaultIHandle = allocIHandle(0);
      contextAdd_xrx(defaultIHandle);
      defaultIHandle->prefix = "stubs";
      defaultIHandle->dirName = "javastubs";
      setDefaultContext(defaultIHandle);
    programName = iluparser_GetProgramName(argv[0]);
    while (optind < argc && argv[optind][0] == '-') {
	char *opt = (argv[optind++])+1;
	Option o = getOption(opt);
	ifname = 0;
	ctx = defaultIHandle;
	if (optionNeedsInterface(o)) {
            if (optind>=argc) usage();
            ifname = argv[optind++]; 
            ctx = getContextN(ifname); 
	}
	switch (o) {
            case OptDebug:
                debugFlag = TRUE;
                break;
            case OptFlat:
            case OptFlat1:
		contextAdd_flatness(ctx, 1);
		break;
            case OptNoPrefix:
            case OptNoPrefix1:
		contextAdd_prefix(ctx, 0);
		break;
            case OptPrefix:
            case OptPrefix1:
		if (optind>=argc) usage();
		contextAdd_prefix(ctx, argv[optind++]);
		break;
            case OptDir:
            case OptDir1:
		if (optind>=argc) usage();
		contextAdd_dirName(ctx, argv[optind++]);
		break;
            case OptSun:
            case OptSun1:
		contextAdd_Sun(ctx);
		break;
            case OptXrx:
            case OptXrx1:
		contextAdd_xrx(ctx);
		break;
            case OptNoServ:
            case OptNoServ1:
		ctx->noServer = 1;
		break;
            case OptDel:
            case OptDel1:
		ctx->useDel = 1;
		break;
            default:
		usage();
	}
    }
    setDefaultContext(defaultIHandle);
    while (optind < argc) {
	char *fileName = argv[optind++];
	list s;
	if ((s = ParseFile(fileName)) == 0) {
            fatal("Couldn't find or parse %s.\n", fileName);
	}
	fprintf(stderr, "done with parsing\n");
	list_enumerate(s, (EnumProc) generate, NULL);
    }
    return 0;
}


/* end */
