/* util.c */
/* Chris Jacobi, January 30, 1997 1:57 pm PST */

/*
 * Copyright (c) 1996 Xerox Corporation.  All Rights Reserved.  
 * Unlimited use, reproduction, and distribution of this software is
 * permitted.  Any copy of this software must include both the above
 * copyright notice of Xerox Corporation and this paragraph.  Any
 * distribution of this software must comply with all applicable United
 * States export control laws.  This software is made available AS IS,
 * and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
 * PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
 * THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
 * CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
 * XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 */

/* $Id: util.c,v 1.42 1997/01/30 22:29:06 jacobi Exp $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "iluptype.h"
#include "shared.h"
#include "stubops.h"
#include "name.h"
#include "util.h"
#include "context.h"

#ifdef WIN32
/* for mkdir */
#include <direct.h>
#endif

/*public*/
char *programName;

/*public*/
Interface currentIfc;

/*public*/
boolean generatingSkeleton;


PUBLIC char *
booleanImage(boolean value)
{
    return value ? "true" : "false";
}


PUBLIC TypeDescription
baseTypeDescription(Type t)
{
    Type base = t;

    for (;;) {
	if (base->description)
	    return base->description;
	base = base->supertype;
    }
}



static char spacing[] = "                                                ";
/*public*/
char *sol = spacing + sizeof spacing - 1;


PUBLIC void
indent(int levelDelta)
{
    sol -= levelDelta * 2;
    if (sol < spacing || spacing + sizeof spacing <= sol)
	fatal("internal error: indentation");
}


PUBLIC void
newline(void)
{
    printf("\n");
}



PUBLIC void
printBanner(const char *part, Interface ifc)
{
    static char *prefixes[2] = {"//", "//"};

    printf("// %s for \"%s\"\n//\n", part, packageName(ifc));
    iluparser_GenerateBoilerplate(stdout, ifc, programName, prefixes);
    printf("\n");
}


PRIVATE void
print1ActualParam(Argument a, int *count)
/* prints an argument in a method call
 * count for position
 */
{
    if (++*count > 1) {
        printf(", ");
    }
    /* "if (a->direction == In) { " not necessary
     * name of argument independent from a->direction 
     */
    printf("%s", argumentName(a));
}


PUBLIC void
printActualParamList(list args)
{
    int count = 0;
    list_enumerate(args, (EnumProc) print1ActualParam, &count);
}
 
 
PRIVATE void
print1FormalArg(Argument a, int *pCount)
/* prints an argument for a method with its type.
 * pCount is argument position.
 */
{
    if (++*pCount > 1) {
        printf(", ");
    }
    if (a->direction == In) {
	printf("%s %s", typeDeclarator(a->type), argumentName(a));
    }
    else {
	printf("%s %s", holderTypeDeclarator(a->type), argumentName(a));
    }
}


PUBLIC void
printFormalArgList(list args)
{
    int count = 0;
    list_enumerate(args, (EnumProc) print1FormalArg, &count);
}


PRIVATE void
print1ExceptionName(Exception e, int *count)
/* Prints an exception name.
 * count is the position of the exception
 */
{
    if (++*count > 1) {
        printf(", ");
    }
    printf("%s", exceptionName(e));
}


PUBLIC void
printExceptionList(set excp_set)
{
    int count = 0;
    list_enumerate(excp_set, (EnumProc) print1ExceptionName, &count);
}


PUBLIC char *
typeInitializer(TypeKind tk) 
/* returns a string containing the initializer for a type */
{
  switch(tk) {
    case byte8_Type:
    case int16_Type:
    case int32_Type:
    case card16_Type:
    case card32_Type:
    case enumeration_Type:
	return "0";
    case int64_Type:
    case card64_Type:
	return "0L";
    case real32_Type:
        return "0F";
    case real64_Type:
        return "0D";
    case real128_Type:
	return "null";
    case char16_Type:
    case char8_Type:
	return "'\x00'";
    case boolean_Type:
	return "false";
    case object_Type:
    case union_Type:
    case sequence_Type:
    case record_Type:
    case optional_Type:
    case array_Type:
	return "null";
    case alias_Type:
        fatal("typeInitializer shouldn't be alias");
        return "null";
    default:
        fatal("typeInitializer unknown case");
        return "null";
    }
}


PUBLIC boolean
matchPointer(void *p1, void *p2)
{
    return p1 == p2;
}


PRIVATE void
PrintFileHeader(const char * path, Interface interface)
{
    printf("// "); printf(path); printf("\n");
    printBanner("Stubs", interface);
    printf("\n\n");
    printf("package %s;\n\n", packageName(interface));
}

static char * separator = "/";


PRIVATE void
assertExist(const char * dirname)
{
 
#ifndef WIN32
	FILE* f;
    f = fopen(dirname, "rb");
    if (f) {
        fclose(f);
		return;
    }
#else
/* xxx note we just make specific declarations and defines for now
since including <windows.h> causes all kind of grief right now*/
__declspec(dllimport)  unsigned long __stdcall GetFileAttributesA (const char*);
#define FILE_ATTRIBUTE_DIRECTORY        0x00000010
	if ((GetFileAttributesA(dirname) & FILE_ATTRIBUTE_DIRECTORY))
		return;
#endif

	fatal("Directory [%s] does not exist.\n", dirname);
}


PRIVATE char*
leftOfX(const char * s, const char * seps)
{
    char * p;
    p = strpbrk(s, seps); /* returns pointer to xx if found */
    if (p) {
        /* a xx is found */
        int idx = p-s;
        if (idx) {    
            /* leftmost character */
            return 0;
        } else {
            /* normal case */
            char * c = iluparser_Malloc(idx+1);
            strncpy(c, s, idx);
            c[idx] = 0;
            return c;
        } 
    } else {
        /* no xx found */
        return copy(s);
    }
}


PRIVATE char*
rightOfX(const char* s, const char* seps)
{    
    char * p;
    int len = strlen(s);
    if (len<=0) return 0;
    p = strpbrk(s, seps); /* returns pointer to xx if found */
    if (p) {
        /* a xx is found */
        int ridx = p-s+1; /*index of character to the right */
        if (ridx>=len) {
            /* rightmost character */
            return 0;
        } else {
            /* normal case */
            char * c = iluparser_Malloc(len-ridx+1);
            strncpy(c, s+ridx, len-ridx);
            c[len-ridx] = 0;
            return c;
        }
    } else {
        /* no xx found */
        return 0;
    }
}

PRIVATE void
createSubdir(const char * dirname)
{
    int retval;
#ifndef WIN32
    retval = mkdir(dirname, (S_IRWXU | S_IRWXG));
#else
	retval = _mkdir(dirname);
#endif
    if (retval == 0) return; /*new directory has been created*/
    if (errno == EEXIST) return; /*directory already exists*/ 
    fatal("Couldn't create directory [%s].\n", dirname);
}


PRIVATE char *
createDirectoryPath(char * existingDir, const char * newpath)
{
    char* xseps = "./\\";
    while (newpath) {
        char * thisStep = leftOfX(newpath, xseps);
        if (existingDir) {
            existingDir = cat3(existingDir, separator, thisStep);
        } else {
            existingDir = thisStep;
        }
        createSubdir(existingDir);
        newpath = rightOfX(newpath, xseps);
    }
    return existingDir;
}


PUBLIC void
NewJavaFile(Interface interface, const char * className) {
    IHandle ih = getContext(interface);
    char * dirname = ih->dirName;
    char * filename = dotCat(className, "java");
    char * islIfName = ih->islIfName;
    char * prefix = ih->prefix;
    if (dirname==0) fatal("Must specify a directory");
    assertExist(dirname);
    if (! ih->flatDir) {
        dirname = createDirectoryPath(dirname, prefix);
        dirname = createDirectoryPath(dirname, islIfName);
            /* ?? not clear dot is the right separator */
    }
    if (dirname) filename = cat3(dirname, separator, filename);
    if (freopen(filename, "w", stdout) == 0 ) {
        fatal("Couldn't create file %s.\n", filename);
    }
    fprintf(stderr, "writing file %s\n", filename);
    PrintFileHeader(filename, interface);
}


PUBLIC Type myUrType(Type t)
{
    /* ur_type reats redefinitions as well as aliases */
    return ur_type(t);
}


PUBLIC void
printLoopIndent(const unsigned long level)
{
    unsigned long i;

    for (i = 0; i < level; i++)
	printf("  ");
}


PUBLIC void
printDocString(const char* indent, const char* s)
{
    if (s) {
        boolean newline = 1;
        boolean watchout = 0;
        char *p;
        printf("\n%s/**", indent);
        for (p = (char *) s; *p != '\0'; p++) {
            if (newline) {
                printf("\n%s * ", indent);
                newline = 0;
                watchout = 0;
            }
            if (*p == '*') {
                putchar('*');
                watchout = 1;
            } else if (*p == '\n') {
                newline = 1;
                watchout = 1;
            } else if (*p == '/') {
                if (watchout) putchar('.'); else putchar('/');
                watchout = 0;
            } else {
                putchar(*p);
                watchout = 0;
            }
        }
        printf("\n%s */\n", indent);
    }
}


/* end */
