/*
Copyright (c) 1991, 1992, 1993, 1994 Xerox Corporation.  All Rights Reserved.

Unlimited use, reproduction, and distribution of this software is
permitted.  Any copy of this software must include both the above
copyright notice of Xerox Corporation and this paragraph.  Any
distribution of this software must comply with all applicable United
States export control laws.  This software is made available AS IS,
and XEROX CORPORATION DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE, AND NOTWITHSTANDING ANY OTHER
PROVISION CONTAINED HEREIN, ANY LIABILITY FOR DAMAGES RESULTING FROM
THE SOFTWARE OR ITS USE IS EXPRESSLY DISCLAIMED, WHETHER ARISING IN
CONTRACT, TORT (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, EVEN IF
XEROX CORPORATION IS ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.

$Id: util.c,v 1.11 1997/01/08 20:40:09 janssen Exp $
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "iluptype.h"
#include "manifest.h"
#include "name.h"
#include "util.h"

char *		programName;
Interface	currentIfc;
boolean		generatingSkeleton;

void
sysFatal(const char *msg)
{
	fprintf(stderr, "Fatal error: ");
	perror(msg);
	exit(1);
}

void
fatal(const char *fmt, ...)
{
	va_list	args;

	va_start(args, fmt);
	fprintf(stderr, "Fatal error: ");
	vfprintf(stderr, fmt, args);
	va_end(args);
	putc('\n', stderr);
	exit(1);
}

/************************************************************************/

boolean
isPrefixOf(const char *prefix, const char *base)
{
	int	i;

	for (i = 0; prefix[i]; i++)
		if (prefix[i] != base[i])
			return FALSE;
	return TRUE;
}

/************************************************************************/

char *
booleanImage(int value)
{
	return value ? "TRUE" : "FALSE";
}

TypeDescription
baseTypeDescription(Type t)
{
	Type	base	= t;

	for (;;)
	{
		if (base->description)
			return base->description;
		base = base->supertype;
	}
}

char *
simpleTypeName(Type t)
{
	TypeDescription	d	= baseTypeDescription(t);

	switch (d->type)
	{
	case boolean_Type:		return "Boolean";
	case byte_Type:			return "Byte";
	case cardinal_Type:		return "Cardinal";
	case character_Type:		return "Character";
	case enumeration_Type:		return "Enum";
	case integer_Type:		return "Integer";
	case longcardinal_Type:		return "LongCardinal";
	case longinteger_Type:		return "LongInteger";
	case longreal_Type:		return "LongReal";
	case real_Type:			return "Real";
	case shortcardinal_Type:	return "ShortCardinal";
	case shortcharacter_Type:	return "Byte";
	case shortinteger_Type:		return "ShortInteger";
	case shortreal_Type:		return "ShortReal";
	case pickle_Type:		return "Pickle";
	default:			return 0;
	}
}

char *
arraySpecialElemTypeName(Type t)
{
	switch (baseTypeDescription(t)->type)
	{
	case byte_Type:			return "Opaque";
	case shortcharacter_Type:	return "StringVec";
	case character_Type:		return "WStringVec";
	default:			return 0;
	}
}

char *
sequenceSpecialElemTypeName(Type t)
{
	switch (baseTypeDescription(t)->type)
	{
	case byte_Type:			return "Bytes";
	case shortcharacter_Type:	return "String";
	case character_Type:		return "WString";
	default:			return 0;
	}
}

static void
countResultArg(Argument a, int *pCount)
{
	if (a->direction == Out || a->direction == InOut)
		*pCount += 1;
}

int
methodResultCount(Procedure m)
{
	int	count	= m->returnType != 0;

	list_enumerate(m->arguments, (EnumProc) countResultArg, &count);
	return count;
}

/************************************************************************/

static char	spacing[] = "                                                ";
char *		sol = spacing + sizeof spacing - 1;

void
indent(int levelDelta)
{
	sol -= levelDelta * 2;
	if (sol < spacing || spacing + sizeof spacing <= sol)
		fatal("internal error: indentation");
}

void
newline(void)
{
	putchar('\n');
}

/************************************************************************/

void
printBanner(const char *part, Interface ifc)
{
	static char *prefixesForPython[2] = { "#", "#" };

	printf("# %s for \"%s\"\n#\n", part, interface_name(ifc));
	iluparser_GenerateBoilerplate(stdout, ifc, programName,
		prefixesForPython);
	printf("\n");
}

void
printImportIfc(const char *ifcName, boolean skelToo)
{
	if (strcmp(ifcName, "ilu") == 0)
		printf("import %s\n", nameModuleIlu);
	else
	{
		char	buffer[1024];
		char *	name	= getImportName(ifcName, buffer);

		printf("import %s; _imported_modules['%s'] = %s; del %s;",
		       name, name, name, name);
		if (skelToo)
		{
		  newline();
		  printf("import %s__skel; _imported_modules['%s__skel'] = %s__skel; del %s__skel;",
			 name, name, name, name);
		}
		newline();
	}
}

void
printImportTable (void)
{
  printf ("_imported_modules = {};\n");
  printf ("import ilu ; _imported_modules['ilu'] = ilu ; del ilu\n");
}

/************************************************************************/

static void
printArg(Argument a, int *pCount)
{
	if (a->direction == In || a->direction == InOut)
		printf("%s%s", ++*pCount > 1 ? ", " : "", getArgumentName(a));
}

void
printArgList(list argList, int nPrevArgs)
{
	int	count	= nPrevArgs;

	list_enumerate(argList, (EnumProc) printArg, &count);
}

static void
printIfcPrefix(Interface ifc)
{
	if (generatingSkeleton || ifc != currentIfc)
		printf("%s.", getInterfaceName(ifc));
}

void
printClassVarName(Type t, const char *varName)
{
	Type	urt	= ur_type(t);

	printIfcPrefix(urt->interface);
	printf("%s", getTypeName(urt));
	if (varName)
		printf(".%s", varName);
}

void
printExceptionName(Exception e)
{
	printIfcPrefix(e->interface);
	printf("%s", getExceptionName(e));
}

void
printTypeIoFuncName(Type t, const char *prefix)
{
	Type	urt	= ur_type(t);

	printIfcPrefix(urt->interface);
	printf(fmtFuncIo, prefix, getTypeName(urt));
}

/* returns the kernel's notion of the typekind for this ILU type,
 * which is not the same as the parser's notion
 */
int
kernelTypeKind(Type t)
{
  enum _ilu_TypeKind_e {
    ilu_byte_tk,
    ilu_boolean_tk,
    ilu_character_tk,
    ilu_shortcharacter_tk,
    ilu_shortinteger_tk,
    ilu_integer_tk,
    ilu_longinteger_tk,
    ilu_shortcardinal_tk,
    ilu_cardinal_tk,
    ilu_longcardinal_tk,
    ilu_real_tk,
    ilu_shortreal_tk,
    ilu_longreal_tk,
    ilu_object_tk,
    ilu_pipe_tk,
    ilu_optional_tk,
    ilu_alias_tk,
    ilu_union_tk,
    ilu_sequence_tk,
    ilu_record_tk,
    ilu_array_tk,
    ilu_enumeration_tk,
    ilu_pickle_tk
    };

  switch (type_kind(t)) {
  case byte_Type:
    return (int) ilu_byte_tk;

  case boolean_Type:
    return (int) ilu_boolean_tk;

  case character_Type:
    return (int) ilu_character_tk;

  case shortcharacter_Type:
    return (int) ilu_shortcharacter_tk;

  case shortinteger_Type:
    return (int) ilu_shortinteger_tk;

  case integer_Type:
    return (int) ilu_integer_tk;

  case longinteger_Type:
    return (int) ilu_longinteger_tk;

  case shortcardinal_Type:
    return (int) ilu_shortcardinal_tk;

  case cardinal_Type:
    return (int) ilu_cardinal_tk;

  case longcardinal_Type:
    return (int) ilu_longcardinal_tk;

  case real_Type:
    return (int) ilu_real_tk;

  case shortreal_Type:
    return (int) ilu_shortreal_tk;

  case longreal_Type:
    return (int) ilu_longreal_tk;

  case pickle_Type:
    return (int) ilu_pickle_tk;

  case object_Type:
    return (int) ilu_object_tk;

  case pipe_Type:
    return (int) ilu_pipe_tk;

  case optional_Type:
    return (int) ilu_optional_tk;

  case alias_Type:
    return (int) ilu_alias_tk;

  case union_Type:
    return (int) ilu_union_tk;

  case sequence_Type:
    return (int) ilu_sequence_tk;

  case record_Type:
    return (int) ilu_record_tk;

  case array_Type:
    return (int) ilu_array_tk;

  case enumeration_Type:
    return (int) ilu_enumeration_tk;

  default:
    return 0;
  }
}
